/*
 * SELM_CouplingOperator_StaggeredGrid1.cpp
 *
 * Paul J. Atzberger
 * http://atzberger.org/
 *
 */

#include "SELM_CouplingOperator_StaggeredGrid1.h"
#include "SELM_CouplingOperator_Types.h"
#include "math.h"
#include <cstdlib>
#include <math.h>
#include "stdio.h"
#include "stdlib.h"
#include "string.h"

using namespace std;

namespace LAMMPS_NS {

// ============ constants definition ===========
int   SELM_CouplingOperator_StaggeredGrid1::TYPE     = SELM_CouplingOperator_Types::TYPE_StaggeredGrid1;
char *SELM_CouplingOperator_StaggeredGrid1::TYPE_STR = SELM_CouplingOperator_Types::TYPE_STR_StaggeredGrid1;
const double SELM_CouplingOperator_StaggeredGrid1::UNIT_pi=3.141592653589793;
const char* SELM_CouplingOperator_StaggeredGrid1::error_str_code="SELM_CouplingOperator_StaggeredGrid1.h";

int SELM_CouplingOperator_StaggeredGrid1::usePeskin  = 1;

SELM_CouplingOperator_StaggeredGrid1::SELM_CouplingOperator_StaggeredGrid1() : SELM_CouplingOperator()
{
  /* generic initializations for this class */
  init();
}
SELM_CouplingOperator_StaggeredGrid1::SELM_CouplingOperator_StaggeredGrid1(int narg, char **arg) : SELM_CouplingOperator(narg, arg)
{
  const char *error_str_func = "SELM_CouplingOperator_StaggeredGrid1()";
  /* generic initializations for this class */
  init();
}



SELM_CouplingOperator_StaggeredGrid1::~SELM_CouplingOperator_StaggeredGrid1() {
}

void SELM_CouplingOperator_StaggeredGrid1::packageError(int code, void *extras) {

  const char* error_str_func="packageError()";

  stringstream message;
  message << "Code = " << code << endl;
  SELM_Package::packageError(error_str_code, error_str_func, message);
}


void SELM_CouplingOperator_StaggeredGrid1::init() {
  type                       = SELM_CouplingOperator_Types::TYPE_StaggeredGrid1;
  strcpy(typeStr, SELM_CouplingOperator_Types::TYPE_STR_StaggeredGrid1);

  upsilon_0                  = 0.0;

  flagWriteSimulationData    = 0;
  saveSkipSimulationData     = 0;
}

void SELM_CouplingOperator_StaggeredGrid1::computeOperatorGamma(SELM_Lagrangian *Lpt,
            SELM_Eulerian *Ept, double** u_in, double* v_out){

	// multiply Peskin Delta function at position x
	// f_in is the Lagrangian force at position x,
	// f_out is Eulerian force at face center (staggered grid)
    const char* error_str_func="computeOperatorGamma()";

	int k1, k2, k3, d, i;
	int n, p1, p2, p3, N12;
	double r[3], delta;
	int p[3], i1, i2, i3;


	int num_dim;
	int numMeshPtsPerDir[3];
	double meshDeltaX;
	double meshCenterX0[3];
	double deltaX3;

	int numControlPts;       // Number of control points.
	double* x;

	//Eulerian data
	if(Ept->type==SELM_Eulerian_Types::TYPE_StaggeredGrid1) {
		SELM_Eulerian_StaggeredGrid1* Ept_data=(SELM_Eulerian_StaggeredGrid1*) Ept;
		num_dim=Ept_data->num_dim;
		for(i=0; i<num_dim; i++) {
		  numMeshPtsPerDir[i]=Ept_data->numMeshPtsPerDir[i];
		  meshCenterX0[i]=Ept_data->meshCenterX0[i];
		}
		meshDeltaX=Ept_data->meshDeltaX;
	} else {
		stringstream message;
		message << "This method is only implemented on staggered grid" << endl;
		SELM_Package::packageError(error_str_code, error_str_func, message);
	}


	//Lagrangian data
	if(Lpt->type==SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {
		SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE* Lpt_data=(SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE*) Lpt;
		numControlPts=Lpt_data->numControlPts;       // Number of control points.
		x=new double[3*numControlPts];
		// shift the computational domain to [0,a]x[0,b]x[0,c]
		for(i=0; i<numControlPts; i++) {
			i1=3*i;
			i2=i1+1;
			i3=i2+1;
			x[i1]=Lpt_data->ptsX[i1]-(meshCenterX0[0]-numMeshPtsPerDir[0]*meshDeltaX/2);
			x[i2]=Lpt_data->ptsX[i2]-(meshCenterX0[1]-numMeshPtsPerDir[1]*meshDeltaX/2);
			x[i3]=Lpt_data->ptsX[i3]-(meshCenterX0[2]-numMeshPtsPerDir[2]*meshDeltaX/2);
		}
	} else if(Lpt->type==SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {
		SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE* Lpt_data=(SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE*) Lpt;
		numControlPts=Lpt_data->numControlPts;       // Number of control points.
		x=new double[3*numControlPts];
		// shift the computational domain to [0,a]x[0,b]x[0,c]
		for(i=0; i<numControlPts; i++) {
			i1=3*i;
			i2=i1+1;
			i3=i2+1;
			x[i1]=Lpt_data->ptsX[i1]-(meshCenterX0[0]-numMeshPtsPerDir[0]*meshDeltaX/2);
			x[i2]=Lpt_data->ptsX[i2]-(meshCenterX0[1]-numMeshPtsPerDir[1]*meshDeltaX/2);
			x[i3]=Lpt_data->ptsX[i3]-(meshCenterX0[2]-numMeshPtsPerDir[2]*meshDeltaX/2);
		}
	} else{
		stringstream message;
		message << "SELM_Lagrangian type is not recognized" << endl;
		SELM_Package::packageError(error_str_code, error_str_func, message);
	}

	//multiply Peskin Delta function at position x
	// u_in is the Eulerian velocity on staggered grid system,
	// v_out is Lagrangian velocity at position x

	N12=numMeshPtsPerDir[0]*numMeshPtsPerDir[1];
	deltaX3=meshDeltaX*meshDeltaX*meshDeltaX;

	for(i=0; i<numControlPts; i++) {
		// find the grid cell position x lines in, since all the values are face centered,
		// u is shifted by half meshDeltaX on y, z direction
		// v is shifted by half meshDeltaX on x, z direction
		// w is shifted by half meshDeltaX on z, y direction
		i1=3*i;
		i2=i1+1;
		i3=i2+1;

		v_out[i1]=0;
		v_out[i2]=0;
		v_out[i3]=0;
		// u component
		p1=(int)floor(x[i1]/meshDeltaX);
		p2=(int)floor(x[i2]/meshDeltaX-0.5);
		p3=(int)floor(x[i3]/meshDeltaX-0.5);
		if(usePeskin) {
		  for(k3=-2; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
				for(k1=-2; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=fabs(x[i1]/meshDeltaX-p[0]);
						r[1]=fabs(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+0.5));
						if((r[0]<2.0) && (r[1]<2.0) && (r[2]<2.0)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]+=numMeshPtsPerDir[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]-=numMeshPtsPerDir[d];
								if(r[d]<1.0) {
									delta=delta*(3.0-2.0*r[d]+sqrt(1.0+4.0*r[d]*(1.0-r[d])))/8.0;
								} else if(r[d]<2.0) {
									delta=delta*(5.0-2.0*r[d]-sqrt(-7.0+4.0*r[d]*(3.0-r[d])))/8.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							v_out[i1]=v_out[i1]+u_in[0][n]*delta/deltaX3;
						}
					}
				}
			 }
		  }
		  //v component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX);
		  p3=(int)floor(x[i3]/meshDeltaX-0.5);
		  for(k3=-2; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
				for(k1=-2; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=fabs(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=fabs(x[i2]/meshDeltaX-p[1]);
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+0.5));
						if((r[0]<2.0) && (r[1]<2.0) && (r[2]<2.0)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<1.0) {
									delta=delta*(3.0-2.0*r[d]+sqrt(1.0+4.0*r[d]*(1.0-r[d])))/8.0;
								} else if(r[d]<2.0) {
									delta=delta*(5.0-2.0*r[d]-sqrt(-7.0+4.0*r[d]*(3.0-r[d])))/8.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							v_out[i2]=v_out[i2]+u_in[1][n]*delta/deltaX3;
						}
					}
				}
			 }
		  }
		  //w component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX-0.5);
		  p3=(int)floor(x[i3]/meshDeltaX-1.0);
		  for(k3=-2; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
				for(k1=-2; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]-1) {
						r[0]=fabs(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=fabs(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+1));
						if((r[0]<2.0) && (r[1]<2.0) && (r[2]<2.0)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<1.0) {
									delta=delta*(3.0-2.0*r[d]+sqrt(1.0+4.0*r[d]*(1.0-r[d])))/8.0;
								} else if(r[d]<2.0) {
									delta=delta*(5.0-2.0*r[d]-sqrt(-7.0+4.0*r[d]*(3.0-r[d])))/8.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							v_out[i3]=v_out[i3]+u_in[2][n]*delta/deltaX3;
						}
					}
				}
			 }
		  }
		} else {
		  for(k3=-1; k3<3; k3++) {
			 for(k2=-1; k2<3; k2++) {
				for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=fabs(x[i1]/meshDeltaX-p[0]);
						r[1]=fabs(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+0.5));
						if((r[0]<1.5) && (r[1]<1.5) && (r[2]<1.5)) {
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]+=numMeshPtsPerDir[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]-=numMeshPtsPerDir[d];
								if(r[d]<0.5) {
									delta=delta*(1.0+sqrt(1.0-3.0*r[d]*r[d]))/3.0;
								} else if(r[d]<1.5) {
									delta=delta*(5.0-3.0*r[d]-sqrt(1.0-3.0*(1.0-r[d])*(1.0-r[d])))/6.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							v_out[i1]=v_out[i1]+u_in[0][n]*delta/deltaX3;
						}
					}
				}
			 }
		  }
		  //v component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX);
		  p3=(int)floor(x[i3]/meshDeltaX-0.5);
		  for(k3=-1; k3<3; k3++) {
			 for(k2=-1; k2<3; k2++) {
				for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=fabs(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=fabs(x[i2]/meshDeltaX-p[1]);
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+0.5));
						if((r[0]<1.5) && (r[1]<1.5) && (r[2]<1.5)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<0.5) {
									delta=delta*(1.0+sqrt(1.0-3.0*r[d]*r[d]))/3.0;
								} else if(r[d]<1.5) {
									delta=delta*(5.0-3.0*r[d]-sqrt(1.0-3.0*(1.0-r[d])*(1.0-r[d])))/6.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							v_out[i2]=v_out[i2]+u_in[1][n]*delta/deltaX3;
						}
					}
				}
			 }
		  }
		  //w component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX-0.5);
		  p3=(int)floor(x[i3]/meshDeltaX-1.0);
		  for(k3=-1; k3<3; k3++) {
			 for(k2=-1; k2<3; k2++) {
				for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]-1) {
						r[0]=fabs(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=fabs(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+1));
						if((r[0]<1.5) && (r[1]<1.5) && (r[2]<1.5)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<0.5) {
									delta=delta*(1.0+sqrt(1.0-3.0*r[d]*r[d]))/3.0;
								} else if(r[d]<1.5) {
									delta=delta*(5.0-3.0*r[d]-sqrt(1.0-3.0*(1.0-r[d])*(1.0-r[d])))/6.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							v_out[i3]=v_out[i3]+u_in[2][n]*delta/deltaX3;
						}
					}
				}
			 }
		  }
		} // end of if(usePeskin)
	}

	delete[] x;
}

void SELM_CouplingOperator_StaggeredGrid1::computeOperatorLambda(SELM_Lagrangian  *Lpt,
            SELM_Eulerian *Ept, double* f_in, double** f_out){
	// multiply Peskin Delta function at position x
	// f_in is the Lagrangian force at position x,
	// f_out is Eulerian force at face center (staggered grid)
    const char* error_str_func="computeOperatorLambda()";

	int k1, k2, k3, d, i;
	int n, p1, p2, p3, N12, Ntotal;
	double r[3], delta;
	int p[3], i1, i2, i3;

	//Eulerian data
	int num_dim;
	int numMeshPtsPerDir[3];
	double meshDeltaX;
	double meshCenterX0[3];
	double deltaX3;


	int numControlPts;       // Number of control points.
	double* x;

	if(Ept->type==SELM_Eulerian_Types::TYPE_StaggeredGrid1) {
		SELM_Eulerian_StaggeredGrid1* Ept_data=(SELM_Eulerian_StaggeredGrid1*) Ept;
		num_dim=Ept_data->num_dim;
		for(i=0; i<num_dim; i++) {
		  numMeshPtsPerDir[i]=Ept_data->numMeshPtsPerDir[i];
		  meshCenterX0[i]=Ept_data->meshCenterX0[i];
		}
		meshDeltaX=Ept_data->meshDeltaX;
	} else {
		stringstream message;
		message << "This method is only implemented on staggered grid" << endl;
		SELM_Package::packageError(error_str_code, error_str_func, message);
	}

	//Lagrangian data
	if(Lpt->type==SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {
		SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE* Lpt_data=(SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE*) Lpt;
		numControlPts=Lpt_data->numControlPts;       // Number of control points.
		x=new double[3*numControlPts];
		// shift the computational domain to [0,a]x[0,b]x[0,c]
		for(i=0; i<numControlPts; i++) {
			i1=3*i;
			i2=i1+1;
			i3=i2+1;
			x[i1]=Lpt_data->ptsX[i1]-(meshCenterX0[0]-numMeshPtsPerDir[0]*meshDeltaX/2);
			x[i2]=Lpt_data->ptsX[i2]-(meshCenterX0[1]-numMeshPtsPerDir[1]*meshDeltaX/2);
			x[i3]=Lpt_data->ptsX[i3]-(meshCenterX0[2]-numMeshPtsPerDir[2]*meshDeltaX/2);
		}
	} else if(Lpt->type==SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {
		SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE* Lpt_data=(SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE*) Lpt;
		numControlPts=Lpt_data->numControlPts;       // Number of control points.
		x=new double[3*numControlPts];
		// shift the computational domain to [0,a]x[0,b]x[0,c]
		for(i=0; i<numControlPts; i++) {
			i1=3*i;
			i2=i1+1;
			i3=i2+1;
			x[i1]=Lpt_data->ptsX[i1]-(meshCenterX0[0]-numMeshPtsPerDir[0]*meshDeltaX/2);
			x[i2]=Lpt_data->ptsX[i2]-(meshCenterX0[1]-numMeshPtsPerDir[1]*meshDeltaX/2);
			x[i3]=Lpt_data->ptsX[i3]-(meshCenterX0[2]-numMeshPtsPerDir[2]*meshDeltaX/2);
		}
	} else {
		stringstream message;
		message << "SELM_Lagrangian type is not recognized" << endl;
		SELM_Package::packageError(error_str_code, error_str_func, message);
	}


	N12=numMeshPtsPerDir[0]*numMeshPtsPerDir[1];
	Ntotal=N12*numMeshPtsPerDir[2];
	deltaX3=meshDeltaX*meshDeltaX*meshDeltaX;
/*
	for(i=0; i<Ntotal-N12; i++)
		f_out[2][i]=0.0;
	for (d=0; d<num_dim-1; d++) {
		for (i=0; i<Ntotal; i++) {
			f_out[d][i]=0.0;
		}
	}*/
	for(i=0; i<numControlPts; i++) {
		// find the grid cell position x lines in, since all the values are face centered,
		// f_u is shifted by half meshDeltaX on y, z direction
		// f_v is shifted by half meshDeltaX on x, z direction
		// f_w is shifted by half meshDeltaX on z, y direction
		i1=3*i;
		i2=i1+1;
		i3=i2+1;
		// f_u component
		p1=(int)floor(x[i1]/meshDeltaX);
		p2=(int)floor(x[i2]/meshDeltaX-0.5);
		p3=(int)floor(x[i3]/meshDeltaX-0.5);
		if(usePeskin) {
  		  for(k3=-2; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
			  	for(k1=-2; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=fabs(x[i1]/meshDeltaX-p[0]);
						r[1]=fabs(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+0.5));
						if((r[0]<2.0) && (r[1]<2.0) && (r[2]<2.0)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<1.0) {
									delta=delta*(3.0-2.0*r[d]+sqrt(1.0+4.0*r[d]*(1.0-r[d])))/8.0;
								} else if(r[d]<2.0) {
									delta=delta*(5.0-2.0*r[d]-sqrt(-7.0+4.0*r[d]*(3.0-r[d])))/8.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							f_out[0][n]=f_out[0][n]+f_in[i1]*delta/deltaX3;
						}
					}
				}
			 }
		  }

  		  // f_v component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX);
		  p3=(int)floor(x[i3]/meshDeltaX-0.5);
		  for(k3=-2; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
				for(k1=-2; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=fabs(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=fabs(x[i2]/meshDeltaX-p[1]);
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+0.5));
						if((r[0]<2.0) && (r[1]<2.0) && (r[2]<2.0)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<1.0) {
									delta=delta*(3.0-2.0*r[d]+sqrt(1.0+4.0*r[d]*(1.0-r[d])))/8.0;
								} else if(r[d]<2.0) {
									delta=delta*(5.0-2.0*r[d]-sqrt(-7.0+4.0*r[d]*(3.0-r[d])))/8.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							f_out[1][n]=f_out[1][n]+f_in[i2]*delta/deltaX3;
						}
					}
				}
			 }
		  }
		  // f_w component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX-0.5);
		  p3=(int)floor(x[i3]/meshDeltaX-1.0);
		  for(k3=-2; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
				for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]-1) {
						r[0]=fabs(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=fabs(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+1.0));
						if((r[0]<2.0) && (r[1]<2.0) && (r[2]<2.0)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<1.0) {
									delta=delta*(3.0-2.0*r[d]+sqrt(1.0+4.0*r[d]*(1.0-r[d])))/8.0;
								} else if(r[d]<2.0) {
									delta=delta*(5.0-2.0*r[d]-sqrt(-7.0+4.0*r[d]*(3.0-r[d])))/8.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							f_out[2][n]=f_out[2][n]+f_in[i3]*delta/deltaX3;
						}
					}
				}
			 }
		  }
		} else {
  		  for(k3=-1; k3<3; k3++) {
			 for(k2=-1; k2<3; k2++) {
			  	for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=fabs(x[i1]/meshDeltaX-p[0]);
						r[1]=fabs(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+0.5));
						if((r[0]<1.5) && (r[1]<1.5) && (r[2]<1.5)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<0.5) {
									delta=delta*(1.0+sqrt(1.0-3.0*r[d]*r[d]))/3.0;
								} else if(r[d]<1.5) {
									delta=delta*(5.0-3.0*r[d]-sqrt(1.0-3.0*(1.0-r[d])*(1.0-r[d])))/6.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							f_out[0][n]=f_out[0][n]+f_in[i1]*delta/deltaX3;
						}
					}
				}
			 }
		  }

	  	  // f_v component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX);
		  p3=(int)floor(x[i3]/meshDeltaX-0.5);
		  for(k3=-1; k3<3; k3++) {
			 for(k2=-1; k2<3; k2++) {
				for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=fabs(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=fabs(x[i2]/meshDeltaX-p[1]);
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+0.5));
						if((r[0]<1.5) && (r[1]<1.5) && (r[2]<1.5)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<0.5) {
									delta=delta*(1.0+sqrt(1.0-3.0*r[d]*r[d]))/3.0;
								} else if(r[d]<1.5) {
									delta=delta*(5.0-3.0*r[d]-sqrt(1.0-3.0*(1.0-r[d])*(1.0-r[d])))/6.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							f_out[1][n]=f_out[1][n]+f_in[i2]*delta/deltaX3;
						}
					}
				}
			 }
		  }
		  // f_w component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX-0.5);
		  p3=(int)floor(x[i3]/meshDeltaX-1.0);
		  for(k3=-1; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
				for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]-1) {
						r[0]=fabs(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=fabs(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+1.0));
						if((r[0]<1.5) && (r[1]<1.5) && (r[2]<1.5)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<0.5) {
									delta=delta*(1.0+sqrt(1.0-3.0*r[d]*r[d]))/3.0;
								} else if(r[d]<1.5) {
									delta=delta*(5.0-3.0*r[d]-sqrt(1.0-3.0*(1.0-r[d])*(1.0-r[d])))/6.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							f_out[2][n]=f_out[2][n]+f_in[i3]*delta/deltaX3;
						}
					}
				}
			 }
		  }
	   }  // end of if(usePeskin)
	}

	delete[] x;
}

void SELM_CouplingOperator_StaggeredGrid1::computeOperatorLambda(int numControlPts, double* x_in,
            SELM_Eulerian *Ept, double* f_in, double** f_out){
	// multiply Peskin Delta function at position x
	// f_in is the Lagrangian force at position x,
	// f_out is Eulerian force at face center (staggered grid)
    const char* error_str_func="computeOperatorLambda()";

	int k1, k2, k3, d, i;
	int n, p1, p2, p3, N12, Ntotal;
	double r[3];
	double delta;
	int p[3], i1, i2, i3;

	//Eulerian data
	int num_dim;
	int numMeshPtsPerDir[3];
	double meshDeltaX;
	double meshCenterX0[3];
	double deltaX3;

	if(Ept->type==SELM_Eulerian_Types::TYPE_StaggeredGrid1) {
		SELM_Eulerian_StaggeredGrid1* Ept_data=(SELM_Eulerian_StaggeredGrid1*) Ept;
		num_dim=Ept_data->num_dim;
		for(i=0; i<num_dim; i++) {
		  numMeshPtsPerDir[i]=Ept_data->numMeshPtsPerDir[i];
		  meshCenterX0[i]=Ept_data->meshCenterX0[i];
		}
		meshDeltaX=Ept_data->meshDeltaX;
	} else {
		stringstream message;
		message << "This method is only implemented on staggered grid" << endl;
		SELM_Package::packageError(error_str_code, error_str_func, message);
	}

	// shift the computational domain to [0,a]x[0,b]x[0,c]
	double* x;
	x=new double[3*numControlPts];

	for(i=0; i<numControlPts; i++) {
		i1=3*i;
		i2=i1+1;
		i3=i2+1;
		x[i1]=x_in[i1]-(meshCenterX0[0]-numMeshPtsPerDir[0]*meshDeltaX/2);
		x[i2]=x_in[i2]-(meshCenterX0[1]-numMeshPtsPerDir[1]*meshDeltaX/2);
		x[i3]=x_in[i3]-(meshCenterX0[2]-numMeshPtsPerDir[2]*meshDeltaX/2);
	}

	N12=numMeshPtsPerDir[0]*numMeshPtsPerDir[1];
	Ntotal=N12*numMeshPtsPerDir[2];
	deltaX3=meshDeltaX*meshDeltaX*meshDeltaX;
/*
	for(i=0; i<Ntotal-N12; i++)
		f_out[2][i]=0.0;
	for (d=0; d<num_dim-1; d++) {
		for (i=0; i<Ntotal; i++) {
			f_out[d][i]=0.0;
		}
	}*/
	for(i=0; i<numControlPts; i++) {
		// find the grid cell position x lines in, since all the values are face centered,
		// f_u is shifted by half meshDeltaX on y, z direction
		// f_v is shifted by half meshDeltaX on x, z direction
		// f_w is shifted by half meshDeltaX on z, y direction
		i1=3*i;
		i2=i1+1;
		i3=i2+1;
		// f_u component
		p1=(int)floor(x[i1]/meshDeltaX);
		p2=(int)floor(x[i2]/meshDeltaX-0.5);
		p3=(int)floor(x[i3]/meshDeltaX-0.5);
		if(usePeskin) {
  		  for(k3=-2; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
			  	for(k1=-2; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=fabs(x[i1]/meshDeltaX-p[0]);
						r[1]=fabs(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+0.5));
						if((r[0]<2.0) && (r[1]<2.0) && (r[2]<2.0)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<1.0) {
									delta=delta*(3.0-2.0*r[d]+sqrt(1.0+4.0*r[d]*(1.0-r[d])))/8.0;
								} else if(r[d]<2.0) {
									delta=delta*(5.0-2.0*r[d]-sqrt(-7.0+4.0*r[d]*(3.0-r[d])))/8.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							f_out[0][n]=f_out[0][n]+f_in[i1]*delta/deltaX3;
						}
					}
				}
			 }
		  }

  		  // f_v component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX);
		  p3=(int)floor(x[i3]/meshDeltaX-0.5);
		  for(k3=-2; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
				for(k1=-2; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=fabs(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=fabs(x[i2]/meshDeltaX-p[1]);
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+0.5));
						if((r[0]<2.0) && (r[1]<2.0) && (r[2]<2.0)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<1.0) {
									delta=delta*(3.0-2.0*r[d]+sqrt(1.0+4.0*r[d]*(1.0-r[d])))/8.0;
								} else if(r[d]<2.0) {
									delta=delta*(5.0-2.0*r[d]-sqrt(-7.0+4.0*r[d]*(3.0-r[d])))/8.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							f_out[1][n]=f_out[1][n]+f_in[i2]*delta/deltaX3;
						}
					}
				}
			 }
		  }
		  // f_w component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX-0.5);
		  p3=(int)floor(x[i3]/meshDeltaX-1.0);
		  for(k3=-2; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
				for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]-1) {
						r[0]=fabs(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=fabs(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+1.0));
						if((r[0]<2.0) && (r[1]<2.0) && (r[2]<2.0)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<1.0) {
									delta=delta*(3.0-2.0*r[d]+sqrt(1.0+4.0*r[d]*(1.0-r[d])))/8.0;
								} else if(r[d]<2.0) {
									delta=delta*(5.0-2.0*r[d]-sqrt(-7.0+4.0*r[d]*(3.0-r[d])))/8.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							f_out[2][n]=f_out[2][n]+f_in[i3]*delta/deltaX3;
						}
					}
				}
			 }
		  }
		} else {
  		  for(k3=-1; k3<3; k3++) {
			 for(k2=-1; k2<3; k2++) {
			  	for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=fabs(x[i1]/meshDeltaX-p[0]);
						r[1]=fabs(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+0.5));
						if((r[0]<1.5) && (r[1]<1.5) && (r[2]<1.5)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<0.5) {
									delta=delta*(1.0+sqrt(1.0-3.0*r[d]*r[d]))/3.0;
								} else if(r[d]<1.5) {
									delta=delta*(5.0-3.0*r[d]-sqrt(1.0-3.0*(1.0-r[d])*(1.0-r[d])))/6.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							f_out[0][n]=f_out[0][n]+f_in[i1]*delta/deltaX3;
						}
					}
				}
			 }
		  }

	  	  // f_v component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX);
		  p3=(int)floor(x[i3]/meshDeltaX-0.5);
		  for(k3=-1; k3<3; k3++) {
			 for(k2=-1; k2<3; k2++) {
				for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=fabs(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=fabs(x[i2]/meshDeltaX-p[1]);
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+0.5));
						if((r[0]<1.5) && (r[1]<1.5) && (r[2]<1.5)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<0.5) {
									delta=delta*(1.0+sqrt(1.0-3.0*r[d]*r[d]))/3.0;
								} else if(r[d]<1.5) {
									delta=delta*(5.0-3.0*r[d]-sqrt(1.0-3.0*(1.0-r[d])*(1.0-r[d])))/6.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							f_out[1][n]=f_out[1][n]+f_in[i2]*delta/deltaX3;
						}
					}
				}
			 }
		  }
		  // f_w component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX-0.5);
		  p3=(int)floor(x[i3]/meshDeltaX-1.0);
		  for(k3=-1; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
				for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]-1) {
						r[0]=fabs(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=fabs(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=fabs(x[i3]/meshDeltaX-(p[2]+1.0));
						if((r[0]<1.5) && (r[1]<1.5) && (r[2]<1.5)){
							delta=1.0;
							for(d=0; d<num_dim; d++) {
								if (p[d]<0) p[d]=numMeshPtsPerDir[d]+p[d];
								if (p[d]>numMeshPtsPerDir[d]-1) p[d]=p[d]-numMeshPtsPerDir[d];
								if(r[d]<0.5) {
									delta=delta*(1.0+sqrt(1.0-3.0*r[d]*r[d]))/3.0;
								} else if(r[d]<1.5) {
									delta=delta*(5.0-3.0*r[d]-sqrt(1.0-3.0*(1.0-r[d])*(1.0-r[d])))/6.0;
								}
							}
							n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
							f_out[2][n]=f_out[2][n]+f_in[i3]*delta/deltaX3;
						}
					}
				}
			 }
		  }
	   }  // end of if(usePeskin)
	}

	delete[] x;
}

void SELM_CouplingOperator_StaggeredGrid1::computeOperatorUpsilon(SELM_Lagrangian  *Lpt,
		double mu, double meshDeltaX, double* v_in, double* f_out){
	// Stokes' law:   F_out = 6\pi*mu*R*v
	// Lagrangian description of velocity and force
	// Peskin Delta function radius =2.0r, Roma Delta function radius=1.5r

	int i, i1, i2, i3;
	double c;
	double R;

	double numControlPts;

	const char* error_str_func="computeOperatorUpsilon()";

	//if(usePeskin) {  /* WARNING!!! need to make this come from outside parameter (XML) */
	//	R=2.0*meshDeltaX;
	//} else {
	//	R=1.5*meshDeltaX;
	//}
	//c=6.0*UNIT_pi*R*mu;
//	c=4.5*mu/(2.0*R*R);
	c = this->upsilon_0;

	if(Lpt->type==SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {
		SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE* Lpt_data=(SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE*) Lpt;
		numControlPts=Lpt_data->numControlPts;
	} else if(Lpt->type==SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {
		SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE* Lpt_data=(SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE*) Lpt;
		numControlPts=Lpt_data->numControlPts;
	} else {
		stringstream message;
		message << "SELM_Lagrangian type is not recognized" << endl;
		SELM_Package::packageError(error_str_code, error_str_func, message);
	}

	for (i=0; i<numControlPts; i++) {
		i1=3*i;
		i2=i1+1;
		i3=i2+1;
		f_out[i1]=c*v_in[i1];
		f_out[i2]=c*v_in[i2];
		f_out[i3]=c*v_in[i3];
	}
}
void SELM_CouplingOperator_StaggeredGrid1::writeSimulationDataToDisk(char *baseFilename, int timeIndex) {

  const char *error_str_func = "writeSimulationDataToDisk()";

  FILE *fid;
  char  filename[1000];

  /* open the file for writing the data */
  sprintf(filename, "SELM_CouplingOperator%.9d._%s", timeIndex, typeStr);
  fid = fopen(filename,"w");

  if (fid == NULL) {
    printf("ERROR: %s : %s \n", error_str_code, error_str_func);
    printf("Could not open file, error occured. \n");
    printf("  filename = %s \n", filename);
    packageError(1, 0);
  }

  fprintf(fid, "-- SELM_CouplingOperator_LAMMPS_SHEAR_UNIFORM1_FFTW3_TABLE1 : Simulation Data -- \n");
  fprintf(fid, "\n");

  /* close the file */
  fclose(fid);
}
void SELM_CouplingOperator_StaggeredGrid1::setup() {

  const char *error_str_func = "setup()";

}

void SELM_CouplingOperator_StaggeredGrid1::parse_ParameterFile(char *filename) {
  const char* error_str_func="parse_ParameterFile()";
  stringstream message;
  message << "This method is not implemented on staggered grid" << endl;
  SELM_Package::packageError(error_str_code, error_str_func, message);
}

int SELM_CouplingOperator_StaggeredGrid1::getOperatorTypeFromStr(char *operatorTypeStr) {
  const char* error_str_func="getOperatorTypeFromStr()";
  stringstream message;
  message << "This method is not implemented on staggered grid" << endl;
  SELM_Package::packageError(error_str_code, error_str_func, message);
}

void SELM_CouplingOperator_StaggeredGrid1::computeOperator(char *couplingOpTypeStr,
  SELM_Lagrangian *SELM_LagrangianData, SELM_Eulerian   *SELM_EulerianData){
  const char* error_str_func="computeOperator()";
  stringstream message;
  message << "This method is not implemented on staggered grid" << endl;
  SELM_Package::packageError(error_str_code, error_str_func, message);
}

void SELM_CouplingOperator_StaggeredGrid1::gradientLambda(int numControlPts, double* x_in,
        double KbT,   SELM_Eulerian *Ept, double** f_out) {
    const char* error_str_func="gradientLambda()";

	int k1, k2, k3, d, i;
	int n, p1, p2, p3, N12, Ntotal;
	double r[3];
	double delta;
	int p[3], i1, i2, i3;
	double meshDeltaX4th;

	//Eulerian data
	int num_dim;
	int numMeshPtsPerDir[3];
	double meshDeltaX;
	double meshCenterX0[3];

	if(Ept->type==SELM_Eulerian_Types::TYPE_StaggeredGrid1) {
		SELM_Eulerian_StaggeredGrid1* Ept_data=(SELM_Eulerian_StaggeredGrid1*) Ept;
		num_dim=Ept_data->num_dim;
		for(i=0; i<num_dim; i++) {
		  numMeshPtsPerDir[i]=Ept_data->numMeshPtsPerDir[i];
		  meshCenterX0[i]=Ept_data->meshCenterX0[i];
		}
		meshDeltaX=Ept_data->meshDeltaX;
		meshDeltaX4th=meshDeltaX*meshDeltaX*meshDeltaX*meshDeltaX*KbT;
	} else {
		stringstream message;
		message << "This method is only implemented on staggered grid" << endl;
		SELM_Package::packageError(error_str_code, error_str_func, message);
	}

	// shift the computational domain to [0,a]x[0,b]x[0,c]
	double* x;
	x=new double[3*numControlPts];

	for(i=0; i<numControlPts; i++) {
		i1=3*i;
		i2=i1+1;
		i3=i2+1;
		x[i1]=x_in[i1]-(meshCenterX0[0]-numMeshPtsPerDir[0]*meshDeltaX/2);
		x[i2]=x_in[i2]-(meshCenterX0[1]-numMeshPtsPerDir[1]*meshDeltaX/2);
		x[i3]=x_in[i3]-(meshCenterX0[2]-numMeshPtsPerDir[2]*meshDeltaX/2);
	}

	N12=numMeshPtsPerDir[0]*numMeshPtsPerDir[1];
	Ntotal=N12*numMeshPtsPerDir[2];

	for(i=0; i<numControlPts; i++) {
		// find the grid cell position x lines in, since all the values are face centered,
		// f_u is shifted by half meshDeltaX on y, z direction
		// f_v is shifted by half meshDeltaX on x, z direction
		// f_w is shifted by half meshDeltaX on z, y direction
		i1=3*i;
		i2=i1+1;
		i3=i2+1;
		// f_u component
		p1=(int)floor(x[i1]/meshDeltaX);
		p2=(int)floor(x[i2]/meshDeltaX-0.5);
		p3=(int)floor(x[i3]/meshDeltaX-0.5);
		if(usePeskin) {
  		  for(k3=-2; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
			  	for(k1=-2; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=-(x[i1]/meshDeltaX-p[0]);
						r[1]=-(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=-(x[i3]/meshDeltaX-(p[2]+0.5));
						n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
						f_out[0][n]-=
							derivativePeskinDelta(r[0])*peskinDelta(r[1])*peskinDelta(r[2])/meshDeltaX4th;
					}
				}
			 }
		  }

  		  // f_v component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX);
		  p3=(int)floor(x[i3]/meshDeltaX-0.5);
		  for(k3=-2; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
				for(k1=-2; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=-(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=-(x[i2]/meshDeltaX-p[1]);
						r[2]=-(x[i3]/meshDeltaX-(p[2]+0.5));
						n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
						f_out[1][n]-=
							peskinDelta(r[0])*derivativePeskinDelta(r[1])*peskinDelta(r[2])/meshDeltaX4th;
					}
				}
			 }
		  }
		  // f_w component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX-0.5);
		  p3=(int)floor(x[i3]/meshDeltaX-1.0);
		  for(k3=-2; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
				for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]-1) {
						r[0]=-(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=-(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=-(x[i3]/meshDeltaX-(p[2]+1.0));
						n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
						f_out[2][n]-=
							peskinDelta(r[0])*peskinDelta(r[1])*derivativePeskinDelta(r[2])/meshDeltaX4th;
					}
				}
			 }
		  }
		} else {
  		  for(k3=-1; k3<3; k3++) {
			 for(k2=-1; k2<3; k2++) {
			  	for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=-(x[i1]/meshDeltaX-p[0]);
						r[1]=-(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=-(x[i3]/meshDeltaX-(p[2]+0.5));
						n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
						f_out[0][n]-=
							derivativeRomaDelta(r[0])*romaDelta(r[1])*romaDelta(r[2])/meshDeltaX4th;

					}
				}
			 }
		  }

	  	  // f_v component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX);
		  p3=(int)floor(x[i3]/meshDeltaX-0.5);
		  for(k3=-1; k3<3; k3++) {
			 for(k2=-1; k2<3; k2++) {
				for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]) {
						r[0]=-(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=-(x[i2]/meshDeltaX-p[1]);
						r[2]=-(x[i3]/meshDeltaX-(p[2]+0.5));
						n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
						f_out[1][n]-=
							romaDelta(r[0])*derivativeRomaDelta(r[1])*romaDelta(r[2])/meshDeltaX4th;
					}
				}
			 }
		  }
		  // f_w component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX-0.5);
		  p3=(int)floor(x[i3]/meshDeltaX-1.0);
		  for(k3=-1; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
				for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					if (p[2]>=0&&p[2]<numMeshPtsPerDir[2]-1) {
						r[0]=-(x[i1]/meshDeltaX-(p[0]+0.5));
						r[1]=-(x[i2]/meshDeltaX-(p[1]+0.5));
						r[2]=-(x[i3]/meshDeltaX-(p[2]+1.0));
						n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
						f_out[2][n]-=
								romaDelta(r[0])*romaDelta(r[1])*derivativeRomaDelta(r[2])/meshDeltaX4th;
					}
				}
			 }
		  }
	   }  // end of if(usePeskin)
	}

	delete[] x;
}

double SELM_CouplingOperator_StaggeredGrid1::peskinDelta(double r){
	if(r>=2||r<=-2) {
		return 0.0;
	} else if (r>-2 && r<=-1) {
		return (5.0+2.0*r-sqrt(-7.0-4.0*r*(3.0+r)))/8.0;
	} else if (r>-1 && r<=0) {
		return (3.0+2.0*r+sqrt(1.0-4.0*r*(1.0+r)))/8.0;
	} else if (r>0 &&  r<=1) {
		return (3.0-2.0*r+sqrt(1.0+4.0*r*(1.0-r)))/8.0;
	} else {
		return (5.0-2.0*r-sqrt(-7.0+4.0*r*(3.0-r)))/8.0;
	}
}

double SELM_CouplingOperator_StaggeredGrid1::derivativePeskinDelta(double r){
	if(r>=2||r<=-2) {
		return 0.0;
	} else if (r>-2 && r<=-1) {
		return (1.0+(3.0+2.0*r)/sqrt(-7.0-4.0*r*(3.0+r)))/4.0;
	} else if (r>-1 && r<=0) {
		return (1.0-(1.0+2.0*r)/sqrt(1.0-4.0*r*(1.0+r)))/4.0;
	} else if (r>0 &&  r<=1) {
		return (1.0-(-1.0+2.0*r)/sqrt(1.0+4.0*r*(1.0-r)))/4.0;
	} else {
		return (1.0+(-3.0+2.0*r)/sqrt(-7.0+4.0*r*(3.0-r)))/4.0;
	}
}

double SELM_CouplingOperator_StaggeredGrid1::romaDelta(double r){
	if(r>=1.5 || r<=-1.5) {
		return 0.0;
	} else if(r>-1.5 && r<=-0.5) {
		return (5.0+3.0*r-sqrt(-3.0*(1.0+r)*(1.0+r)+1.0))/6.0;
	} else if(r>-0.5 && r<0.5) {
		return (1.0+sqrt(-3.0*r*r+1.0))/3.0;
	} else {
		return (5.0-3.0*r-sqrt(-3.0*(1.0-r)*(1.0-r)+1.0))/6.0;
	}
}

double SELM_CouplingOperator_StaggeredGrid1::derivativeRomaDelta(double r){
	if(r>=1.5 || r<=-1.5) {
		return 0.0;
	} else if(r>-1.5 && r<=-0.5) {
		return (1.0+(1.0+r)/sqrt(-3.0*(1.0+r)*(1.0+r)+1.0))/2.0;
	} else if(r>-0.5 && r<0.5) {
		return -r/sqrt(-3.0*r*r+1.0);
	} else {
		return (-1.0+(1.0-r)/sqrt(-3.0*(1.0-r)*(1.0-r)+1.0))/2.0;
	}
}

}
