/*
 * SELM_Eulerian_Uniform1_Periodic.cpp
 *
 * http://atzberger.org/
 *
 */

#include "SELM_Eulerian.h"
#include "SELM_Eulerian_Types.h"
#include "SELM_Eulerian_Uniform1_Periodic.h"

#include <cstdlib>
#include <iostream>
#include <cstdio>
#include <sstream>
#include <cmath>
#include "string.h"

// LAMMPS
#include "domain.h"

using namespace std;

namespace LAMMPS_NS {

// ============ constants definition ===========
int SELM_Eulerian_Uniform1_Periodic::flag_use_setup         = 0;
double SELM_Eulerian_Uniform1_Periodic::UNIT_pi             = 3.141592653589793;
int SELM_Eulerian_Uniform1_Periodic::TYPE                   = SELM_Eulerian_Types::TYPE_Uniform1_Periodic;
char* SELM_Eulerian_Uniform1_Periodic::TYPE_STR             = SELM_Eulerian_Types::TYPE_STR_Uniform1_Periodic;
const char *SELM_Eulerian_Uniform1_Periodic::error_str_code = __FILE__;//"SELM_Eulerian_Uniform1_Periodic.cpp";

// output flags
const int    SELM_Eulerian_Uniform1_Periodic::OUTPUTFLAG_TYPE_NULL                = 0;
const char*  SELM_Eulerian_Uniform1_Periodic::OUTPUTFLAG_TYPESTR_NULL             = "NULL";

const int    SELM_Eulerian_Uniform1_Periodic::OUTPUTFLAG_TYPE_SELM                = 1;
const char*  SELM_Eulerian_Uniform1_Periodic::OUTPUTFLAG_TYPESTR_SELM             = "SELM";

const int    SELM_Eulerian_Uniform1_Periodic::OUTPUTFLAG_TYPE_VTK                 = 2;
const char*  SELM_Eulerian_Uniform1_Periodic::OUTPUTFLAG_TYPESTR_VTK              = "vtk";

const int    SELM_Eulerian_Uniform1_Periodic::OUTPUTFLAG_TYPE_VTK_LEGACY          = 3;
const char*  SELM_Eulerian_Uniform1_Periodic::OUTPUTFLAG_TYPESTR_VTK_LEGACY       = "VTK_legacy";

const int    SELM_Eulerian_Uniform1_Periodic::OUTPUTFLAG_TYPE_VELOCITY            = 4;
const char*  SELM_Eulerian_Uniform1_Periodic::OUTPUTFLAG_TYPESTR_VELOCITY         = "velocity";

const int    SELM_Eulerian_Uniform1_Periodic::OUTPUTFLAG_TYPE_FORCE_DENSITY       = 5;
const char*  SELM_Eulerian_Uniform1_Periodic::OUTPUTFLAG_TYPESTR_FORCE_DENSITY    = "force_density";

const int    SELM_Eulerian_Uniform1_Periodic::OUTPUTFLAG_TYPE_PRESSURE            = 6;
const char*  SELM_Eulerian_Uniform1_Periodic::OUTPUTFLAG_TYPESTR_PRESSURE         = "pressure";

const int    SELM_Eulerian_Uniform1_Periodic::OUTPUTFLAG_TOTAL_NUM                = 5;


SELM_Eulerian_Uniform1_Periodic::SELM_Eulerian_Uniform1_Periodic() {
  // generic initialization
  init();
}

SELM_Eulerian_Uniform1_Periodic::SELM_Eulerian_Uniform1_Periodic(int narg, char **arg) : SELM_Eulerian(narg, arg)
{

  const char *error_str_func = "SELM_Eulerian_Uniform1_Periodic()";

  // generic initialization
  init();
}


SELM_Eulerian_Uniform1_Periodic::~SELM_Eulerian_Uniform1_Periodic() {

  if(flag_use_setup) {
    deleteSetup();
  }

}

void SELM_Eulerian_Uniform1_Periodic::init() {

  /* identify the type of the mesh */
  type = SELM_Eulerian_Uniform1_Periodic::TYPE;
  strcpy(typeStr, SELM_Eulerian_Uniform1_Periodic::TYPE_STR);

  flagWriteSimulationData    = 0;
  saveSkipSimulationData     = 0;

  //flagWriteFluidVel_VTK      = 0;
  //flagWriteFluidForce_VTK    = 0;
  //flagWriteFluidPressure_VTK = 0;


  // setup output flags
  for (int i = 0; i < OUTPUTFLAG_TOTAL_NUM; i++) {
    outputFlagsStr[i][0] = 0;
    outputFlags[i]       = 0;
  }
  strcpy(outputFlagsStr[OUTPUTFLAG_TYPE_NULL],             OUTPUTFLAG_TYPESTR_NULL);
  strcpy(outputFlagsStr[OUTPUTFLAG_TYPE_SELM],             OUTPUTFLAG_TYPESTR_SELM);
  strcpy(outputFlagsStr[OUTPUTFLAG_TYPE_VTK],              OUTPUTFLAG_TYPESTR_VTK);
  strcpy(outputFlagsStr[OUTPUTFLAG_TYPE_VTK_LEGACY],       OUTPUTFLAG_TYPESTR_VTK_LEGACY);
  strcpy(outputFlagsStr[OUTPUTFLAG_TYPE_VELOCITY],         OUTPUTFLAG_TYPESTR_VELOCITY);
  strcpy(outputFlagsStr[OUTPUTFLAG_TYPE_FORCE_DENSITY],    OUTPUTFLAG_TYPESTR_FORCE_DENSITY);
  strcpy(outputFlagsStr[OUTPUTFLAG_TYPE_PRESSURE],         OUTPUTFLAG_TYPESTR_PRESSURE);

  resetSimulationOutputFlags(); // default values

}

void SELM_Eulerian_Uniform1_Periodic::setup() {

  const char *error_str_func = "setup()";

  flag_use_setup = 1;

  // determine the domain size and meshDeltaX based on
  // the number of grid points specified in each direction.
  double meshDeltaX1 = lammps->domain->xprd/numMeshPtsPerDir[0];
  double meshDeltaX2 = lammps->domain->yprd/numMeshPtsPerDir[1];
  double meshDeltaX3 = lammps->domain->zprd/numMeshPtsPerDir[2];
  double avgDx = (meshDeltaX1 + meshDeltaX2 + meshDeltaX3)/3.0;

  if ((fabs(meshDeltaX1 - avgDx) > 1e-4) || (fabs(meshDeltaX2 - avgDx) > 1e-4) || (fabs(meshDeltaX3 - avgDx) > 1e-4)) {
    stringstream message;
    message << "The number of points specified in each direction must yield" << endl;
    message << "mesh with the same aspect ratio as the LAMMPS domain." << endl;
    message << "meshDeltaX1 = " << meshDeltaX1 << endl;
    message << "meshDeltaX2 = " << meshDeltaX2 << endl;
    message << "meshDeltaX3 = " << meshDeltaX3 << endl;
    message << "numMeshPtsPerDir[0] = " << numMeshPtsPerDir[0] << endl;
    message << "numMeshPtsPerDir[1] = " << numMeshPtsPerDir[1] << endl;
    message << "numMeshPtsPerDir[2] = " << numMeshPtsPerDir[2] << endl;
    message << "LAMMPS L1           = " << lammps->domain->xprd << endl;
    message << "LAMMPS L2           = " << lammps->domain->yprd << endl;
    message << "LAMMPS L3           = " << lammps->domain->zprd << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  meshDeltaX = avgDx;

  meshL[0] = meshDeltaX*numMeshPtsPerDir[0];
  meshL[1] = meshDeltaX*numMeshPtsPerDir[1];
  meshL[2] = meshDeltaX*numMeshPtsPerDir[2];

  //allocate velocity, force density and pressure arrays
  //allocate velocity u, v, w, force  on face center
  //allocate pressure at cell center.
  int N = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];

  for (int d = 0; d < num_dim; d++) {

    u_main[d]                 = new double[N];
    u_stochastic[d]           = new double[N];

    u[d]                      = new double[N];

    fluidForceDensity[d]      = new double[N];
    fluidStochForceDensity[d] = new double[N];

    for(int i = 0; i < N; i++) {
      u_main[d][i]         = 0.0;
      u_stochastic[d][i]           = 0.0;
      u[d][i]                      = 0.0;
      fluidForceDensity[d][i]      = 0.0;
      fluidStochForceDensity[d][i] = 0.0;
	} /* end i loop */

  } /* end d loop */

  p_main = new double[N];
  p_stochastic   = new double[N];
  p              = new double[N];

  for(int i = 0; i < N; i++) {
	p_main[i] = 0.0;
	p_stochastic[i]   = 0.0;
	p[i]              = 0.0;
  } /* end i loop */

}

void SELM_Eulerian_Uniform1_Periodic::deleteSetup() {

  for (int d = 0; d < num_dim; d++) {
  	delete[] u_main[d];
  	delete[] u_stochastic[d];
  	delete[] u[d];
  	delete[] fluidForceDensity[d];
  }

  delete[] p_main;
  delete[] p_stochastic;
  delete[] p;
}

void SELM_Eulerian_Uniform1_Periodic::packageError(int code, void *extras) {
    exit(code);
}

void SELM_Eulerian_Uniform1_Periodic::writeSimulationDataToDisk(char *baseFilename, int timeIndex) {

	const char *error_str_func = "writeSimulationDataToDisk()";

	int    d, n, k1, k2, k3, N12, I;
	double meshLengths[3];

    char  filename[1000];

    // ============================= SELM file format data ========================
    if (outputFlags[OUTPUTFLAG_TYPE_SELM]) {
      // nothing to write
    }

    // ============================= VTK file format data (.vtr) ========================
    if (outputFlags[OUTPUTFLAG_TYPE_VTK]) {
      writeVTK(baseFilename, timeIndex);
    }

    // ============================= VTK LEGACY file format data ========================
    if (outputFlags[OUTPUTFLAG_TYPE_VTK_LEGACY]) {
    //if (flagWriteFluidVel_VTK) {
      writeVTKLegacy(baseFilename, timeIndex);
    } // end VTK Legacy


}


void SELM_Eulerian_Uniform1_Periodic::setSimulationOutputFlags(char *outputFlagStr) {

  char parseStr[1000];
  char tokenStr[100][100]; // assumes not more than 100 tokens!

  if (strcmp(outputFlagStr,"all") == 0) {

    // set all output flags to one
    for (int i = 0; i < OUTPUTFLAG_TOTAL_NUM; i++) {
      outputFlags[i] = 1;
    }

  } else { // loop over tokens and selectively set output
    // loop over the tokens of the string and set the flags
    strcpy(parseStr,outputFlagStr); // avoids destroying original string
    char *p        = strtok(outputFlagStr, " ");
    int  k         = 0;
    int  numTokens = 0;
    while (p) {
      //printf ("Token: %s\n", p);
      strcpy(tokenStr[k],p);
      p = strtok(NULL, " ");
      k++;
    }
    numTokens = k;

    // reset all flags to default values
    resetSimulationOutputFlags();

    for (k = 0; k < numTokens; k++) {
      for (int i = 0; i < OUTPUTFLAG_TOTAL_NUM; i++) {
        if (strcmp(tokenStr[k],outputFlagsStr[i]) == 0) {
          outputFlags[i] = 1;
        }
      }
    }

  } /* end else */

}

void SELM_Eulerian_Uniform1_Periodic::resetSimulationOutputFlags() {

  // set all to zero
  for (int i = 0; i < OUTPUTFLAG_TOTAL_NUM; i++) {
    outputFlags[i] = 0;
  }

  // set default values that are on
  outputFlags[OUTPUTFLAG_TYPE_VTK]           = 1;
  outputFlags[OUTPUTFLAG_TYPE_FORCE_DENSITY] = 1;
  outputFlags[OUTPUTFLAG_TYPE_VELOCITY]      = 1;
  outputFlags[OUTPUTFLAG_TYPE_PRESSURE]      = 0;

}

void SELM_Eulerian_Uniform1_Periodic::userAppl_writeVecFieldVTKFile(const char   *filename,
                                                                  int     num_dim,
                                                                  int    *numMeshPtsPerDir,
                                                                  double *meshLengths,
                                                                  const char   *vec_name,
                                                                  double **vec_array) {
  FILE  *fid;
  int    k;
  int    d;
  int    J[3];
  double X[3];
  double deltaX;
  int    index;

  int     numIndices;
  int    *indices;

  fid = fopen(filename,"w");

  fprintf(fid,"# vtk DataFile Version 1.0\n");
  fprintf(fid,"Vector Field %s. \n",vec_name);
  fprintf(fid,"ASCII\n");
  fprintf(fid,"\n");

  numIndices = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];
  indices=new int[numIndices];
  for(int k=0; k<numIndices; k++) {
	  indices[k]=k;
  }

  if (numIndices >= 0) {
      fprintf(fid,"DATASET UNSTRUCTURED_GRID\n");
      fprintf(fid,"POINTS %d float\n",numIndices);
      for (k = 0; k < numIndices; k++) {
    	  index = indices[k];
    	  J[2]    = index/(numMeshPtsPerDir[1]*numMeshPtsPerDir[0]);
    	  J[1]    = (index - J[2]*(numMeshPtsPerDir[1]*numMeshPtsPerDir[0]))
            		  / numMeshPtsPerDir[0];
    	  J[0]    = (index - J[2]*(numMeshPtsPerDir[1]*numMeshPtsPerDir[0])
    			  - J[1]*numMeshPtsPerDir[0]);

    	  for (d = 0; d < num_dim; d++) {
    		  deltaX = meshLengths[d]/numMeshPtsPerDir[d];
    		  X[d]   = (J[d]+0.5-numMeshPtsPerDir[d]/2.0)*deltaX+meshCenterX0[d];
    	  }

    	  // write the point location
    	  fprintf(fid,"%g %g %g", X[0], X[1], X[2]);
    	  fprintf(fid,"\n");
      }
      fprintf(fid,"\n");

      fprintf(fid,"CELLS %d %d\n",numIndices,2*numIndices);
      for (k = 0; k < numIndices; k++) {
    	  fprintf(fid,"1 %d \n",k);
      }
      fprintf(fid,"\n");

      fprintf(fid,"CELL_TYPES %d \n",numIndices);
      for (k = 0; k < numIndices; k++) {
    	  fprintf(fid,"1 \n");
      }
      fprintf(fid,"\n");

      fprintf(fid,"POINT_DATA %d \n",numIndices);
      fprintf(fid,"\n");

      fprintf(fid,"VECTORS %s float \n",vec_name);
      for (k = 0; k < numIndices; k++) {
    	  index = indices[k];
    	  for (d = 0; d < num_dim; d++) {
    		  fprintf(fid,"%g ", vec_array[d][index]);
    	  }
    	  fprintf(fid,"\n");
      }
      fprintf(fid,"\n");

  } // end of if numIndices>0
  fclose(fid);
  delete[] indices;
}

void SELM_Eulerian_Uniform1_Periodic::userAppl_writeScalarFieldVTKFile(const char   *filename,
                                                                  int     num_dim,
                                                                  int    *numMeshPtsPerDir,
                                                                  double *meshLengths,
                                                                  const char   *scalar_name,
                                                                  double *scalar_array) {
  FILE  *fid;
  int    k;
  int    d;
  int    J[3];
  double X[3];
  double deltaX;
  int    index;

  int     numIndices;
  int    *indices;

  fid = fopen(filename,"w");

  fprintf(fid,"# vtk DataFile Version 1.0\n");
  fprintf(fid,"Scalar Field %s. \n",scalar_name);
  fprintf(fid,"ASCII\n");
  fprintf(fid,"\n");

  numIndices = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];
  indices=new int[numIndices];
  for(int k=0; k<numIndices; k++) {
	  indices[k]=k;
  }

  if (numIndices >= 0) {
      fprintf(fid,"DATASET UNSTRUCTURED_GRID\n");
      fprintf(fid,"POINTS %d float\n",numIndices);
      for (k = 0; k < numIndices; k++) {
    	  index = indices[k];
    	  J[2]    = index/(numMeshPtsPerDir[1]*numMeshPtsPerDir[0]);
    	  J[1]    = (index - J[2]*(numMeshPtsPerDir[1]*numMeshPtsPerDir[0]))
            		  / numMeshPtsPerDir[0];
    	  J[0]    = (index - J[2]*(numMeshPtsPerDir[1]*numMeshPtsPerDir[0])
    			  - J[1]*numMeshPtsPerDir[0]);

    	  for (d = 0; d < num_dim; d++) {
    		  deltaX = meshLengths[d]/numMeshPtsPerDir[d];
    		  X[d]   = (J[d]+0.5-numMeshPtsPerDir[d]/2.0)*deltaX+meshCenterX0[d];
    	  }

    	  // write the point location
    	  fprintf(fid,"%g %g %g", X[0], X[1], X[2]);
    	  fprintf(fid,"\n");
      }
      fprintf(fid,"\n");

      fprintf(fid,"CELLS %d %d\n",numIndices,2*numIndices);
      for (k = 0; k < numIndices; k++) {
    	  fprintf(fid,"1 %d \n",k);
      }
      fprintf(fid,"\n");

      fprintf(fid,"CELL_TYPES %d \n",numIndices);
      for (k = 0; k < numIndices; k++) {
    	  fprintf(fid,"1 \n");
      }
      fprintf(fid,"\n");

      fprintf(fid,"POINT_DATA %d \n",numIndices);
      fprintf(fid,"\n");

      fprintf(fid,"Scalar_Vectors %s float \n",scalar_name);
      for (k = 0; k < numIndices; k++) {
    	  index = indices[k];
  		  fprintf(fid,"%g ", scalar_array[index]);
    	  fprintf(fid,"\n");
      }
      fprintf(fid,"\n");
  } // end of if numIndices>0
  fclose(fid);
  delete[] indices;
}

void SELM_Eulerian_Uniform1_Periodic::userAppl_writeVecFieldXFaceVTKFile(const char   *filename,
                                                                  int     num_dim,
                                                                  int    *numMeshPtsPerDir,
                                                                  double *meshLengths,
                                                                  const char   *vec_name,
                                                                  double **vec_array) {
  FILE  *fid;
  int    k;
  int    d;
  int    J[3];
  double X[3];
  double deltaX;
  int    index;

  int     numIndices;
  int    *indices;

  fid = fopen(filename,"w");

  fprintf(fid,"# vtk DataFile Version 1.0\n");
  fprintf(fid,"Vector Field %s. \n",vec_name);
  fprintf(fid,"ASCII\n");
  fprintf(fid,"\n");

  numIndices = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];
  indices=new int[numIndices];
  for(int k=0; k<numIndices; k++) {
	  indices[k]=k;
  }

  if (numIndices >= 0) {
      fprintf(fid,"DATASET UNSTRUCTURED_GRID\n");
      fprintf(fid,"POINTS %d float\n",numIndices);
      for (k = 0; k < numIndices; k++) {
    	  index = indices[k];
    	  J[2]    = index/(numMeshPtsPerDir[1]*numMeshPtsPerDir[0]);
    	  J[1]    = (index - J[2]*(numMeshPtsPerDir[1]*numMeshPtsPerDir[0]))
            		  / numMeshPtsPerDir[0];
    	  J[0]    = (index - J[2]*(numMeshPtsPerDir[1]*numMeshPtsPerDir[0])
    			  - J[1]*numMeshPtsPerDir[0]);

    	  for (d = 1; d < num_dim; d++) {
    		  deltaX = meshLengths[d]/numMeshPtsPerDir[d];
    		  X[d]   = (J[d]+0.5-numMeshPtsPerDir[d]/2.0)*deltaX+meshCenterX0[d];
    	  }
    	  X[0]=J[0]*deltaX;

    	  // write the point location
    	  fprintf(fid,"%g %g %g", X[0], X[1], X[2]);
    	  fprintf(fid,"\n");
      }
      fprintf(fid,"\n");

      fprintf(fid,"CELLS %d %d\n",numIndices,2*numIndices);
      for (k = 0; k < numIndices; k++) {
    	  fprintf(fid,"1 %d \n",k);
      }
      fprintf(fid,"\n");

      fprintf(fid,"CELL_TYPES %d \n",numIndices);
      for (k = 0; k < numIndices; k++) {
    	  fprintf(fid,"1 \n");
      }
      fprintf(fid,"\n");

      fprintf(fid,"POINT_DATA %d \n",numIndices);
      fprintf(fid,"\n");

      fprintf(fid,"VECTORS %s float \n",vec_name);
      for (k = 0; k < numIndices; k++) {
    	  index = indices[k];
    	  for (d = 0; d < num_dim; d++) {
    		  fprintf(fid,"%g ", vec_array[d][index]);
    	  }
    	  fprintf(fid,"\n");
      }
      fprintf(fid,"\n");

  } // end of if numIndices>0
  fclose(fid);
  delete[] indices;
}


void SELM_Eulerian_Uniform1_Periodic::writeVTKLegacy(char *baseFilename, int timeIndex) {

    int    d, n, k1, k2, k3, N12, I;
    double meshLengths[3];

    char  filename[1000];

    for (d = 0; d < num_dim; d++) {
      meshLengths[d] = numMeshPtsPerDir[d]*meshDeltaX;
    }

    int totalMeshPts;
    totalMeshPts=numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];
    N12=numMeshPtsPerDir[0]*numMeshPtsPerDir[1];

    if (outputFlags[OUTPUTFLAG_TYPE_VELOCITY]) {

      sprintf(filename,"%s_FluidVel_%.9d.vtk", nameStr, timeIndex);

      //interpolate the cell faces velocity to cell center velocity
      double* u_c[3];

      for (d = 0; d < num_dim; d++) {
          u_c[d]=new double[totalMeshPts];
      }
      //calculate the kinetic energy
      double totalKe=0;
      double rho=602.2142*meshDeltaX*meshDeltaX*meshDeltaX;
      for(k3=0; k3<numMeshPtsPerDir[2]; k3++) {
          for(k2=0; k2<numMeshPtsPerDir[1]; k2++) {
              for(k1=0; k1<numMeshPtsPerDir[0]; k1++) {
                  n=k3*N12+k2*numMeshPtsPerDir[0]+k1;
                  if(k1==numMeshPtsPerDir[0]-1) {
                      I=k3*N12+k2*numMeshPtsPerDir[0];
                  } else {
                      I=n+1;
                  }
                  u_c[0][n]=(u_main[0][I]+u_main[0][n])/2;

                  if(k2==numMeshPtsPerDir[1]-1) {
                      I=k3*N12+k1;
                  } else {
                      I=n+numMeshPtsPerDir[0];
                  }
                  u_c[1][n]=(u_main[1][I]+u_main[1][n])/2;

                  if(k3==numMeshPtsPerDir[2]-1) {
                      I=k2*numMeshPtsPerDir[0]+k1;
                  } else {
                      I=n+N12;
                  }
                  u_c[2][n]=(u_main[2][I]+u_main[2][n])/2;

                  totalKe=rho*(u_c[0][n]*u_c[0][n]+u_c[1][n]*u_c[1][n]+u_c[2][n]*u_c[2][n])/2.0;
              }
          }
      }
      FILE* fid;
      fid = fopen("LiquidKineticEnergy.dat","a");
      fprintf(fid,"%g\n", totalKe);
      fclose(fid);
      userAppl_writeVecFieldVTKFile( filename, num_dim, numMeshPtsPerDir, meshLengths,"fluid_velocity",u_c);

      for (d = 0; d < num_dim; d++) {
          delete[]  u_c[d];
      }
  } // end of flagWriteFluidVel_VTK

  if (outputFlags[OUTPUTFLAG_TYPE_FORCE_DENSITY]) {
  //if (flagWriteFluidForce_VTK) {
    sprintf(filename,"%s_FluidForce_%.9d.vtk", nameStr, timeIndex);
    //interpolate the cell faces forces to cell center velocity
    double* f_c[3];

    for (d = 0; d < num_dim; d++) {
        f_c[d]=new double[totalMeshPts];
    }

    for(k3=0; k3<numMeshPtsPerDir[2]; k3++) {
        for(k2=0; k2<numMeshPtsPerDir[1]; k2++) {
            for(k1=0; k1<numMeshPtsPerDir[0]; k1++) {
                n=k3*N12+k2*numMeshPtsPerDir[0]+k1;
                if(k1==numMeshPtsPerDir[0]-1) {
                    I=k3*N12+k2*numMeshPtsPerDir[0];
                } else {
                    I=n+1;
                }
                f_c[0][n]=(fluidForceDensity[0][I]+fluidForceDensity[0][n])/2;

                if(k2==numMeshPtsPerDir[1]-1) {
                    I=k3*N12+k1;
                } else {
                    I=n+numMeshPtsPerDir[0];
                }
                f_c[1][n]=(fluidForceDensity[1][I]+fluidForceDensity[1][n])/2;

                if(k3==numMeshPtsPerDir[2]-1) {
                    I=k2*numMeshPtsPerDir[0]+k1;
                } else {
                    I=n+N12;
                }
                f_c[1][n]=(fluidForceDensity[1][I]+fluidForceDensity[1][n])/2;
            }
        }
    }

    userAppl_writeVecFieldVTKFile((char *)filename,
                                       num_dim,
                                       numMeshPtsPerDir,
                                       meshLengths,
                                       "fluid_forceDensity",
                                       f_c);
  }// end of flagWriteFluidForce_VTK

  if (outputFlags[OUTPUTFLAG_TYPE_PRESSURE]) {
  //if (flagWriteFluidPressure_VTK) {
    sprintf(filename,"%s_Pressure_%.9d.vtk", nameStr, timeIndex);

    userAppl_writeScalarFieldVTKFile((char *)filename,
                                       num_dim,
                                       numMeshPtsPerDir,
                                       meshLengths,
                                       "fluid_Pressure",
                                       p_main);
  } // end pressure

}

void SELM_Eulerian_Uniform1_Periodic::writeVTK(char *baseFilename, int timeIndex) {

 const char *error_str_func = "writeVTK(baseFilename, timeIndex)";

 char filename[1000];

 sprintf(filename,"%s_%s_%.9d.vtr", baseFilename, nameStr, timeIndex);

 writeVTK(filename);

}

void SELM_Eulerian_Uniform1_Periodic::writeVTK(char *filename) {

 const char *error_str_func = "writeVTK(filename)";

 int a = 1;

 stringstream extrasStr;

 ofstream fid;
 fid.open(filename);

 if (!fid.is_open()) {  // if file is not open
   stringstream message;
   message << "Could not open file to write error occured." << endl;
   message << "  filename = " << filename << endl;
   SELM_Package::packageError(error_str_code, error_str_func, message);
 }

 Atz_XML_Package::writeXMLHeader(fid);

 // VTKFile
 extrasStr.str("");
 extrasStr << "type=\""       << "RectilinearGrid" << "\" ";
 extrasStr << "version=\""    << "0.1" << "\" ";
 extrasStr << "byte_order=\"" << "LittleEndian" << "\"";
 Atz_XML_Package::writeTagStart(fid,
                                "VTKFile",
                                extrasStr.str().c_str());

 // RectilinearGrid
 extrasStr.str("");
 extrasStr << "WholeExtent=\"";
 for (int d = 0; d < num_dim; d++) {
   extrasStr << 0 << " ";
   extrasStr << numMeshPtsPerDir[d] - 1;
   if (d != num_dim - 1) {
     extrasStr << " ";
   }
 }
 extrasStr << "\"";
 Atz_XML_Package::writeTagStart(fid,
                                "RectilinearGrid",
                                extrasStr.str().c_str());

 // Piece
 extrasStr.str("");
 extrasStr << "Extent=\"";
 for (int d = 0; d < num_dim; d++) {
   extrasStr << 0 << " ";
   extrasStr << numMeshPtsPerDir[d] - 1;
   if (d != num_dim - 1) {
     extrasStr << " ";
   }
 }
 extrasStr << "\"";
 Atz_XML_Package::writeTagStart(fid,
                                "Piece",
                                extrasStr.str().c_str());

 // PointData
 Atz_XML_Package::writeTagStart(fid,"PointData");

 // DataArrays

 // Velocity field
 if (outputFlags[this->OUTPUTFLAG_TYPE_VELOCITY]) {
   // DataArray
   extrasStr.str("");
   extrasStr << "type=\"" << "Float32" << "\" ";
   extrasStr << "Name=\"" << "Velocity" << "\" ";
   extrasStr << "NumberOfComponents=\"" << num_dim << "\" ";
   extrasStr << "format=\"" << "ascii" << "\"";
   Atz_XML_Package::writeTagStart(fid,"DataArray",
                                  extrasStr.str().c_str());

   int numMeshPts = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];
   for (int I = 0; I < numMeshPts; I++) {
     for (int d = 0; d < num_dim; d++) {
       fid << u_main[d][I] << " ";
     }
   }
   fid << endl;

   Atz_XML_Package::writeTagEnd(fid,"DataArray");
 }

 // Force density field
 if (outputFlags[this->OUTPUTFLAG_TYPE_FORCE_DENSITY]) {
    // DataArray
    extrasStr.str("");
    extrasStr << "type=\"" << "Float32" << "\" ";
    extrasStr << "Name=\"" << "Force_Density" << "\" ";
    extrasStr << "NumberOfComponents=\"" << num_dim << "\" ";
    extrasStr << "format=\"" << "ascii" << "\"";
    Atz_XML_Package::writeTagStart(fid,"DataArray",
                                   extrasStr.str().c_str());

    int numMeshPts = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];
    for (int I = 0; I < numMeshPts; I++) {
      for (int d = 0; d < num_dim; d++) {
        fid << fluidForceDensity[d][I] << " ";
      }
    }
    fid << endl;

    Atz_XML_Package::writeTagEnd(fid,"DataArray");
 }

 // Pressure field
 if (outputFlags[OUTPUTFLAG_TYPE_PRESSURE]) {
   // DataArray
   extrasStr.str("");
   extrasStr << "type=\"" << "Float32" << "\" ";
   extrasStr << "Name=\"" << "Pressure" << "\" ";
   extrasStr << "NumberOfComponents=\"" << 1 << "\" ";
   extrasStr << "format=\"" << "ascii" << "\"";
   Atz_XML_Package::writeTagStart(fid,"DataArray",
                                  extrasStr.str().c_str());

   int numMeshPts = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];
   for (int I = 0; I < numMeshPts; I++) {
     fid << this->p_main[I] << " ";
   }
   fid << endl;

   Atz_XML_Package::writeTagEnd(fid,"DataArray");
 }

 Atz_XML_Package::writeTagEnd(fid,"PointData");

 // CellData
 Atz_XML_Package::writeTagStart(fid,"CellData");
 Atz_XML_Package::writeTagEnd(fid,"CellData");


 // Coordinates
 Atz_XML_Package::writeTagStart(fid,"Coordinates");

 // DataArrays
 char dimStr[3][10];
 strcpy(dimStr[0],"x");
 strcpy(dimStr[1],"y");
 strcpy(dimStr[2],"z");
 for (int d = 0; d < num_dim; d++) {

   // DataArray
   extrasStr.str("");
   extrasStr << "type=\"" << "Float32" << "\" ";
   extrasStr << "Name=\"" << dimStr[d] << "\" ";
   extrasStr << "NumberOfComponents=\"" << 1 << "\" ";
   extrasStr << "format=\"" << "ascii" << "\"";
   Atz_XML_Package::writeTagStart(fid,"DataArray",
                                  extrasStr.str().c_str());

   double X_base = meshCenterX0[d] - (numMeshPtsPerDir[d]*meshDeltaX/2.0) + 0.5*meshDeltaX;
   for (int I = 0; I < numMeshPtsPerDir[d]; I++) {
     fid << X_base + I*meshDeltaX << " ";
   }
   fid << endl;

   Atz_XML_Package::writeTagEnd(fid,"DataArray");

 } // end d loop

 Atz_XML_Package::writeTagEnd(fid,"Coordinates");

 Atz_XML_Package::writeTagEnd(fid,"Piece");

 Atz_XML_Package::writeTagEnd(fid,"RectilinearGrid");

 Atz_XML_Package::writeTagEnd(fid,"VTKFile");

 fid.close();

}




void SELM_Eulerian_Uniform1_Periodic::writeVTK(char *filename, const char *vecFieldName, double **vecField) {

  SELM_Eulerian_Uniform1_Periodic::writeVTK_vecField(filename,
                                                        vecFieldName,
                                                        vecField,
                                                        num_dim,
                                                        meshDeltaX,
                                                        numMeshPtsPerDir[0],
                                                        numMeshPtsPerDir[1],
                                                        numMeshPtsPerDir[2],
                                                        meshCenterX0[0],
                                                        meshCenterX0[1],
                                                        meshCenterX0[2]);
}

int SELM_Eulerian_Uniform1_Periodic::writeVTK_vecField(char        *filename,
                                                          const char  *vecFieldName,
                                                          double     **vecField,
                                                          int          num_dim,
                                                          int          numMeshPtsPerDir_0,
                                                          int          numMeshPtsPerDir_1,
                                                          int          numMeshPtsPerDir_2) {

  // set some default values to save time for debugging
  return SELM_Eulerian_Uniform1_Periodic::writeVTK_vecField(filename,
                                                        vecFieldName,
                                                        vecField,
                                                        num_dim,
                                                        1.0,
                                                        numMeshPtsPerDir_0,
                                                        numMeshPtsPerDir_1,
                                                        numMeshPtsPerDir_2,
                                                        0.0,
                                                        0.0,
                                                        0.0);

}

int SELM_Eulerian_Uniform1_Periodic::writeVTK_vecField(char        *filename,
                                                          const char  *vecFieldName,
                                                          double     **vecField,
                                                          int          num_dim,
                                                          double       meshDeltaX,
                                                          int          numMeshPtsPerDir_0,
                                                          int          numMeshPtsPerDir_1,
                                                          int          numMeshPtsPerDir_2,
                                                          double       meshCenterX0_0,
                                                          double       meshCenterX0_1,
                                                          double       meshCenterX0_2) {

  const char *error_str_func = "writeVTK_vecField(filename, vecFieldName, vecField)";

  int    numMeshPtsPerDir[3];
  double meshCenterX0[3];

  stringstream extrasStr;

  numMeshPtsPerDir[0] = numMeshPtsPerDir_0;
  numMeshPtsPerDir[1] = numMeshPtsPerDir_1;
  numMeshPtsPerDir[2] = numMeshPtsPerDir_2;

  meshCenterX0[0]     = meshCenterX0_0;
  meshCenterX0[1]     = meshCenterX0_1;
  meshCenterX0[2]     = meshCenterX0_2;

  ofstream fid;
  fid.open(filename);

  if (!fid.is_open()) {  // if file is not open
   stringstream message;
   message << "Could not open file to write error occured." << endl;
   message << "  filename = " << filename << endl;
   SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  Atz_XML_Package::writeXMLHeader(fid);

  // VTKFile
  extrasStr.str("");
  extrasStr << "type=\""       << "RectilinearGrid" << "\" ";
  extrasStr << "version=\""    << "0.1" << "\" ";
  extrasStr << "byte_order=\"" << "LittleEndian" << "\"";
  Atz_XML_Package::writeTagStart(fid,
                                "VTKFile",
                                extrasStr.str().c_str());

  // RectilinearGrid
  extrasStr.str("");
  extrasStr << "WholeExtent=\"";
  for (int d = 0; d < num_dim; d++) {
   extrasStr << 0 << " ";
   extrasStr << numMeshPtsPerDir[d] - 1;
   if (d != num_dim - 1) {
     extrasStr << " ";
   }
  }
  extrasStr << "\"";
  Atz_XML_Package::writeTagStart(fid,
                                 "RectilinearGrid",
                                 extrasStr.str().c_str());

  // Piece
  extrasStr.str("");
  extrasStr << "Extent=\"";
  for (int d = 0; d < num_dim; d++) {
   extrasStr << 0 << " ";
   extrasStr << numMeshPtsPerDir[d] - 1;
   if (d != num_dim - 1) {
     extrasStr << " ";
   }
  }
  extrasStr << "\"";
  Atz_XML_Package::writeTagStart(fid,
                                 "Piece",
                                 extrasStr.str().c_str());

  // PointData
  Atz_XML_Package::writeTagStart(fid,"PointData");

  // DataArrays

  // Grid field values
  // DataArray
  extrasStr.str("");
  extrasStr << "type=\"" << "Float32" << "\" ";
  extrasStr << "Name=\"" << vecFieldName << "\" ";
  extrasStr << "NumberOfComponents=\"" << num_dim << "\" ";
  extrasStr << "format=\"" << "ascii" << "\"";
  Atz_XML_Package::writeTagStart(fid,"DataArray",
                                 extrasStr.str().c_str());

  int numMeshPts = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];
  for (int I = 0; I < numMeshPts; I++) {
   for (int d = 0; d < num_dim; d++) {
     fid << vecField[d][I] << " ";
   }
  }
  fid << endl;

  Atz_XML_Package::writeTagEnd(fid,"DataArray");

  Atz_XML_Package::writeTagEnd(fid,"PointData");

  // CellData
  Atz_XML_Package::writeTagStart(fid,"CellData");
  Atz_XML_Package::writeTagEnd(fid,"CellData");


  // Coordinates
  Atz_XML_Package::writeTagStart(fid,"Coordinates");

  // DataArrays
  char dimStr[3][10];
  strcpy(dimStr[0],"x");
  strcpy(dimStr[1],"y");
  strcpy(dimStr[2],"z");
  for (int d = 0; d < num_dim; d++) {

   // DataArray
   extrasStr.str("");
   extrasStr << "type=\"" << "Float32" << "\" ";
   extrasStr << "Name=\"" << dimStr[d] << "\" ";
   extrasStr << "NumberOfComponents=\"" << 1 << "\" ";
   extrasStr << "format=\"" << "ascii" << "\"";
   Atz_XML_Package::writeTagStart(fid,"DataArray",
                                  extrasStr.str().c_str());

   double X_base = meshCenterX0[d] - (numMeshPtsPerDir[d]*meshDeltaX/2.0) + 0.5*meshDeltaX;
   for (int I = 0; I < numMeshPtsPerDir[d]; I++) {
     fid << X_base + I*meshDeltaX << " ";
   }
   fid << endl;

   Atz_XML_Package::writeTagEnd(fid,"DataArray");

  } // end d loop

  Atz_XML_Package::writeTagEnd(fid,"Coordinates");

  Atz_XML_Package::writeTagEnd(fid,"Piece");

  Atz_XML_Package::writeTagEnd(fid,"RectilinearGrid");

  Atz_XML_Package::writeTagEnd(fid,"VTKFile");

  fid.close();

  return 0;

}


} // end name-space



