/* ----------------------------------------------------------------------

  Stochastic Eulerian Lagrangian Methods

  Paul J. Atzberger
  http://atzberger.org/
 

------------------------------------------------------------------------- */

#ifndef FFT_FFTW /* determine if FFTW package specified */
  #warning "No FFTW package specified for SELM codes.  The SELM functionality will be disabled."
#else
  #define USE_PACKAGE_FFTW3
#endif

/* standard C/C++ includes */
#include <cstdlib>
#include <math.h>
#include "stdio.h"
#include "string.h"

/* SELM includes */
#include "fix_SELM.h"
#include "SELM_Parser1.h"
#include "SELM_Integrator.h"
#include "SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3.h"
#include "SELM_Eulerian.h"
#include "SELM_Eulerian_Types.h"
#include "SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3.h"
#include "SELM_Lagrangian.h"
#include "SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE.h"
#include "SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE.h"
#include "SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID.h"
#include "SELM_CouplingOperator_Types.h"
#include "SELM_CouplingOperator_LAMMPS_SHEAR_UNIFORM1_FFTW3_TABLE1.h"

/* LAMMPS includes */
//#include "respa.h"
#include "atom.h"
#include "update.h"
#include "comm.h"
#include "irregular.h"
#include "domain.h"
#include "lattice.h"
#include "force.h"
#include "modify.h"
#include "math_const.h"
#include "kspace.h"
#include "input.h"
#include "variable.h"
#include "error.h"

#ifdef USE_PACKAGE_FFTW3
  #include "fftw3.h"
#endif

using namespace LAMMPS_NS;
using namespace FixConst;

/* ============ constants definition =========== */
double SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::UNIT_pi  = 3.141592653589793;

int    SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::TYPE     = SELM_Integrator_Types::TYPE_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3;
char*  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::TYPE_STR = SELM_Integrator_Types::TYPE_STR_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3;

char*  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::SHEAR_MODE_TYPE_STR_NULL      = "NULL";
char*  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::SHEAR_MODE_TYPE_STR_RM_SHEAR1 = "RM_SHEAR1";
char*  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::SHEAR_MODE_TYPE_STR_RM_OSC1   = "RM_OSC1";

char* SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::error_str_code = "SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3.cpp";

enum{NONE=0,FINAL,DELTA,SCALE,VEL,ERATE,TRATE,VOLUME,WIGGLE,VARIABLE};

/* ---------------------------------------------------------------------- */
SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3() : SELM_Integrator()
{

  //printf("WARNING: Creating SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3() (will need to setup from data at later stage). \n");
  //exit(1);

  init();

  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras
      = (SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_ExtrasType *)
        malloc(sizeof(SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_ExtrasType));

}

SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3(int narg, char **arg) : SELM_Integrator(narg, arg)
{

  const char *error_str_func = "SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3()";

  /* generic initialization */
  init();
  
  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras
      = (SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_ExtrasType *)
        malloc(sizeof(SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_ExtrasType));

  printf("Creating SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3(int narg, char **arg) : for this type of constructor not yet implemented. \n");
  //SELM_Package::packageError();
  std::exit(1);

}

SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3(LAMMPS *lmps, FixSELM *fx)
{

  /* generic initialization */
  init();
  
  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras
      = (SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_ExtrasType *)
        malloc(sizeof(SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_ExtrasType));  

  setGlobalRefs(lmps, fx);

}


SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::~SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3()
{

  // -- from fix_deform.cpp
  if(set) {
    for (int i = 0; i < 6; i++) {
      delete [] set[i].hstr;
      delete [] set[i].hratestr;
    }
  }
  delete [] set;

  delete irregular;
 
  // reset domain's h_rate = 0.0, since this fix may have made it non-zero
  double *h_rate = lammps->domain->h_rate;
  double *h_ratelo = lammps->domain->h_ratelo;

  h_rate[0] = h_rate[1] = h_rate[2] =
  h_rate[3] = h_rate[4] = h_rate[5] = 0.0;
  h_ratelo[0] = h_ratelo[1] = h_ratelo[2] = 0.0;
  // --
  
}

void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::setup_internals()
{

  const char *error_str_func = "setup_internals()";
  
  Atom *atom = lammps->atom;
  Domain *domain = lammps->domain;
  Update *update = lammps->update;

  int flagShearMode;
  
  int index;

  int    shearDir;
  int    shearVelDir;
  double shearRate;
  double shearDist;
  double shearDist_last;

  double shearOmega;
  double shearRateAmplitude;
  
  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3 *SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3;
  //SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE;

  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3::SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_ExtrasType *SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras;

  ShearData_RM_SHEAR1_Type *shearData_RM_SHEAR1;
  ShearData_RM_OSC1_Type   *shearData_RM_OSC1;

  /* generic initialization */
  //init();
  //setGlobalRefs(lmps, fx);
 
  flagShearMode = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagShearMode;

  /* de-reference variables */
  switch (flagShearMode) {

    case SHEAR_MODE_TYPE_RM_SHEAR1:

      shearData_RM_SHEAR1 = (ShearData_RM_SHEAR1_Type *)
          SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->shearData;

      /* retrieve shear values required to generate the flow */
      shearDir       = shearData_RM_SHEAR1->shearDir;
      shearVelDir    = shearData_RM_SHEAR1->shearVelDir;
      shearRate      = shearData_RM_SHEAR1->shearRate;
      shearDist      = shearData_RM_SHEAR1->shearDist;
      shearDist_last = shearData_RM_SHEAR1->shearDist_last;
      
      SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3
        = (SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3 *)fixSELM->SELM_Eulerian_List[0];

      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras
        = SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3->SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras;

      /* setup the mesh in accordance of the current shear conditions */
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDir
        = shearDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearVelDir
        = shearVelDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearRate
        = shearRate;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist
        = shearDist;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist_last
        = shearDist_last;

      break;

    case SHEAR_MODE_TYPE_RM_OSC1:

      shearData_RM_OSC1
        = (ShearData_RM_OSC1_Type *)
        SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->shearData;
       
      /* update the shear rate (oscillation) */
      shearDir           = shearData_RM_OSC1->shearDir;
      shearVelDir        = shearData_RM_OSC1->shearVelDir;
      shearOmega         = shearData_RM_OSC1->shearOmega;
      shearRateAmplitude = shearData_RM_OSC1->shearRateAmplitude;
      shearRate          = shearData_RM_OSC1->shearRate;
      shearDist          = shearData_RM_OSC1->shearDist;
      shearDist_last     = shearData_RM_OSC1->shearDist_last;
      
      SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3
        = (SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3 *)fixSELM->SELM_Eulerian_List[0];

      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras
        = SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3->SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras;      

      /* setup the mesh in accordance of the current shear conditions */
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDir
        = shearDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearVelDir
        = shearVelDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearRate
        = shearRate;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist
        = shearDist;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist_last
        = shearDist_last;

      break;

    } /* end of switch */
    
  //printf("flagShearMode = %d",flagShearMode);
  
  /* create "set" data structure from "fix_deform.cpp" and initialize for bookkeeping. */  
  // create Set for storing box book-keeping information
  set = new Set[6];
  memset(set,0,6*sizeof(Set));  // PJA: sets all values to zero
  for (int i = 0; i < 6; i++) { // to be extra safe we set styles explicitly to NONE
    set[i].style = NONE;
  }
  
  /* -- start of "fix deform" create() related items */
  /* setup domain settings so signals other LAMMPS classes of modifications
     (we follow approaches used in "fix deform" class).
   */
  fixSELM->no_change_box = 1;
  fixSELM->restart_global = 1;
  fixSELM->pre_exchange_migrate = 1;

  // // read options from end of input line
  // // no x remap effectively moves atoms within box, so set restart_pbc
  // options(narg-iarg,&arg[iarg]);
  // from options()
  flipflag = 1; // flips allowed once box gets too big
  remapflag = Domain::V_REMAP; // remap the velocity of particles when bonds cross the boundary

  if (remapflag != Domain::X_REMAP) fixSELM->restart_pbc = 1;
  fixSELM->restart_pbc = 1;

  // setup dimflags used by other classes to check for volume-change conflicts
  for (int i = 0; i < 6; i++)
    if (set[i].style == NONE) dimflag[i] = 0;
    else dimflag[i] = 1;

  //if (dimflag[0]) box_change |= BOX_CHANGE_X;
  //if (dimflag[1]) box_change |= BOX_CHANGE_Y;
  //if (dimflag[2]) box_change |= BOX_CHANGE_Z;
  //if (dimflag[3]) box_change |= BOX_CHANGE_YZ;
  if (dimflag[4]) fixSELM->box_change |= fixSELM->BOX_CHANGE_XZ;
  //if (dimflag[5]) box_change |= BOX_CHANGE_XY;

  fixSELM->box_change |= fixSELM->BOX_CHANGE_XZ;
  
  // // reneighboring only forced if flips can occur due to shape changes
  //if (flipflag && (set[3].style || set[4].style || set[5].style))
  // force_reneighbor = 1;  
  //next_reneighbor = -1;
  
  fixSELM->force_reneighbor = 1;
  fixSELM->next_reneighbor = -1;
  
  flip = 0;
  
  //if (force_reneighbor) irregular = new Irregular(lmp);
  //else irregular = nullptr;  
  irregular = new Irregular(lammps);
          
  /* further initialization will occur from init_from_fix() call */
  
}

void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::init_from_fix()
{

  const char *error_str_func = "init_from_fix()";
  
  Atom *atom = lammps->atom;
  Domain *domain = lammps->domain;
  Update *update = lammps->update;

  int flagShearMode;
  
  int index;

  int    shearDir;
  int    shearVelDir;
  double shearRate;
  double shearDist;
  double shearDist_last;

  double shearOmega;
  double shearRateAmplitude;
  
  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3 *SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3;
  //SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE;

  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3::SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_ExtrasType *SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras;

  ShearData_RM_SHEAR1_Type *shearData_RM_SHEAR1;
  ShearData_RM_OSC1_Type   *shearData_RM_OSC1;

  /* generic initialization */
  flagShearMode = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagShearMode;

  /* de-reference variables */
  /* PJA: Note: below is not currently really needed. */
  switch (flagShearMode) {

    case SHEAR_MODE_TYPE_RM_SHEAR1:
    
      SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3
        = (SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3 *)fixSELM->SELM_Eulerian_List[0];

      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras
        = SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3->SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras;    

      shearData_RM_SHEAR1 = (ShearData_RM_SHEAR1_Type *)
          SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->shearData;

      /* retrieve shear values required to generate the flow */
      shearDir       = shearData_RM_SHEAR1->shearDir;
      shearVelDir    = shearData_RM_SHEAR1->shearVelDir;
      shearRate      = shearData_RM_SHEAR1->shearRate;
      shearDist      = shearData_RM_SHEAR1->shearDist;
      shearDist_last = shearData_RM_SHEAR1->shearDist_last;
      
      /* setup the mesh in accordance of the current shear conditions */
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDir
        = shearDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearVelDir
        = shearVelDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearRate
        = shearRate;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist
        = shearDist;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist_last
        = shearDist_last;

      break;

    case SHEAR_MODE_TYPE_RM_OSC1:
    
      SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3
        = (SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3 *)fixSELM->SELM_Eulerian_List[0];

      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras
        = SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3->SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras;        

      shearData_RM_OSC1
        = (ShearData_RM_OSC1_Type *)
        SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->shearData;

      /* update the shear rate (oscillation) */
      shearDir           = shearData_RM_OSC1->shearDir;
      shearVelDir        = shearData_RM_OSC1->shearVelDir;      
      shearOmega         = shearData_RM_OSC1->shearOmega;
      shearRateAmplitude = shearData_RM_OSC1->shearRateAmplitude;
      shearRate          = shearData_RM_OSC1->shearRate;
      shearDist          = shearData_RM_OSC1->shearDist;
      shearDist_last     = shearData_RM_OSC1->shearDist_last;
      
      SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3
        = (SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3 *)fixSELM->SELM_Eulerian_List[0];

      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras
        = SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3->SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras;

      /* setup the mesh in accordance of the current shear conditions */
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDir
        = shearDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearVelDir
        = shearVelDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearRate
        = shearRate;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist
        = shearDist;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist_last
        = shearDist_last;

      break;

    } /* end of switch */
    
  // setup values for set and related init  
  triclinic = domain->triclinic;  // required to be one or issue error
  if (triclinic == 0) {
    stringstream message; message << "requires a 'triclinic' domain." << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }
   
  index = 4; // xz shear
  set[index].style = VEL;
  //set[index].vel = utils::numeric(FLERR,arg[iarg+2],false,lmp);
  //set[index].vel = shearRate*(domain->boxhi[2] - domain->boxlo[2]); // atz_set the shear rate here for now (might need conversion later) 
  set[index].vel = shearRate*(domain->boxhi[2] - domain->boxlo[2]); // atz_set the shear rate here for now (might need conversion later) 
    
  //---- init()
  
  // Kspace setting  
  if (lammps->force->kspace) kspace_flag = 1;
  else kspace_flag = 0;
  
  double delt = (update->endstep - update->beginstep)*update->dt;

  // set start/stop values for box size and shape
  // if single run, start is current values
  // if multiple runs enabled via run start/stop settings,
  //   start is value when fix deform was issued
  // if VARIABLE or NONE, no need to set
  for (int i = 0; i < 3; i++) {
  
    if (update->firststep == update->beginstep) {
      set[i].lo_start = domain->boxlo[i];
      set[i].hi_start = domain->boxhi[i];
      set[i].vol_start = domain->xprd * domain->yprd * domain->zprd;
    } else {
      set[i].lo_start = set[i].lo_initial;
      set[i].hi_start = set[i].hi_initial;
      set[i].vol_start = set[i].vol_initial;
    }  
    
    if (set[i].style == VEL) {
      set[i].lo_stop = set[i].lo_start - 0.5*delt*set[i].vel;
      set[i].hi_stop = set[i].hi_start + 0.5*delt*set[i].vel;
    }

  }

  for (int i = 3; i < 6; i++) {
       
    if (update->firststep == update->beginstep) {
      if (i == 5) set[i].tilt_start = domain->xy;
      else if (i == 4) set[i].tilt_start = domain->xz;
      else if (i == 3) set[i].tilt_start = domain->yz;
    } else set[i].tilt_start = set[i].tilt_initial;
    
    if (set[i].style == VEL) {
      set[i].tilt_stop = set[i].tilt_start + delt*set[i].vel;
    }
    
  }

  /* init procedure */  
  // set domain->h_rate values for use by domain and other fixes/computes
  // initialize all rates to 0.0
  // cannot set here for TRATE,VOLUME,WIGGLE,VARIABLE since not constant
  h_rate = domain->h_rate;
  h_ratelo = domain->h_ratelo;
  
  for (int i = 0; i < 3; i++) {
    h_rate[i] = h_ratelo[i] = 0.0;
  }
    
  /* for xz we have */ 
  for (int i = 3; i < 6; i++) {
    h_rate[i] = 0.0;
    if (set[i].style == FINAL || set[i].style == DELTA ||
        set[i].style == VEL || set[i].style == ERATE) {
      if (delt != 0.0)
        h_rate[i] = (set[i].tilt_stop - set[i].tilt_start) / delt;
      else h_rate[i] = 0.0;
    }
  } 

  flip = 0; // init flag to zero for the initialization 

  /* -- end of fix deform init() related items */
  
  /* -- also added this code, since domain explicitly looks for fix_deform in domain.cpp */
  /*
  deform_flag = deform_vremap = deform_groupbit = 0;
  for (int i = 0; i < modify->nfix; i++)
    if (strcmp(modify->fix[i]->style,"deform") == 0) {
      deform_flag = 1;
      if (((FixDeform *) modify->fix[i])->remapflag == Domain::V_REMAP) {
        deform_vremap = 1;
        deform_groupbit = modify->fix[i]->groupbit;
      }
    }
  */

  domain->deform_flag = 1;
  domain->deform_vremap = 1;
  domain->deform_groupbit = fixSELM->groupbit;
  
  // region inits
  // for (int i = 0; i < nregion; i++) regions[i]->init();
  
  /*
  printf("Creating SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3 : not yet implemented. \n");
  exit(1);
  */

}

void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::pre_exchange()
{

  Atom *atom = lammps->atom;
  Domain *domain = lammps->domain;

  /* codes from "fix_deform.cpp" to help handle atom flipping */
  if (flip == 0) return;

  domain->yz = set[3].tilt_target = set[3].tilt_flip;
  domain->xz = set[4].tilt_target = set[4].tilt_flip;
  domain->xy = set[5].tilt_target = set[5].tilt_flip;
  
  // atz -- (added to move the bottom of the box to make shear symmetric, with the flip)
  double lx = domain->boxhi[0] - domain->boxlo[0];
  domain->boxlo[0] = set[0].lo_target = set[0].lo_start - 0.5*domain->xz;
  domain->boxhi[0] = set[0].hi_target = set[0].lo_target + lx;
  //set[0].lo_target = set[0].lo_start + delta*(set[0].lo_stop - set[0].lo_start);
  //set[0].hi_target = set[0].lo_target + lx;
  // --
  
  domain->set_global_box();
  domain->set_local_box();

  domain->image_flip(flipxy,flipxz,flipyz);

  double **x = atom->x;
  imageint *image = atom->image;
  int nlocal = atom->nlocal;
  for (int i = 0; i < nlocal; i++) domain->remap(x[i],image[i]);

  domain->x2lamda(atom->nlocal);
  irregular->migrate_atoms();
  domain->lamda2x(atom->nlocal);

  flip = 0;

  // @@@ Possibly update the SELM codes with the flip information, if needed.
  
}


// handled above
//int SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::setmask()
//{
//  /* based on codes from "fix_deform.cpp" to help handle atom flipping */
//  int mask = 0;
//  if (force_reneighbor) mask |= PRE_EXCHANGE;
//  mask |= END_OF_STEP;
//  return mask;
//}

void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::end_of_step() {
  /* based on codes from "fix_deform.cpp", simplified for the shear xz case. */
  
  const char *error_str_func = "end_of_step()";
  
  Atom *atom = lammps->atom;
  Domain *domain = lammps->domain;  
  Update *update = lammps->update;
    
  ShearData_RM_SHEAR1_Type *shearData_RM_SHEAR1;
  ShearData_RM_OSC1_Type   *shearData_RM_OSC1;
  
  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3                                                       *SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3;
  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3::SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_ExtrasType *SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras;
      
  int flagShearMode = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagShearMode;
  
  /* de-reference variables */
  SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3
    = (SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3 *)fixSELM->SELM_Eulerian_List[0]; // assumes type

  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras 
    = SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3->SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras;
  
  if (flagShearMode == SHEAR_MODE_TYPE_RM_SHEAR1) {
      shearData_RM_SHEAR1 = (ShearData_RM_SHEAR1_Type *)
          SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->shearData;
  } else if (flagShearMode == SHEAR_MODE_TYPE_RM_OSC1) {
      shearData_RM_OSC1
        = (ShearData_RM_OSC1_Type *)
        SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->shearData;
  } else {
    stringstream message; message << "Unknown case of flagShearMode = " << flagShearMode << endl; 
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  // -- related to fix_deform.cpp --
  int i;  
  double delta = update->ntimestep - update->beginstep;
  
  if (delta != 0.0) delta /= update->endstep - update->beginstep;
  
  for (i = 0; i < 3; i++) {
    if (set[i].style == NONE) {
      set[i].lo_target = domain->boxlo[i];
      set[i].hi_target = domain->boxhi[i];
    }
  }
  
  // set new box size
  // for triclinic, set new box shape
  double *h = domain->h;
  h_rate = domain->h_rate;
  h_ratelo = domain->h_ratelo;
  
  if (triclinic) {
           
    for (i = 3; i < 6; i++) {

      if (set[i].style == NONE) {
        if (i == 5) set[i].tilt_target = domain->xy;
        else if (i == 4) set[i].tilt_target = domain->xz;
        else if (i == 3) set[i].tilt_target = domain->yz;
      } else if ((set[i].style == VEL) && (flagShearMode == SHEAR_MODE_TYPE_RM_SHEAR1)) { // we cut other cases, this is what gets called for VEL
        set[i].tilt_target = set[i].tilt_start + delta*(set[i].tilt_stop - set[i].tilt_start);  
        // shear rate assumed constant, so nothing to update for SELM codes currently.       
      } else if ((set[i].style == VEL) && (flagShearMode == SHEAR_MODE_TYPE_RM_OSC1)) {

        double dt = update->dt; double t = update->ntimestep*dt;
        double omega = shearData_RM_OSC1->shearOmega;
        double A = shearData_RM_OSC1->shearRateAmplitude;
        double shearRate;

        set[i].tilt_target = set[i].tilt_start + A*sin(2.0*UNIT_pi*omega*t); // note this is different than my 2013 paper, we control via displacement now (allows omega = 0.0)
        
        double tilt_vel = (A*2.0*UNIT_pi*omega)*cos(2.0*UNIT_pi*omega*t);
        double lz = domain->boxhi[2] - domain->boxlo[2];
        
        shearRate = tilt_vel/lz;  // the shearRate = vel/L 
                
        //h_rate[i] = TWOPI/set[i].tperiod * set[i].amplitude *
        //  cos(TWOPI*delt/set[i].tperiod);
        //h_ratelo[i] = -0.5*h_rate[i];
        h_rate[2] = tilt_vel; // shearRate;
        h_ratelo[2] = -0.5*h_rate[2]; // PJA: double-check the rates here...

        // update the target values for SELM codes
        //SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist_last 
        //  = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist;
        //SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist
        //  = set[i].tilt_target;
        shearData_RM_OSC1->shearRate = shearRate;
        SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearRate = shearRate;        
      }

      // tilt_target can be large positive or large negative value
      // add/subtract box lengths until tilt_target is closest to current value
      int idenom = 0;
      
      if (i == 5) idenom = 0;
      else if (i == 4) idenom = 0;
      else if (i == 3) idenom = 1;
      
      double denom = set[idenom].hi_target - set[idenom].lo_target;

      double current = h[i]/h[idenom];

      while (set[i].tilt_target/denom - current > 0.0)
        set[i].tilt_target -= denom;
      while (set[i].tilt_target/denom - current < 0.0)
        set[i].tilt_target += denom;
      if (fabs(set[i].tilt_target/denom - 1.0 - current) <
          fabs(set[i].tilt_target/denom - current))
        set[i].tilt_target -= denom;      
    }

    // atz -- (added to move the bottom of the box to make shear symmetric)
    double lx = domain->boxhi[0] - domain->boxlo[0];
    set[0].lo_target = set[0].lo_start - 0.5*set[4].tilt_target;
    set[0].hi_target = set[0].lo_target + lx;
    //set[0].lo_target = set[0].lo_start + delta*(set[0].lo_stop - set[0].lo_start);
    //set[0].hi_target = set[0].lo_target + lx;
    // --
 
  }

  // if any tilt ratios exceed 0.5, set flip = 1 and compute new tilt values
  // do not flip in x or y if non-periodic (can tilt but not flip)
  //   this is b/c the box length would be changed (dramatically) by flip
  // if xz tilt exceeded, adjust C vector by one A vector
  // flip is performed on next timestep, before reneighboring in pre-exchange()
  if (triclinic && flipflag) {
    double xprd = set[0].hi_target - set[0].lo_target;
    double yprd = set[1].hi_target - set[1].lo_target;
    double xprdinv = 1.0 / xprd;
    double yprdinv = 1.0 / yprd;
    if (set[3].tilt_target*yprdinv < -0.5 ||
                                     set[3].tilt_target*yprdinv > 0.5 ||
        set[4].tilt_target*xprdinv < -0.5 ||
                                     set[4].tilt_target*xprdinv > 0.5 ||
        set[5].tilt_target*xprdinv < -0.5 ||
                                     set[5].tilt_target*xprdinv > 0.5) {
      set[3].tilt_flip = set[3].tilt_target;
      set[4].tilt_flip = set[4].tilt_target;
      set[5].tilt_flip = set[5].tilt_target;

      flipxy = flipxz = flipyz = 0;

      if (domain->yperiodic) {
        if (set[3].tilt_flip*yprdinv < -0.5) {
          set[3].tilt_flip += yprd;
          set[4].tilt_flip += set[5].tilt_flip;
          flipyz = 1;
        } else if (set[3].tilt_flip*yprdinv > 0.5) {
          set[3].tilt_flip -= yprd;
          set[4].tilt_flip -= set[5].tilt_flip;
          flipyz = -1;
        }
      }
      if (domain->xperiodic) {
        if (set[4].tilt_flip*xprdinv < -0.5) {
          set[4].tilt_flip += xprd;
          flipxz = 1;
        }
        if (set[4].tilt_flip*xprdinv > 0.5) {
          set[4].tilt_flip -= xprd;
          flipxz = -1;
        }
        if (set[5].tilt_flip*xprdinv < -0.5) {
          set[5].tilt_flip += xprd;
          flipxy = 1;
        }
        if (set[5].tilt_flip*xprdinv > 0.5) {
          set[5].tilt_flip -= xprd;
          flipxy = -1;
        }
      }

      flip = 0;
      if (flipxy || flipxz || flipyz) flip = 1;
      if (flip) fixSELM->next_reneighbor = update->ntimestep + 1;
    }
  }

  // convert atoms and rigid bodies to lamda coords
  if (remapflag == Domain::X_REMAP) {
    double **x = atom->x;
    int *mask = atom->mask;
    int nlocal = atom->nlocal;

    for (i = 0; i < nlocal; i++)
      if (mask[i] & fixSELM->groupbit)
        domain->x2lamda(x[i],x[i]);
  }

  // reset global and local box to new size/shape
  // only if deform fix is controlling the dimension
  if (triclinic) {
    if (set[4].style) {
      domain->xz = set[4].tilt_target;
      domain->boxlo[0] = set[0].lo_target;
      domain->boxhi[0] = set[0].hi_target;
    }
    // @@@ change the xlo and xhi, so bottom also moves
  }

  domain->set_global_box();
  domain->set_local_box();

  // convert atoms and rigid bodies back to box coords
  if (remapflag == Domain::X_REMAP) {
    double **x = atom->x;
    int *mask = atom->mask;
    int nlocal = atom->nlocal;

    for (i = 0; i < nlocal; i++)
      if (mask[i] & fixSELM->groupbit)
        domain->lamda2x(x[i],x[i]);
  }
  
  // redo KSpace coeffs since box has changed
  if (kspace_flag) lammps->force->kspace->setup();

}

void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::setGlobalRefs(LAMMPS *lmps, FixSELM *fx) {

  /* setup the LAMMPS related information */
  lammps  = lmps;
  fixSELM = fx;

  //if (force_reneighbor)
  //  irregular = new Irregular(lammps);
  // (setup in constructor)
  //else
  //  irregular = NULL;

}

void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::init() {

  type = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::TYPE;
  strcpy(typeStr, SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::TYPE_STR);

}


void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::parse_ParameterFile(char *baseFilename) {

  const char *error_str_func = "parse_ParameterFile()";

//  char filename[1000];
//
//  SELM_Parser1 *parser;
//
//  int j, k;
//
//  int paramIndex;
//  int areAllSetFlag;
//
//  char typeString[1000];
//
//  int numIndicesCase_RM_SHEAR1;
//  int indicesCase_RM_SHEAR1[1000];
//  int ptI_case_RM_SHEAR1;
//
//  int numIndicesCase_RM_OSC1;
//  int indicesCase_RM_OSC1[1000];
//  int ptI_case_RM_OSC1;
//
//  int numIndices;
//  int *indices;
//
//  int flagParseFailed;
//
//  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_ParamsType *params;
//
//  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params
//  = (SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_ParamsType *)
//  malloc(sizeof(SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_ParamsType));
//
//  params = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params;
//
//  sprintf(filename, "%s.SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3", baseFilename);
//
//  parser = new SELM_Parser1();
//
//  /* Setup the parsing data structure */
//  SELM_Parser1::paramSpecificationType *paramSpecification =
//      (SELM_Parser1::paramSpecificationType *) malloc(sizeof(SELM_Parser1::paramSpecificationType));
//
//  paramSpecification->paramDescrList
//  = (SELM_Parser1::paramDescrType *) malloc(sizeof(SELM_Parser1::paramDescrType) * SELM_Parser1::MAX_NUM_PARAMS);
//
//  paramIndex = 0;
//
//  /* initialize the parameters consistency set */
//  ptI_case_RM_OSC1 = 0;
//  ptI_case_RM_SHEAR1 = 0;
//
//  numIndicesCase_RM_OSC1 = 0;
//  numIndicesCase_RM_SHEAR1 = 0;
//
//  /* initialize the parameters */
//  params->maxTimeStepIndex = 0;
//  params->deltaT = 0;
//  params->mu = 0;
//  params->rho = 0;
//  params->KB = 0;
//  params->T = 0;
//  params->flagIncompressibleFluid = 0;
//  params->flagShearMode = 0;
//  params->flagStochasticDriving = 0;
//  params->shearOmega = 0;
//  params->shearRateAmplitude = 0;
//
//  /* setup the parser information */
//  strcpy(paramSpecification->paramDescrList[paramIndex].paramName,
//         "maxTimeStepIndex");
//  paramSpecification->paramDescrList[paramIndex].paramType = SELM_Parser1::PARAMTYPE_INT;
//  paramSpecification->paramDescrList[paramIndex].paramSetFlag = 0;
//  paramSpecification->paramDescrList[paramIndex].paramVar
//  = &params->maxTimeStepIndex;
//  paramSpecification->paramDescrList[paramIndex].paramExtras = NULL;
//  indicesCase_RM_SHEAR1[ptI_case_RM_SHEAR1] = paramIndex;
//  ptI_case_RM_SHEAR1++;
//  indicesCase_RM_OSC1[ptI_case_RM_OSC1] = paramIndex;
//  ptI_case_RM_OSC1++;
//  paramIndex++;
//
//  strcpy(paramSpecification->paramDescrList[paramIndex].paramName, "deltaT");
//  paramSpecification->paramDescrList[paramIndex].paramType = SELM_Parser1::PARAMTYPE_DOUBLE;
//  paramSpecification->paramDescrList[paramIndex].paramSetFlag = 0;
//  paramSpecification->paramDescrList[paramIndex].paramVar = &params->deltaT;
//  paramSpecification->paramDescrList[paramIndex].paramExtras = NULL;
//  indicesCase_RM_SHEAR1[ptI_case_RM_SHEAR1] = paramIndex;
//  ptI_case_RM_SHEAR1++;
//  indicesCase_RM_OSC1[ptI_case_RM_OSC1] = paramIndex;
//  ptI_case_RM_OSC1++;
//  paramIndex++;
//
//  strcpy(paramSpecification->paramDescrList[paramIndex].paramName, "mu");
//  paramSpecification->paramDescrList[paramIndex].paramType = SELM_Parser1::PARAMTYPE_DOUBLE;
//  paramSpecification->paramDescrList[paramIndex].paramSetFlag = 0;
//  paramSpecification->paramDescrList[paramIndex].paramVar = &params->mu;
//  paramSpecification->paramDescrList[paramIndex].paramExtras = NULL;
//  indicesCase_RM_SHEAR1[ptI_case_RM_SHEAR1] = paramIndex;
//  ptI_case_RM_SHEAR1++;
//  indicesCase_RM_OSC1[ptI_case_RM_OSC1] = paramIndex;
//  ptI_case_RM_OSC1++;
//  paramIndex++;
//
//  strcpy(paramSpecification->paramDescrList[paramIndex].paramName, "rho");
//  paramSpecification->paramDescrList[paramIndex].paramType = SELM_Parser1::PARAMTYPE_DOUBLE;
//  paramSpecification->paramDescrList[paramIndex].paramSetFlag = 0;
//  paramSpecification->paramDescrList[paramIndex].paramVar = &params->rho;
//  paramSpecification->paramDescrList[paramIndex].paramExtras = NULL;
//  indicesCase_RM_SHEAR1[ptI_case_RM_SHEAR1] = paramIndex;
//  ptI_case_RM_SHEAR1++;
//  indicesCase_RM_OSC1[ptI_case_RM_OSC1] = paramIndex;
//  ptI_case_RM_OSC1++;
//  paramIndex++;
//
//  strcpy(paramSpecification->paramDescrList[paramIndex].paramName, "KB");
//  paramSpecification->paramDescrList[paramIndex].paramType = SELM_Parser1::PARAMTYPE_DOUBLE;
//  paramSpecification->paramDescrList[paramIndex].paramSetFlag = 0;
//  paramSpecification->paramDescrList[paramIndex].paramVar = &params->KB;
//  paramSpecification->paramDescrList[paramIndex].paramExtras = NULL;
//  indicesCase_RM_SHEAR1[ptI_case_RM_SHEAR1] = paramIndex;
//  ptI_case_RM_SHEAR1++;
//  indicesCase_RM_OSC1[ptI_case_RM_OSC1] = paramIndex;
//  ptI_case_RM_OSC1++;
//  paramIndex++;
//
//  strcpy(paramSpecification->paramDescrList[paramIndex].paramName, "T");
//  paramSpecification->paramDescrList[paramIndex].paramType = SELM_Parser1::PARAMTYPE_DOUBLE;
//  paramSpecification->paramDescrList[paramIndex].paramSetFlag = 0;
//  paramSpecification->paramDescrList[paramIndex].paramVar = &params->T;
//  paramSpecification->paramDescrList[paramIndex].paramExtras = NULL;
//  indicesCase_RM_SHEAR1[ptI_case_RM_SHEAR1] = paramIndex;
//  ptI_case_RM_SHEAR1++;
//  indicesCase_RM_OSC1[ptI_case_RM_OSC1] = paramIndex;
//  ptI_case_RM_OSC1++;
//  paramIndex++;
//
//  strcpy(paramSpecification->paramDescrList[paramIndex].paramName,
//         "flagShearMode");
//  paramSpecification->paramDescrList[paramIndex].paramType = SELM_Parser1::PARAMTYPE_INT;
//  paramSpecification->paramDescrList[paramIndex].paramSetFlag = 0;
//  paramSpecification->paramDescrList[paramIndex].paramVar
//  = &params->flagShearMode;
//  paramSpecification->paramDescrList[paramIndex].paramExtras = NULL;
//  indicesCase_RM_SHEAR1[ptI_case_RM_SHEAR1] = paramIndex;
//  ptI_case_RM_SHEAR1++;
//  indicesCase_RM_OSC1[ptI_case_RM_OSC1] = paramIndex;
//  ptI_case_RM_OSC1++;
//  paramIndex++;
//
//  strcpy(paramSpecification->paramDescrList[paramIndex].paramName, "shearOmega");
//  paramSpecification->paramDescrList[paramIndex].paramType = SELM_Parser1::PARAMTYPE_DOUBLE;
//  paramSpecification->paramDescrList[paramIndex].paramSetFlag = 0;
//  paramSpecification->paramDescrList[paramIndex].paramVar = &params->shearOmega;
//  paramSpecification->paramDescrList[paramIndex].paramExtras = NULL;
//  indicesCase_RM_OSC1[ptI_case_RM_OSC1] = paramIndex;
//  ptI_case_RM_OSC1++;
//  paramIndex++;
//
//  strcpy(paramSpecification->paramDescrList[paramIndex].paramName,
//         "shearRateAmplitude");
//  paramSpecification->paramDescrList[paramIndex].paramType = SELM_Parser1::PARAMTYPE_DOUBLE;
//  paramSpecification->paramDescrList[paramIndex].paramSetFlag = 0;
//  paramSpecification->paramDescrList[paramIndex].paramVar
//  = &params->shearRateAmplitude;
//  paramSpecification->paramDescrList[paramIndex].paramExtras = NULL;
//  indicesCase_RM_OSC1[ptI_case_RM_OSC1] = paramIndex;
//  ptI_case_RM_OSC1++;
//  paramIndex++;
//
//  strcpy(paramSpecification->paramDescrList[paramIndex].paramName,
//         "flagStochasticDriving");
//  paramSpecification->paramDescrList[paramIndex].paramType = SELM_Parser1::PARAMTYPE_INT;
//  paramSpecification->paramDescrList[paramIndex].paramSetFlag = 0;
//  paramSpecification->paramDescrList[paramIndex].paramVar
//  = &params->flagStochasticDriving;
//  paramSpecification->paramDescrList[paramIndex].paramExtras = NULL;
//  indicesCase_RM_SHEAR1[ptI_case_RM_SHEAR1] = paramIndex;
//  ptI_case_RM_SHEAR1++;
//  indicesCase_RM_OSC1[ptI_case_RM_OSC1] = paramIndex;
//  ptI_case_RM_OSC1++;
//  paramIndex++;
//
//  strcpy(paramSpecification->paramDescrList[paramIndex].paramName,
//         "flagIncompressibleFluid");
//  paramSpecification->paramDescrList[paramIndex].paramType = SELM_Parser1::PARAMTYPE_INT;
//  paramSpecification->paramDescrList[paramIndex].paramSetFlag = 0;
//  paramSpecification->paramDescrList[paramIndex].paramVar
//  = &params->flagIncompressibleFluid;
//  paramSpecification->paramDescrList[paramIndex].paramExtras = NULL;
//  indicesCase_RM_SHEAR1[ptI_case_RM_SHEAR1] = paramIndex;
//  ptI_case_RM_SHEAR1++;
//  indicesCase_RM_OSC1[ptI_case_RM_OSC1] = paramIndex;
//  ptI_case_RM_OSC1++;
//  paramIndex++;
//
//  strcpy(paramSpecification->paramDescrList[paramIndex].paramName,
//         "flagWriteSimulationData");
//  paramSpecification->paramDescrList[paramIndex].paramType = SELM_Parser1::PARAMTYPE_INT;
//  paramSpecification->paramDescrList[paramIndex].paramSetFlag = 0;
//  paramSpecification->paramDescrList[paramIndex].paramVar
//  = &params->flagWriteSimulationData;
//  paramSpecification->paramDescrList[paramIndex].paramExtras = NULL;
//  indicesCase_RM_SHEAR1[ptI_case_RM_SHEAR1] = paramIndex;
//  ptI_case_RM_SHEAR1++;
//  indicesCase_RM_OSC1[ptI_case_RM_OSC1] = paramIndex;
//  ptI_case_RM_OSC1++;
//  paramIndex++;
//
//  strcpy(paramSpecification->paramDescrList[paramIndex].paramName,
//         "saveSkipSimulationData");
//  paramSpecification->paramDescrList[paramIndex].paramType = SELM_Parser1::PARAMTYPE_INT;
//  paramSpecification->paramDescrList[paramIndex].paramSetFlag = 0;
//  paramSpecification->paramDescrList[paramIndex].paramVar
//  = &params->saveSkipSimulationData;
//  paramSpecification->paramDescrList[paramIndex].paramExtras = NULL;
//  indicesCase_RM_SHEAR1[ptI_case_RM_SHEAR1] = paramIndex;
//  ptI_case_RM_SHEAR1++;
//  indicesCase_RM_OSC1[ptI_case_RM_OSC1] = paramIndex;
//  ptI_case_RM_OSC1++;
//  paramIndex++;
//
//  numIndicesCase_RM_SHEAR1 = ptI_case_RM_SHEAR1;
//  numIndicesCase_RM_OSC1 = ptI_case_RM_OSC1;
//
//  paramSpecification->numParams = paramIndex;
//
//  /* Parse the parameters */
//  parser->parseParameters(filename, paramSpecification);
//
//  /* Check whethor all parameters were set */
//  areAllSetFlag = parser->areAllParametersSet(paramSpecification);
//
//  switch (params->flagShearMode) {
//
//  case SHEAR_MODE_TYPE_RM_SHEAR1:
//
//    flagParseFailed = 0;
//    for (k = 0; k < numIndicesCase_RM_SHEAR1; k++) {
//      if (paramSpecification->paramDescrList[indicesCase_RM_SHEAR1[k]].paramSetFlag
//          == 0) {
//        flagParseFailed = 1;
//      }
//    }
//    break;
//
//  case SHEAR_MODE_TYPE_RM_OSC1:
//
//    flagParseFailed = 0;
//    for (k = 0; k < numIndicesCase_RM_OSC1; k++) {
//      if (paramSpecification->paramDescrList[indicesCase_RM_OSC1[k]].paramSetFlag
//          == 0) {
//        flagParseFailed = 1;
//      }
//    }
//    break;
//
//  } /* end of switch */
//
//  if (flagParseFailed) {
//    printf("ERROR: %s : %s \n", error_str_code, error_str_func);
//    printf("The following Parameters were not set: \n");
//
//    switch (params->flagShearMode) {
//    case SHEAR_MODE_TYPE_RM_SHEAR1:
//      numIndices = numIndicesCase_RM_SHEAR1;
//      indices = indicesCase_RM_SHEAR1;
//      break;
//    case SHEAR_MODE_TYPE_RM_OSC1:
//      numIndices = numIndicesCase_RM_OSC1;
//      indices = indicesCase_RM_OSC1;
//      break;
//    } /* end of switch */
//
//    for (k = 0; k < numIndices; k++) {
//      switch (paramSpecification->paramDescrList[indices[k]].paramType) {
//      case SELM_Parser1::PARAMTYPE_INT:
//        strcpy(typeString, "Integer");
//        break;
//      case SELM_Parser1::PARAMTYPE_DOUBLE:
//        strcpy(typeString, "Double");
//        break;
//      case SELM_Parser1::PARAMTYPE_STRING:
//        strcpy(typeString, "String");
//        break;
//      case SELM_Parser1::PARAMTYPE_INT_LIST:
//        strcpy(typeString, "Integer List");
//        break;
//      case SELM_Parser1::PARAMTYPE_DOUBLE_LIST:
//        strcpy(typeString, "Double List");
//        break;
//      case SELM_Parser1::PARAMTYPE_STRING_LIST:
//        strcpy(typeString, "String List (\0 termination tag)");
//        break;
//      }
//
//      printf("%s : %s \n",
//             paramSpecification->paramDescrList[indices[k]].paramName,
//             typeString);
//    } /* end of k loop */
//
//    /* exit with error */
//    exit(1);
//
//  }
//
//  /* Show parameters parsed */
//  printf("Parameters for SELM_Integrator_%s: \n", typeStr);
//  parser->printParameters(paramSpecification);
//
//  /* Clean up the parsing data structures */
//  free(paramSpecification->paramDescrList);
//  free(paramSpecification);
//
//  delete parser;

}

void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::setup() {

  /* setup the integrator data structure */
  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras
    = (SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_ExtrasType *)
      malloc(sizeof(SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_ExtrasType));

  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->deltaT
  = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params->deltaT;

  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->maxTimeStepIndex
    = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params->maxTimeStepIndex;

  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->mu
    = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params->mu;

  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->rho
      = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params->rho;

  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->KB
      = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params->KB;

  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->T
        = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params->T;

  strcpy(SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagShearModeStr,
         SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params->flagShearModeStr);

  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagShearMode
  = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params->flagShearMode;

  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->shearData
  = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params->shearData;

  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->maxTimeStepIndex
  = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params->maxTimeStepIndex;

  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagStochasticDriving
  = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params->flagStochasticDriving;

  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagIncompressibleFluid
  = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params->flagIncompressibleFluid;

  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagUpdateControlPts           = 1;

  SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagInitializedNumericalMethod = 1;

  flagWriteSimulationData = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params->flagWriteSimulationData;
  saveSkipSimulationData  = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params->saveSkipSimulationData;

  free(SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Params);
  
  // now since initialized, setup the internal init() for the class
  setup_internals();
  
}

void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::packageError(int code, void *extras) {
    std::exit(code);
}


void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::set_LAMMPS_mask(int *mask_ptr) {

  /* this determines if LAMMPS calls both the initial and final integrator commands */
  /* this command augments the mask flags to control this behavior */
  // WARNING: Might need to add the above bits to ensure integrates also initial() and final() time-step 

  //(*mask_ptr) = setmask(); // set the mask value (handled now below to make more explicit) 

  (*mask_ptr) = 0;
  if (fixSELM->force_reneighbor) (*mask_ptr) |= PRE_EXCHANGE;
  (*mask_ptr) |= END_OF_STEP;

  (*mask_ptr) |= INITIAL_INTEGRATE; /* tells LAMMPS to call integrate_initial via fix_SELM */
  (*mask_ptr) |= FINAL_INTEGRATE;   /* tells LAMMPS to call integrate_final via fix_SELM */

}

void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::integrate_initialize() {

  const char *error_str_func = "integrate_initialize()";

}


void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::integrate_initial() {

  const char *error_str_func = "integrate_initial()";

  /* synchronize the lammps domain with the fluid domain for shear deformation */
  syncShearDomainWithLammpsDomain();

  /* now done internally within fluid-structure update */
  /* syncronize the lammps data and control points data to one another */
  //for (int I = 0; I < fixSELM->SELM_Lagrangian_List_N; I++) {
  //  fixSELM->SELM_Lagrangian_List[I]->setControlPtsDataFromLammpsData();
  //}

  /* perform the calculations before the integrator is called */
  IB_appl1_start_time_step_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3();

  SELM_updateFluidAndStructures_initial();

  /* now done internally during update */
  /* synchronize the lammps data and control points data to one another */
  //for (int I = 0; I < fixSELM->SELM_Lagrangian_List_N; I++) {
  //  fixSELM->SELM_Lagrangian_List[I]->setLammpsDataFromControlPtsData();
  //}

  /* synchronize the lammps domain with the fluid domain for shear deformation */
  /* setLammpsDomainFromShearDomain(); */

}


void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::integrate_final() {

  const char *error_str_func = "integrate_final()";

  SELM_updateFluidAndStructures_final();

}


void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::SELM_updateParticlesTest_initial()
{

  const char *error_str_func = "SELM_updateParticlesTest_initial()";


  Atom   *atom    = lammps->atom;
  int    igroup   = fixSELM->igroup;
  int    groupbit = fixSELM->groupbit;

  double **x      = atom->x;
  double **v      = atom->v;
  double **f      = atom->f;
  double *rmass   = atom->rmass;
  double *mass    = atom->mass;
  int    *type    = atom->type;
  int    *mask    = atom->mask;
  int     nlocal  = atom->nlocal;

  double  dtfm    = 1.0/0.0;
  double  dtv     = 1.0/0.0;
  double  dtf     = 1.0/0.0;

  if (igroup == atom->firstgroup) nlocal = atom->nfirst;

  if (rmass) {
    for (int i = 0; i < nlocal; i++)
      if (mask[i] & groupbit) {
        dtfm = dtf / rmass[i];
        v[i][0] += dtfm * f[i][0];
        v[i][1] += dtfm * f[i][1];
        v[i][2] += dtfm * f[i][2];
        x[i][0] += dtv * v[i][0];
        x[i][1] += dtv * v[i][1];
        x[i][2] += dtv * v[i][2];
      }

  } else {
    for (int i = 0; i < nlocal; i++)
      if (mask[i] & groupbit) {
        dtfm = dtf / mass[type[i]];
        v[i][0] += dtfm * f[i][0];
        v[i][1] += dtfm * f[i][1];
        v[i][2] += dtfm * f[i][2];
        x[i][0] += dtv * v[i][0];
        x[i][1] += dtv * v[i][1];
        x[i][2] += dtv * v[i][2];
      }
  }

  /*
   % get parameter values
    num_dim             = dataParticles.num_dim;
    numParticles        = dataParticles.numParticles;
    deltaT2             = dataParticles.deltaT2;

    funcForce           = dataParticles.funcForce;
    funcForceExtras     = dataParticles.funcForceExtras;

    funcOpUpsilon       = dataParticles.funcOpUpsilon;
    funcOpUpsilonExtras = dataParticles.funcOpUpsilonExtras;

    Gamma_u_val         = dataParticles.Gamma_u_val;

    gamma               = dataParticles.gamma;
    mass                = dataParticles.mass;
    deltaT2             = dataParticles.deltaT2;
    KBT                 = dataParticles.KBT;

    %sigma_Lngv      = dataParticles.sigma_Lngv;

    sigma_Lngv          = sqrt(2*KBT*(gamma/(mass^2))*0.5*deltaT2);

    % integrate V from n to n + 1/2
    F_thm_half1_dT  = sigma_Lngv*randn(num_dim, numParticles);
    F_drag_half1_dT = 0.5*deltaT2*(-gamma/mass)*(V_n  - Gamma_u_val);

    V_nph           = V_n + F_drag_half1_dT + F_thm_half1_dT;

    F_n             = feval(funcForce, X_n, funcForceExtras);
    V_nph           = V_nph + (0.5*deltaT2*F_n/mass);

    % integrate X from n to n + 1
    X_np1           = X_n + deltaT2*V_nph;

    % integrate V from n + 1/2 to n + 1
    F_thm_half2_dT  = sigma_Lngv*randn(num_dim, numParticles);
    F_drag_half2_dT = 0.5*deltaT2*(-gamma/mass)*(V_nph - Gamma_u_val);

    V_np1           = V_nph + F_drag_half2_dT + F_thm_half2_dT;

    F_np1           = feval(funcForce, X_np1, funcForceExtras);
    V_np1           = V_np1 + (0.5*deltaT2*F_np1/mass);

    outputParticles                 = [];
    outputParticles.F_thm_half1_dT  = F_thm_half1_dT;
    outputParticles.F_thm_half2_dT  = F_thm_half2_dT;
    outputParticles.F_drag_half1_dT = F_drag_half1_dT;
    outputParticles.F_drag_half2_dT = F_drag_half2_dT;

 */


  /*Compute the fluid velocity of the control points from the FFTW3 uniform mesh. */
  /*
    void SELM_computeControlPtsVel_FFTW3(int                 num_dim,
                                             int                 numMeshPtsPerDir[3],
                                             double              meshDeltaX,
                                             double             *meshCenterX0,
                                             fftw_complex       *u_m[3],
                                             controlPtsDataType *SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE,
                                             int                 flagPeriodicCondOnPts);
   */

  /* == Compute the fluid velocity of the control points for FFTW3 uniform mesh
       with Lees-Edwards boundary conditions. */
  /*
    void SELM_computeControlPtsVel_SHEAR_FFTW3(int                 num_dim,
                                                   int                 numMeshPtsPerDir[3],
                                                   double              meshDeltaX,
                                                   double             *meshCenterX0,
                                                   fftw_complex       *u_m[3],
                                                   controlPtsDataType *SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE,
                                                   double              shearRate,
                                                   int                 shearDir,
                                                   int                 shearVelDir,
                                                   double              shearDist,
                                                   int                 flagPeriodicCondOnPts);
   */



}


void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::SELM_updateParticlesTest_final()
{

  Atom   *atom    = lammps->atom;
  int    igroup   = fixSELM->igroup;
  int    groupbit = fixSELM->groupbit;

  double **x      = atom->x;
  double **v      = atom->v;
  double **f      = atom->f;
  double *rmass   = atom->rmass;
  double *mass    = atom->mass;
  int    *type    = atom->type;
  int    *mask    = atom->mask;
  int     nlocal  = atom->nlocal;

}


void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::SELM_updateFluidAndStructures_initial()
{

  const char *error_str_func = "SELM_updateFluidAndStructures_initial()";

  Domain *domain = lammps->domain;
  Atom   *atom    = lammps->atom;

  int d;
  int i, j, k, m, ell;
  int I;
  int k1, k2, k3;
  int j1, j2, j3;
  double a_k;
  double c_k1, c_k2, c_k3;
  int num_dim;
  int *numMeshPtsPerDir;
  double meshDeltaX;
  double deltaT;
  double g_k_real[3];
  double g_k_imag[3];

  double g_k_dot_u_real;
  double g_k_dot_u_imag;

  double g_k_dot_f_real;
  double g_k_dot_f_imag;

  double g_k_sq;

  double g_k_dot_u_real_over_g_k_sq;
  double g_k_dot_u_imag_over_g_k_sq;

  double g_k_dot_f_real_over_g_k_sq;
  double g_k_dot_f_imag_over_g_k_sq;

  double vec_proj_real[3];
  double vec_proj_imag[3];

  double normalizeFluidDensity_k;

  double mu;
  double rho;
  double KB;
  double T;

  int flagShearMode;

  int    shearDir;
  int    shearVelDir;
  double shearRate;
  double shearDist;
  double shearDist_last;

  double shearOmega;
  double shearRateAmplitude;

  int a, b;

  int N_a, N_b;

  double shearFlow_k_j[2][3];

  double vec_L[3];
  int vec_k[3];
  int vec_j[3];

  double X_sq;
  double X[3];

  int k_a, k_b;

  double L_a, L_b;
  double rep_d_2_w_j__d_q_a__d_q_b;
  double a_k_j[3];

  int delta_a_velDir;
  int delta_b_velDir;

  int delta_a_ell;
  int delta_b_ell;

  int delta_ell_shearDir;

  int flag;

  double sigma_sq;
  
  double dot_gamma_t;

  /*
  int timeIndex = aibControls->timeIndex;
  */
  int timeIndex;

  double curTime;

  double L_shearDir;
  double L_shearVelDir;

  int N_shear;
  int k_shear;

  int flagTestProblem;

  int flagUseFiniteDifference;
  int flagUseSpectralRep;

  int imageI;

  double *meshCenterX0;
  double  meshBaseX0[3];

  double shearAdj;

  double scaleFactor;

  double shearDistMod;

  int    flagImposeUnitCellRep;
  double X1[3];
  double XX1[3];
  double shearX1[3];
  double shearXX1[3];
  double periodL[3];

  int    flagDebugViaControlPtsUpdate = 0;

  double *controlPt_X;

  int flipToLessShearedUnitCell = 0; /* flag indicating if equivalent less sheared
                                        unit cell used, see comments below */
                                   
  int flag_shear_remap_here = 0; /* when 1 performs remap as part of this routine, to try to mimic behaviors of "fix deform" case */

  ShearData_RM_SHEAR1_Type *shearData_RM_SHEAR1;
  ShearData_RM_OSC1_Type   *shearData_RM_OSC1;

  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3                                                       *SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3;
  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3::SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_ExtrasType *SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras;

  /* WARNING: PJA: Why computing Gamma here... should justify with comment and why only 0 coupling operator. */
  /*
  SELM_CouplingOperator *couplingOp = fixSELM->SELM_CouplingOperator_List[0];
  if (fixSELM->SELM_Lagrangian_List[0]->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {
    SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *lagrangian = (SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *)SELM_LagrangianData;    
    couplingOp->computeOperatorGamma(lagrangian, eulerian, eulerianV, pseudoVel);
  } else if (fixSELM->SELM_Lagrangian_List[0]->type == SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) { 
    SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *lagrangian = (SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *)SELM_LagrangianData;
    couplingOp->computeOperatorGamma(lagrangian, eulerian, eulerianV, pseudoVel);
  } else if (fixSELM->SELM_Lagrangian_List[0]->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID) { 
    SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *lagrangian = (SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *)SELM_LagrangianData;
    couplingOp->computeOperatorGamma(lagrangian, eulerian, eulerianV, pseudoVel);    
  } else {
    stringstream message;
    message << "Lagrangian Type is not recognized" << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }
  */

  timeIndex = lammps->update->ntimestep; /* ### */
  /* deltaT = lammps->update->dt; */ /* LAMMPS overrides the deltaT given in the XML file */

#ifdef USE_PACKAGE_FFTW3

  /* == Dereference the data (WARNING: assumes only one Eulerian DOF for now) */
  if (fixSELM->SELM_Eulerian_List[0]->type == SELM_Eulerian_Types::TYPE_LAMMPS_SHEAR_UNIFORM1_FFTW3) {

    SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3
      = (SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3 *)fixSELM->SELM_Eulerian_List[0];

    SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras
      = SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3->SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras;

    num_dim          = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->num_dim;
    meshDeltaX       = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->meshDeltaX;
    numMeshPtsPerDir = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->numMeshPtsPerDir;
    meshCenterX0     = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->meshCenterX0;

  } else {
    stringstream message;
    message << "Expecting mesh type: " << SELM_Eulerian_Types::TYPE_STR_LAMMPS_SHEAR_UNIFORM1_FFTW3 << endl;
    message << "Instead mesh type was: " << fixSELM->SELM_Eulerian_List[0]->typeStr << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  if (this->type == SELM_Integrator_Types::TYPE_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3) {

    /* time stepping parameters */
    deltaT = lammps->update->dt;
    SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->deltaT = deltaT; /* LAMMPS overrides the deltaT given in the XML file */ 
    
    curTime       = timeIndex*deltaT;
    flagShearMode = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagShearMode;

    /* fluid parameters */
    mu  = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->mu;
    rho = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->rho;
    KB  = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->KB;
    T   = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->T;

    switch (flagShearMode) {

    case SHEAR_MODE_TYPE_RM_SHEAR1:

      shearData_RM_SHEAR1 = (ShearData_RM_SHEAR1_Type *)
          SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->shearData;

      /* retrieve shear values required to generate the flow */
      shearDir       = shearData_RM_SHEAR1->shearDir;
      shearVelDir    = shearData_RM_SHEAR1->shearVelDir;
      shearRate      = shearData_RM_SHEAR1->shearRate;
      shearDist      = shearData_RM_SHEAR1->shearDist;
      shearDist_last = shearData_RM_SHEAR1->shearDist_last;

      /* setup the mesh in accordance of the current shear conditions */
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDir
        = shearDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearVelDir
        = shearVelDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearRate
        = shearRate;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist
        = shearDist;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist_last
        = shearDist_last;

      break;

    case SHEAR_MODE_TYPE_RM_OSC1:

      shearData_RM_OSC1
        = (ShearData_RM_OSC1_Type *)
        SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->shearData;

      /* update the shear rate (oscillation) */
      shearDir           = shearData_RM_OSC1->shearDir;
      shearVelDir        = shearData_RM_OSC1->shearVelDir;      
      shearOmega         = shearData_RM_OSC1->shearOmega;
      shearRateAmplitude = shearData_RM_OSC1->shearRateAmplitude;
      shearRate          = shearData_RM_OSC1->shearRate;
      shearDist          = shearData_RM_OSC1->shearDist;
      shearDist_last     = shearData_RM_OSC1->shearDist_last;

      /* calculate the new shear rate */
      //shearRate                    = shearRateAmplitude*cos(curTime * shearOmega);  // PJA: New way is to have end_of_step(), init() set rates.
      //shearData_RM_OSC1->shearRate = shearRate;

      /* setup the mesh in accordance of the current shear conditions */
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDir
        = shearDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearVelDir
        = shearVelDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearRate
        = shearRate;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist
        = shearDist;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist_last
        = shearDist_last;

      break;

    } /* end of switch */

  } else {
    stringstream message;
    message << "Expecting time stepper: " << SELM_Integrator_Types::TYPE_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3 << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }
  
  /* == Update the Control Points from the LAMMPS data */
  /* synchronize the lammps data and control points data to one another */
  for (int I = 0; I < fixSELM->SELM_Lagrangian_List_N; I++) {
    fixSELM->SELM_Lagrangian_List[I]->setControlPtsDataFromLammpsData();
  }
  
  /* == Add the Pseudoforce to the force density of the fluid */
  /* compute the pseudo-force term arising from the shear boundary condition */
  addPseudoForceTerm(SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->num_dim,
                     SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->meshDeltaX,
                     SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->numMeshPtsPerDir,
                     mu,shearDir,shearVelDir,shearRate,
                     SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidForceDensity_m);

  flagTestProblem = 0;
  switch (flagTestProblem) {

  case 1:
    printf("WARNING: %s : %s \n", error_str_code, error_str_func);
    printf("Test problem flag adds term to the force density field. \n");
    printf("flagTestProblem = %d \n", flagTestProblem);

    scaleFactor = 1e-2;
    sigma_sq = 3 * 3;
    meshCenterX0 = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->meshCenterX0;
    for (d = 0; d < num_dim; d++) {
      vec_L[d] = numMeshPtsPerDir[d] * meshDeltaX;
      meshBaseX0[d] = meshCenterX0[d] - (vec_L[d] / 2.0);
    }

    L_shearDir = numMeshPtsPerDir[shearDir] * meshDeltaX;

    for (j1 = 0; j1 < numMeshPtsPerDir[0]; j1++) {
      for (j2 = 0; j2 < numMeshPtsPerDir[1]; j2++) {
        for (j3 = 0; j3 < numMeshPtsPerDir[2]; j3++) {

          I = (j3 * numMeshPtsPerDir[1] * numMeshPtsPerDir[0]) + (j2
              * numMeshPtsPerDir[0]) + j1;

          vec_j[0] = j1;
          vec_j[1] = j2;
          vec_j[2] = j3;
          for (d = 0; d < num_dim; d++) {
            X[d] = vec_j[d] * meshDeltaX + meshBaseX0[d];
          } /* end of d loop */

          shearAdj = (shearDist / L_shearDir) * (X[shearDir]
                                                   - meshCenterX0[shearDir]);
          X[shearVelDir] += shearAdj;

          X_sq = 0;
          for (d = 0; d < num_dim; d++) {
            X_sq += (X[d] - meshCenterX0[d]) * (X[d] - meshCenterX0[d]);
          }

          for (d = 0; d < num_dim; d++) {

            if (d == 0) {
              SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidForceDensity_m[d][I][0]
                                                                                         += -scaleFactor * mu * ((X_sq / (sigma_sq * sigma_sq)) - (3.0
                                                                                             / sigma_sq)) * exp(-X_sq / (2 * sigma_sq));
              SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidForceDensity_m[d][I][1]
                                                                                         += 0;
            }
          } /* end of d loop */

        } /* end of j3 loop */
      } /* end of j2 loop */
    } /* end of j1 loop */

    break;

  case 2:
    printf("WARNING: %s : %s \n", error_str_code, error_str_func);
    printf("Test problem flag adds term to the force density field. \n");
    printf("flagTestProblem = %d \n", flagTestProblem);

    scaleFactor = 1e-2;
    sigma_sq = 3 * 3;
    meshCenterX0 = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->meshCenterX0;
    for (d = 0; d < num_dim; d++) {
      vec_L[d] = numMeshPtsPerDir[d] * meshDeltaX;
      meshBaseX0[d] = meshCenterX0[d] - (vec_L[d] / 2.0);
    }

    L_shearDir = numMeshPtsPerDir[shearDir] * meshDeltaX;

    for (j1 = 0; j1 < numMeshPtsPerDir[0]; j1++) {
      for (j2 = 0; j2 < numMeshPtsPerDir[1]; j2++) {
        for (j3 = 0; j3 < numMeshPtsPerDir[2]; j3++) {

          I = (j3 * numMeshPtsPerDir[1] * numMeshPtsPerDir[0]) + (j2
              * numMeshPtsPerDir[0]) + j1;

          vec_j[0] = j1;
          vec_j[1] = j2;
          vec_j[2] = j3;
          for (d = 0; d < num_dim; d++) {
            X[d] = vec_j[d] * meshDeltaX + meshBaseX0[d];
          } /* end of d loop */

          shearAdj = (shearDist / L_shearDir) * (X[shearDir]
                                                   - meshCenterX0[shearDir]);
          X[shearVelDir] += shearAdj;

          for (d = 0; d < num_dim; d++) {

            X_sq = 0;
            for (m = 0; m < num_dim; m++) {
              if (m != d) {
                X_sq += (X[m] - meshCenterX0[m]) * (X[m] - meshCenterX0[m]);
              }
            }

            if (d != shearDir) {
              SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidForceDensity_m[d][I][0]
                                                                                         += -scaleFactor * mu * ((X_sq / (sigma_sq * sigma_sq)) - (2.0
                                                                                             / sigma_sq)) * exp(-X_sq / (2 * sigma_sq));
              SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidForceDensity_m[d][I][1]
                                                                                         += 0;
            }

          } /* end of d loop */

        } /* end of j3 loop */
      } /* end of j2 loop */
    } /* end of j1 loop */

    break;

  } /* end of switch */

  /* == Compute the DFT of the force density of the fluid */
  normalizeFluidDensity_k = 1;
  for (d = 0; d < num_dim; d++) {
    fftw_execute(SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidForceDensity_DFT_plan[d]);
    normalizeFluidDensity_k = normalizeFluidDensity_k * numMeshPtsPerDir[d];
  }

  /* == Compute the steady-state fluid velocity */
  if (num_dim == 2) {

    stringstream message;
    message << "2D Steady-State Sheared Stokes flow not implemented." << endl;
    message << "Fluid domain is periodic with Lees-Edwards boundary conditions." << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);

    for (k2 = 0; k2 < numMeshPtsPerDir[1]; k2++) {
      for (k1 = 0; k1 < numMeshPtsPerDir[0]; k1++) {

        I = k2 * numMeshPtsPerDir[0] + k1;

        if ((k1 != 0) || (k2 != 0)) {
          a = 2.0 * mu / (meshDeltaX * meshDeltaX);
          c_k1 = 1 - cos(2 * UNIT_pi * k1 / numMeshPtsPerDir[0]);
          c_k2 = 1 - cos(2 * UNIT_pi * k2 / numMeshPtsPerDir[1]);

          a_k = a * (c_k1 + c_k2);

          for (d = 0; d < num_dim; d++) {
            SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_k[d][I][0]
                                                                      = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidForceDensity_k[d][I][0]
                                                                                                                                      / (a_k * normalizeFluidDensity_k);

            SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_k[d][I][1]
                                                                      = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidForceDensity_k[d][I][1]
                                                                                                                                      / (a_k * normalizeFluidDensity_k);
          }

        } else { /* set the zero mode to zero */

          for (d = 0; d < num_dim; d++) {
            SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_k[d][I][0] = 0.0;
            SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_k[d][I][1] = 0.0;
          }

        }

        if (SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagIncompressibleFluid) {
          printf("WARNING: %s : %s", error_str_code, error_str_func);
          printf(
              "Incompressible computation not implemented in current version of the codes. \n");

          printf("WARNING: %s : %s", error_str_code, error_str_func);
          printf(
              "NOTE: \"Stokes Paradox\" Occurs for Steady-State Incompressible \n");
          printf("Fluid Velocity Fields in 2D! \n");
        }

      } /* end k1 loop */
    } /* end k2 loop */

  } /* end num_dim == 2 */

  /* == Compute the steady-state fluid velocity */
  if (num_dim == 3) {
    /*
       printf("WARNING: %s : %s \n",error_str_code, error_str_func);
       printf("3D Steady-state sheared Stokes flow not debugged. \n");
       printf("Fluid domain is periodic with Lees-Edwards boundary conditions. \n");

       if (shearDir != 2) {
       printf("WARNING: %s : %s \n",error_str_code, error_str_func);
       printf("Only the shear direction z-axis is currently debugged. \n");
       }

       if (shearVelDir != 0) {
       printf("WARNING: %s : %s \n",error_str_code, error_str_func);
       printf("Only the shear velocity direction x-axis is currently debugged. \n");
       }
     */
    /* check number mesh points in each direction is odd. */
    /*
       flag = 0;
       for (j = 0; j < num_dim; j++) {
       if (numMeshPtsPerDir[j] % 2 == 0) {
       flag = 1;
       }
       }

       if (flag) {
       printf("ERROR: %s : %s \n", error_str_code, error_str_func);
       printf("Spectral representation of the differential operator \n");
       printf("requires an odd number of mesh points in each direction. \n");
       for (j = 0; j < num_dim; j++) {
       printf("numMeshPtsPerDir[%d] = %d \n", j, numMeshPtsPerDir[j]);
       }
       packageError(1,0);
       }
     */

    /* compute the steady-state Stokes flow: w = L^{-1}g */
    dot_gamma_t = shearDist/L_shearDir;
    vec_L[0] = numMeshPtsPerDir[0]*meshDeltaX; vec_L[1] = numMeshPtsPerDir[1]*meshDeltaX; vec_L[2] = numMeshPtsPerDir[2]*meshDeltaX;
    
    for (k3 = 0; k3 < numMeshPtsPerDir[2]; k3++) {
      for (k2 = 0; k2 < numMeshPtsPerDir[1]; k2++) {
        for (k1 = 0; k1 < numMeshPtsPerDir[0]; k1++) {

          I = (k3*numMeshPtsPerDir[1]*numMeshPtsPerDir[0]) + (k2*numMeshPtsPerDir[0]) + k1;

          vec_k[0] = k1; vec_k[1] = k2; vec_k[2] = k3;

          flagUseFiniteDifference = 1;
          if (flagUseFiniteDifference) {

            if ((k1 != 0) || (k2 != 0) || (k3 != 0)) {

              /*
                 Stokes flow subject to Lees-Edwards boundary
                 conditions (periodic images are sheared
                 relative to base image).
                 Finite difference approximation of the Laplacian
                 operator is used under the change of variable
                 which moves in the coordinate system of a
                 sheared box deforming with the flow.  This
                 coordinate system allows for the flow field
                 to be represented on a periodic domain with
                 a jump boundary condition.  The Laplacian
                 becomes under this change of variable for
                 shear along the z-axis (j = 2):

                 TeX:
                 $\mathbf{x} = \phi(\mathbf{q}) = (q_1 + \dot{\gamma}tq_3,q_2,q_3)$,
                 $\mathbf{q} = \phi^{-1}(\mathbf{x}) = (x_1 - \dot{\gamma}tx_3,x_2,x_3)$
                 and Laplacian can be expressed as
                 $$\mathcal{L}_{\mathbf{x}} \mathbf{u}^{(j)} 
                 = \sum_{\ell=1}^3 
                 \frac{\partial^2 \mathbf{u}^{(j)} }
                 {\partial \mathbf{x}^{(\ell),2}}
                 = \mathcal{L}_{\mathbf{q}}(t) \mathbf{w}^{(d)}
                 = \sum_{\ell=1}^3 \sum_{a,b = 1}^3 
                 (\delta_{\ell,a} - \dot{\gamma}t \delta_{\ell,3} \delta_{a,1})
                 \frac{\partial^2 \mathbf{w}^{(j)}}
                 {\partial \mathbf{q}^{(a)} \partial \mathbf{q}^{(b)}}
                 (\delta_{\ell,b} - \dot{\gamma}t \delta_{\ell,3} \delta_{b,1})$$

                 We shall use a finite difference approximation of the
                 partial derivatives.  The boundary conditions are
                 incorporated by adding the appropriate jump condition
                 to the values used in the finite difference stencil.
                 Mathematically, this can be taken into account by
                 introducing a source term in the steady-state Stokes
                 equations (move known values to RHS in the linear
                 equations) or even by a linear flow in quasi-steady 
                 regime.  The matrix is then cylcic and inverted
                 using the FFT.

               */
              L_shearDir = numMeshPtsPerDir[shearDir]*meshDeltaX;
              for (j = 0; j < num_dim; j++) {

                /* compute the fourier representation
                   of the finite difference operator */
                a_k_j[j] = 0.0;
                for (ell = 0; ell < num_dim; ell++) {
                  for (a = 0; a < num_dim; a++) {
                    for (b = 0; b < num_dim; b++) {

                      k_a = vec_k[a]; k_b = vec_k[b];

                      /* adjust the wavenumber
                         for use in the spectral
                         representation.  We use
                         that modes are equivalent
                         on the lattice with wavevector
                         modulo N.
                       */
                      N_a = numMeshPtsPerDir[a]; N_b = numMeshPtsPerDir[b];

                      //if (k_a > (N_a - 1) / 2) {
                      //  k_a = k_a - N_a;
                      //}

                      //if (k_b > (N_b - 1) / 2) {
                      //  k_b = k_b - N_b;
                      //}

                      //L_a = vec_L[a];
                      //L_b = vec_L[b];
                      
                      if (a == ell) {
                        delta_a_ell = 1;
                      } else {
                        delta_a_ell = 0;
                      }
                      
                     if (b == ell) {
                        delta_b_ell = 1;
                      } else {
                        delta_b_ell = 0;
                      }

                      if (a == shearVelDir) {
                        delta_a_velDir = 1;
                      } else {
                        delta_a_velDir = 0;
                      }

                      if (b == shearVelDir) {
                        delta_b_velDir = 1;
                      } else {
                        delta_b_velDir = 0;
                      }

                      if (ell == shearDir) {
                        delta_ell_shearDir = 1;
                      } else {
                        delta_ell_shearDir = 0;
                      }

                      if (a != b) { 
                        /* uses composition of central
                           differences in q_a and q_b directions
                           to achieve second order accuracy */                        
                        rep_d_2_w_j__d_q_a__d_q_b = -(sin(2*UNIT_pi*k_a/N_a)*sin(2*UNIT_pi*k_b/N_b))
                                                  /(meshDeltaX*meshDeltaX);

                      } else { /* uses second order central differences in q_a direction */
                        rep_d_2_w_j__d_q_a__d_q_b = -2.0*(1 - cos(2*UNIT_pi*k_a/N_a))/(meshDeltaX*meshDeltaX);
                      }

                      a_k_j[j] += mu*(delta_a_ell - dot_gamma_t*delta_ell_shearDir*delta_a_velDir)
                                    *(delta_b_ell - dot_gamma_t*delta_ell_shearDir*delta_b_velDir)
                                    *rep_d_2_w_j__d_q_a__d_q_b;

                    } /* end of b loop */
                  } /* end of a loop */

                } /* end of ell loop */

              } /* end of j loop */

              for (j = 0; j < num_dim; j++) {

                if (a_k_j[j] > 1e-9) {

                  stringstream message;

                  message << "a_k_j[" << j << "] = " << a_k_j[j] << " > 0 failed to hold." << endl;
                  message << "Rough check on the definiteness of the difference operator." << endl;
                  message << "need to be careful of round-off errors here." << endl;

                  SELM_Package::packageError(error_str_code, error_str_func, message);

                }

                SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_k[j][I][0]
                  = -SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidForceDensity_k[j][I][0]/(a_k_j[j]*normalizeFluidDensity_k);
                SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_k[j][I][1]
                  = -SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidForceDensity_k[j][I][1]/(a_k_j[j]*normalizeFluidDensity_k);
              }

            } else { /* set the zero mode to zero */

              for (d = 0; d < num_dim; d++) {
                SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_k[d][I][0] = 0.0;
                SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_k[d][I][1] = 0.0;
              }

            }

          } /* end of use finite difference approximation */

        } /* end k1 loop */
      } /* end k2 loop */
    } /* end k3 loop */

  } /* end num_dim == 3 */

  /* == Compute the effective stochastic forcing (time integrated average) for the fluid */
  if (SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagStochasticDriving) {

    /* -- Compute the the time averaged stochastic fluctuations of the fluid */
    /*
       printf("WARNING: %s : %s \n", error_str_code, error_str_func);
       printf("Stochastic forcing of the fluid in sheared conditions not yet debugged. \n");
     */
    computeTimeAvgStochFluct(SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->num_dim,
                             SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->meshDeltaX,
                             SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->numMeshPtsPerDir,
                             mu,KB,T,
                             SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->deltaT,
                             shearRate,shearDir,shearVelDir,shearDist,
                             /*IB_appl1Extras->gaussDevState,*/
                             fixSELM->random,
                             SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidStochForceDensity_k);

    /* -- Add the time average stochastic fluctuations to the fluid velocity field */
    for (k3 = 0; k3 < numMeshPtsPerDir[2]; k3++) {
      for (k2 = 0; k2 < numMeshPtsPerDir[1]; k2++) {
        for (k1 = 0; k1 < numMeshPtsPerDir[0]; k1++) {

          I = (k3*numMeshPtsPerDir[1]*numMeshPtsPerDir[0]) + (k2*numMeshPtsPerDir[0]) + k1;

          for (d = 0; d < num_dim; d++) {
            SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_k[d][I][0]
                                                                      += SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidStochForceDensity_k[d][I][0];
            SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_k[d][I][1]
                                                                      += SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidStochForceDensity_k[d][I][1];
          } /* end of d loop */

        } /* end k1 loop */
      } /* end k2 loop */
    } /* end k3 loop */

  } /* end of check for flagStochasticDriving */

  /* == For incompressible fluids apply projection operator */
  if (SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagIncompressibleFluid) {

     projectField(SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->num_dim,
                  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->meshDeltaX,
                  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->numMeshPtsPerDir,
                  shearDir,shearVelDir,shearDist,
                  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_k);

    if (SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->flagUseFluidPressure) {

      if (timeIndex == 1) {
        printf("WARNING: %s : %s \n", error_str_code, error_str_func);
        printf("Computing the explicit pressure field p(x) for the sheared fluid for output. \n");
        // The explicit pressure field below for visualization may not yet be debugged fully (double-check)
        // The simulation calculation uses instead projection operator, so does not get used there.
        // This is maninly for visualization purposes.
      }

      /* Compute pressure associated with the force density.  This only includes
         pressure for the 'deterministic part' of the force field. */
      computePressure(SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->num_dim,
                      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->meshDeltaX,
                      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->numMeshPtsPerDir,
                      shearDir,shearVelDir,shearDist,
                      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidForceDensity_k,
                      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidPressure_k);

    } /*end of flagComputePressureField */

  } /* end of flagIncompressibleFluid */

  /* == Compute the IDFT of the velocity field of the fluid */
  for (d = 0; d < num_dim; d++) {
    fftw_execute(SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_IDFT_plan[d]);
  }

  /* == Compute the IDFT of the pressure field of the fluid */
  if ((SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagIncompressibleFluid)
      && (SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->flagUseFluidPressure)) {
    fftw_execute(SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidPressure_IDFT_plan);
  }

  /* == Update the control point locations (if flag set) */
  if (SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagUpdateControlPts) {

    /* The average velocity is computed from the sheared four point region about the
       structure mapped to the sheared coordinate plane, which requires positions of
       the structures assumed to be described in physical space to be adjusted accordingly
       to get the corresponding structure positions in the sheared coordinate frame.
       The velocity field on the periodic mesh also needs to be handled carefully
       with periodic images being adjusted with the appropriate Lees-Edwards jumps
       when crossing a periodic boundary.
       This is now done in the subroutine IB_appl1_computeControlPtsVel_SHEAR_FFTW3.
     */
    /*
    computeControlPtsVel_SHEAR_FFTW3(SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->num_dim,
                                     SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->numMeshPtsPerDir,
                                     SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->meshDeltaX,
                                     SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->meshCenterX0,
                                     SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_m,
                                     SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE,
                                     shearRate,
                                     shearDir,
                                     shearVelDir,
                                     shearDist);
                                     */

    /* @@@ TODO: change this to looping over the coupling operators and allowing the coupling operators
     * to figure out how to determine the lagrangian velocities
     * */

//    /************************************************************************************************/
//    /* Old way of computing the lagrangian particle velocities */
//    /* The new way uses a loop over the coupling operators (see below) */
//    /************************************************************************************************/
//    /* loop over the Lagrangian DOF */
//    for (int lagrangianI = 0;lagrangianI < fixSELM->SELM_Lagrangian_List_N; lagrangianI++) {
//
//      if (fixSELM->SELM_Lagrangian_List[lagrangianI]->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {
//
//        SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE = (SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *)fixSELM->SELM_Lagrangian_List[lagrangianI];
//
//        /* TODO: determine coupling operator to use for this Lagrangian DOF @@@ */
//
//        computeControlPtsVel_SHEAR_FFTW3(SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE,
//                                         SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3,
//                                         fixSELM->SELM_CouplingOperator_List[0]); /* WARNING: assumes only one coupling operator */
//
//        for (k = 0; k < SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->numControlPts; k++) {
//          for (d = 0; d < num_dim; d++) {
//            SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->ptsX[(k * num_dim) + d]
//              += SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[(k* num_dim) + d] * deltaT;
//          }
//        }
//
//        /**********************************************************************************/

        /************************************************************************************************/
        /* Loop over the coupling operators to determine the Lagrangian velocities via Lambda operator.
        /************************************************************************************************/
        /* loop over the coupling operators DOF */
        for (int couplingOpI = 0;couplingOpI < fixSELM->SELM_CouplingOperator_List_N; couplingOpI++) {

          SELM_CouplingOperator *couplingOp = fixSELM->SELM_CouplingOperator_List[couplingOpI];

          /* maybe make below more generic */
          if (couplingOp->type == SELM_CouplingOperator_Types::TYPE_LAMMPS_SHEAR_UNIFORM1_FFTW3_TABLE1) {

            SELM_Lagrangian *lagrangian = NULL;
            SELM_Eulerian   *eulerian   = NULL;

            SELM_CouplingOperator_LAMMPS_SHEAR_UNIFORM1_FFTW3_TABLE1 *op
              = (SELM_CouplingOperator_LAMMPS_SHEAR_UNIFORM1_FFTW3_TABLE1 *)couplingOp;

            for (int k = 0; k < op->numCoupleList; k++) {
              lagrangian = op->lagrangianList[k];
              eulerian   = op->eulerianList[k];

              computeControlPtsVel_SHEAR_FFTW3(lagrangian,eulerian,couplingOp);
            }

            /* update the lagrangian DOF */
            if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {

              SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE
                = (SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *)lagrangian;

              for (k = 0; k < SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->numControlPts; k++) {
                for (d = 0; d < num_dim; d++) {
                  SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->ptsX[(k * num_dim) + d]
                    += SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[(k* num_dim) + d]*deltaT;
                }
              }

              /* Impose that the control point locations are always described within the
               * rectangular unit cell with Lees-Edwards boundary conditions.  Note the
               * Stokes flow is considered in the sheared unit cell.  We note that this is
               * also important since stress calculations must be handled carefully when
               * bonds cross a periodic boundary.
               */
              flagImposeUnitCellRep = 0;  /* @@@ TODO: Check if we need to impose the unit cell representation */
              if (flagImposeUnitCellRep == 1) {

                for (d = 0; d < num_dim; d++) {
                  periodL[d] = numMeshPtsPerDir[d] * meshDeltaX;
                }

                k = floor(fabs(shearDist / periodL[shearVelDir]));
                if (shearDist > 0) {
                  shearDistMod = shearDist - k*periodL[shearVelDir];
                } else {
                  shearDistMod = shearDist + k*periodL[shearVelDir];
                }

                for (k = 0; k < SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->numControlPts; k++) {
                  controlPt_X = (SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->ptsX + (k * num_dim));

                  bri1_unitCellRectImageShearPeriodic((double *) periodL,
                                                      (double *) meshCenterX0, shearDir,
                                                      shearVelDir, shearDistMod, num_dim,
                                                      controlPt_X, (double *) XX1);

                  for (d = 0; d < num_dim; d++) {
                    SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->ptsX[num_dim * k + d] = XX1[d];
                  }

                } /* end of k loop */

              } /* end flagImposeUnitCellRep */

              /* Adjustments may be made later to the structures to avoid excessive drifting in
               * stress calculations for linear shearing.
               */
               
            } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {

              SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE
                = (SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *)lagrangian;

              for (k = 0; k < SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->numControlPts; k++) {
                for (d = 0; d < num_dim; d++) {
                  SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->ptsX[(k * num_dim) + d]
                    += SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Vel[(k* num_dim) + d] * deltaT;
                }
              }

              /* Impose that the control point locations are always described within the
               * rectangular unit cell with Lees-Edwards boundary conditions.  Note the
               * Stokes flow is considered in the sheared unit cell.  We note that this is
               * also important since stress calculations must be handled carefully when
               * bonds cross a periodic boundary.
               */
              flagImposeUnitCellRep = 0;  /* @@@ TODO: Check if we need to impose the unit cell representation */
              if (flagImposeUnitCellRep == 1) {

                for (d = 0; d < num_dim; d++) {
                  periodL[d] = numMeshPtsPerDir[d] * meshDeltaX;
                }

                k = floor(fabs(shearDist / periodL[shearVelDir]));
                if (shearDist > 0) {
                  shearDistMod = shearDist - k*periodL[shearVelDir];
                } else {
                  shearDistMod = shearDist + k*periodL[shearVelDir];
                }

                for (k = 0; k < SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->numControlPts; k++) {
                  controlPt_X = (SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->ptsX + (k * num_dim));

                  bri1_unitCellRectImageShearPeriodic((double *) periodL,
                                                      (double *) meshCenterX0, shearDir,
                                                      shearVelDir, shearDistMod, num_dim,
                                                      controlPt_X, (double *) XX1);

                  for (d = 0; d < num_dim; d++) {
                    SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->ptsX[num_dim * k + d] = XX1[d];
                  }

                } /* end of k loop */

              } /* end flagImposeUnitCellRep */

              /* Adjustments may be made later to the structures to avoid excessive drifting in
               * stress calculations for linear shearing.
               */

            } else {
              stringstream message;
              message << "Lagrangian Type is not recognized" << endl;
              SELM_Package::packageError(error_str_code, error_str_func, message);
            }

          } else { /* end of TYPE_LAMMPS_SHEAR_UNIFORM1_FFTW3_TABLE1 */
            stringstream message;
            message << "Integrator does not know how to handle coupling operator type. " << endl;
            message << "  couplingOp->typeStr = " << couplingOp->typeStr << endl;
            SELM_Package::packageError(error_str_code, error_str_func, message);
          }

        } /* end of couplingOpI loop */
        
        /* synchronize the lammps data and control points data to one another */
        /* @optimize: only update the points that changed from above, for now we update all points */
        for (int I = 0; I < fixSELM->SELM_Lagrangian_List_N; I++) {
          fixSELM->SELM_Lagrangian_List[I]->setLammpsDataFromControlPtsData();
        }

     } /* end flagUpdateControlPts */

     /**********************************************************************************/

//        flagDebugViaControlPtsUpdate = 0; /* set this flag to one to get ad hoc Brownian dynamics for testing */
//        if (flagDebugViaControlPtsUpdate) {
//
//          if (timeIndex - 1  % 100 == 0) {
//            stringstream message;
//            message << "Debugging by overriding the control pts integrator to be ad hoc Brownian Dynamics" << endl;
//            message << "This code should be commented out or debug flag set to zero for production runs for SELM integrator!!!" << endl;
//            message << "If you receive this message your simulations are not running with proper SELM dynamics." << endl;
//            message << "Contact the developers immediately." << endl;
//            SELM_Package::packageWarning(error_str_code, error_str_func, message);
//          }
//
//          for (k = 0; k < SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->numControlPts; k++) {
//            for (d = 0; d < num_dim; d++) {
//              SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->ptsX[(k * num_dim) + d]
//              += 1e-6*SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Force[(k* num_dim) + d]*deltaT;
//              //+= SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[(k* num_dim) + d] * deltaT;
//            }
//          }
//
//        }

  /* == Update the shearing of the coordinate system for use in next time step */
  /* PJA: New way, we sync with LAMMPS and use the domain->xz*/
  shearDist_last = shearDist;
  L_shearDir = numMeshPtsPerDir[shearDir]*meshDeltaX;
  //shearDist += deltaT*shearRate*L_shearDir; /* this gives the total shift distance
  //                                           between the top and bottom shear planes */
  shearDist = lammps->domain->xz; // we use this for the shear displacement (directly from LAMMPS)  

  /* By symmetry of the system we can always choose the coordinate system for
     the propogator of the system which is sheared at most by one mesh spacing.
     The basic idea for the quasi-steady-state Stokes flow is to treat the
     fluid mesh always for a box which is sheared at most one period in the
     shear velocity direction.  In other words, for any arbitrary shear displacement
     the unit cell can always be mapped to one sheared by at most one period in
     the shear velocity direction by using the periodicity.

     WARNING: It may be important not to modify the shearDist variable as this could
     impact how the structure deformations are interpreted over the mesh.
  */
  flagShearMode = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagShearMode;
  switch (flagShearMode) {

  case SHEAR_MODE_TYPE_RM_SHEAR1:

    shearData_RM_SHEAR1
      = (ShearData_RM_SHEAR1_Type *)
    SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->shearData;

    flipToLessShearedUnitCell = 0;

    /* PJA: New way is to directly use the LAMMPS domain->xz value (commented out below)*/
    
    /* we only allow the shearDist to be at most the domain size, since
     * this is geometrically equivalent to tiling by a less sheared domain
     * that is more computationally convenient to use.  We may be able to further
     * improve the calculation by only allowing shears of half the domain size,
     * since this also yields geometrically equivalent tilings of space.
     */
    //L_shearVelDir = meshDeltaX*numMeshPtsPerDir[shearVelDir];
    //N_shear       = floor(fabs((shearDist - (-0.5*L_shearVelDir)) / L_shearVelDir));
    //if (N_shear != 0) {
    //  if (shearDist - (-0.5*L_shearVelDir) > 0) {
    //    shearDist = shearDist - N_shear * L_shearVelDir;
    //  } else {
    //    shearDist = shearDist + N_shear * L_shearVelDir;
    //  }
    //  flipToLessShearedUnitCell = 1; /* indicates domain was "flipped" to have less shear */
    //}

    /*
    if (N_shear != 0) {
      printf("WARNING: %s : %s \n", error_str_code, error_str_func);
      printf(
          "Adjusted the tracked shear distance since the shear deformation \n");
      printf(
          "of the unit cell now exceeds one unit in the shear velocity direction. \n");
      printf("timeIndex = %d \n", timeIndex);
      printf("old shearDist = %g \n", shearDist + N_shear * L_shearVelDir);
      printf("new shearDist = %g \n", shearDist);
    }
    */

    /* record the change in shearDist */
    shearData_RM_SHEAR1->shearDist_last = shearDist_last;
    shearData_RM_SHEAR1->shearDist      = shearDist;

    SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist_last
    = shearData_RM_SHEAR1->shearDist_last;

    SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist
    = shearData_RM_SHEAR1->shearDist;

    /*old way was to try to UPDATE THE LAAMPS DATA STRUCTURE and CALL UPDATES (commented out below) */
    /*   
    if (domain->triclinic) {
      if ((shearVelDir == 0) && (shearDir == 2)) {
        // this is now handled similar to fix_deform.cpp
      } else {
        stringstream message;
        message << "Use of Lees-Edwards-style shear of simulation domain is only implemented" << endl;
        message << "for the case when the shear velocity is in the x-direction" << endl;
        message << "and the shear varies in the z-direction." << endl;
        message << "Need to have shearVelDir = 0 and shearDir = 2. " << endl;
        message << "Simulation had:" << endl;
        message << "shearVelDir = " << shearVelDir << endl;
        message << "shearDir    = " << shearDir <<endl;
        SELM_Package::packageError(error_str_code, error_str_func, message);
      }
    }
    */

    break;

  case SHEAR_MODE_TYPE_RM_OSC1:

    shearData_RM_OSC1
      = (ShearData_RM_OSC1_Type *)
      SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->shearData;

    /* PJA: New way is to directly use the LAMMPS domain value for the shearDist */

    /* need to change this to only allow 1/2 the total domain length */
    /*
    L_shearVelDir = meshDeltaX * numMeshPtsPerDir[shearVelDir];
    N_shear = floor(fabs(shearDist / L_shearVelDir));
    if (N_shear != 0) {
      printf("WARNING: %s : %s \n", error_str_code, error_str_func);
      printf("The shear deformation exceeds one cell period. \n");
      printf("This behavior was not anticipated for the oscillitory \n");
      printf("deformations of the unit cell and may yield poor accuracy.\n");
      printf(
          "in the discretization of the Laplacian (overly distorted mesh). \n");
      printf(
          "If this issue is not to be a problem for the mesh resolution and deformation invoked \n");
      printf("this warning message can be removed. \n");
      printf("timeIndex = %d \n", timeIndex);
      printf("old shearDist = %g \n", shearDist);
      //packageError(1,0); 
    }

    if (shearDist > 0) {
      shearDist = shearDist - N_shear*L_shearVelDir;
    } else {
      shearDist = shearDist + N_shear*L_shearVelDir;
    }

    if (N_shear != 0) {
      printf("new shearDist = %g \n", shearDist);
    }
    */

    /* record the change in shearDist */
    shearData_RM_OSC1->shearDist_last = shearDist_last;
    shearData_RM_OSC1->shearDist      = shearDist;

    SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist_last
      = shearData_RM_OSC1->shearDist_last;

    SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist
      = shearData_RM_OSC1->shearDist;

    /*UPDATE THE LAAMPS DATA STRUCTURE and CALL UPDATES (see prior case above) */
    
    /* new way is to directly use LAMMPS values for the domain */

    break;

  } /* end of switch */

#else

  stringstream message;
  message << "Compiler flag USE_PACKAGE_FFTW3 was not set." << endl;
  message << "None of the FFTW3 routines were compiled in the codes." << endl;
  SELM_Package::packageError(error_str_code, error_str_func, message);

#endif

  /* Compute the force density arising from the control points for FFTW3 uniform mesh. */
  /*
    void SELM_applyControlPtsForceToMesh_FFTW3(int                 num_dim,
                                                   int                 numMeshPtsPerDir[3],
                                                   double              meshDeltaX,
                                                   double             *meshCenterX0,
                                                   fftw_complex       *f_m[3],
                                                   controlPtsDataType *SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE,
                                                   int                 flagPeriodicCondOnPts);
   */

}


void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::SELM_updateFluidAndStructures_final()
{

  const char *error_str_func = "SELM_updateFluidAndStructures_final()";

}


/*====================================== Private ===================================== */
void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::addPseudoForceTerm(int num_dim,
                                                                          double meshDeltaX,
                                                                          int *numMeshPtsPerDir,
                                                                          double mu,
                                                                          int shearDir,
                                                                          int shearVelDir,
                                                                          double shearRate,
                                                                          fftw_complex **fluidForceDensity_m) {

  const char *error_str_func = "addPseudoForceTerm()";

  int d;
  int a, b;
  int i, j, k;
  int j1, j2, j3;
  int k1, k2, k3;
  int m1, m2, m3;

  int vec_m_top[3];
  int vec_m_bot[3];

  int I_bot;
  int I_top;

  double L_shearDir;

  double contr_top_d_2_e_d_x_a_d_x_b;
  double contr_bot_d_2_e_d_x_a_d_x_b;

  double normalizeFluidDensity_k;

  if (num_dim == 3) {

    /* compute the contribution from the finite difference stencil along the top and
       bottom shear plane of the domain
     */
    for (j1 = 0; j1 < numMeshPtsPerDir[(shearDir + 1) % num_dim]; j1++) {
      for (j2 = 0; j2 < numMeshPtsPerDir[(shearDir + 2) % num_dim]; j2++) {

        vec_m_top[(shearDir + 1) % num_dim] = j1;
        vec_m_top[(shearDir + 2) % num_dim] = j2;
        vec_m_top[(shearDir + 0) % num_dim] = (numMeshPtsPerDir[shearDir] - 1);

        vec_m_bot[(shearDir + 1) % num_dim] = j1;
        vec_m_bot[(shearDir + 2) % num_dim] = j2;
        vec_m_bot[(shearDir + 0) % num_dim] = 0;

        m1 = vec_m_top[0];
        m2 = vec_m_top[1];
        m3 = vec_m_top[2];
        I_top = (m3 * numMeshPtsPerDir[1] * numMeshPtsPerDir[0]) + (m2
            * numMeshPtsPerDir[0]) + m1;

        m1 = vec_m_bot[0];
        m2 = vec_m_bot[1];
        m3 = vec_m_bot[2];
        I_bot = (m3 * numMeshPtsPerDir[1] * numMeshPtsPerDir[0]) + (m2
            * numMeshPtsPerDir[0]) + m1;

        L_shearDir = numMeshPtsPerDir[shearDir] * meshDeltaX;

        contr_top_d_2_e_d_x_a_d_x_b = 0.0;
        contr_bot_d_2_e_d_x_a_d_x_b = 0.0;
        for (a = 0; a < num_dim; a++) {
          for (b = 0; b < num_dim; b++) {

            if (a != b) {
              contr_top_d_2_e_d_x_a_d_x_b += 0.0;
              contr_bot_d_2_e_d_x_a_d_x_b += 0.0;
            } else { /* a == b */

              /* only contributes in the shear direction */
              if (a == shearDir) {
                contr_top_d_2_e_d_x_a_d_x_b += mu * shearRate * L_shearDir
                    / (meshDeltaX * meshDeltaX);
                contr_bot_d_2_e_d_x_a_d_x_b += -mu * shearRate * L_shearDir
                    / (meshDeltaX * meshDeltaX);
              }

            } /* end else a == b */

          } /* end of b loop */
        } /* end of a loop */

        fluidForceDensity_m[shearVelDir][I_top][0] += contr_top_d_2_e_d_x_a_d_x_b;
        fluidForceDensity_m[shearVelDir][I_top][1] += 0.0;
        fluidForceDensity_m[shearVelDir][I_bot][0] += contr_bot_d_2_e_d_x_a_d_x_b;
        fluidForceDensity_m[shearVelDir][I_bot][1] += 0.0;

      } /* end of j2 loop */
    } /* end of j1 loop */

  } else {/* end num_dim == 3 */

    stringstream message;

    message << "num_dim = " << num_dim << endl;
    message << "Pseudo force computation at shear boundary not implemented in this case." << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);

  }

}

void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::computeTimeAvgStochFluct(int num_dim,
                                                                         double meshDeltaX,
                                                                         int *numMeshPtsPerDir,
                                                                         double mu,
                                                                         double KB,
                                                                         double T,
                                                                         double deltaT,
                                                                         double shearRate,
                                                                         int shearDir,
                                                                         int shearVelDir,
                                                                         double shearDist,
                                                                         /* gaussDevStateType *gaussDevState, */
                                                                         class RanMars *random,
                                                                         fftw_complex **fluidStochForceDensity_k) {

  const char *error_str_func = "computeTimeAvgStochFluct()";

  int d;
  int i, j, k, ell;
  int I;
  int I1, I2;
  int k1, k2, k3;
  int kk1, kk2, kk3;

  double G_real_kk;
  double G_imag_kk;

  double sqrt_G_real_kk;
  double sqrt_G_imag_kk;

  double eta_real;
  double eta_imag;

  /* double normalizeFluidDensity_k; */

  double meshVolume;
  double cellVolume;

  int a, b;

  int N_a, N_b;

  double shearFlow_k_j[2][3];

  double vec_L[3];
  int vec_k[3];

  int k_a, k_b;

  double L_a, L_b;
  double rep_d_2_w_j__d_q_a__d_q_b;
  double a_k_j[3];

  int delta_a_velDir;
  int delta_b_velDir;

  int delta_a_ell;
  int delta_b_ell;

  int delta_ell_shearDir;

  int flag;

  double L_shearDir;
  
  double dot_gamma_t;

  int N_shear;
  int k_shear;

  int flagUseFiniteDifference;
  int flagUseSpectralRep;

  int prodNumMeshPts;

  /* == Compute the DFT of the force density of the fluid */
  /* normalizeFluidDensity_k = 1;
     for (d = 0; d < num_dim; d++) {
     normalizeFluidDensity_k
     = normalizeFluidDensity_k*numMeshPtsPerDir[d];
     }
   */

  if (num_dim == 3) {

    prodNumMeshPts = 1;
    for (d = 0; d < num_dim; d++) {
      prodNumMeshPts = prodNumMeshPts * numMeshPtsPerDir[d];
    }

    meshVolume = 1.0;
    for (d = 0; d < num_dim; d++) {
      meshVolume = meshVolume * numMeshPtsPerDir[d] * meshDeltaX;
    }

    cellVolume = 1.0;
    for (d = 0; d < num_dim; d++) {
      cellVolume = cellVolume * meshDeltaX;
    }

    L_shearDir = numMeshPtsPerDir[shearDir] * meshDeltaX;
    dot_gamma_t = shearDist/L_shearDir;
    vec_L[0] = numMeshPtsPerDir[0]*meshDeltaX; vec_L[1] = numMeshPtsPerDir[1]*meshDeltaX; vec_L[2] = numMeshPtsPerDir[2]*meshDeltaX;

    for (k3 = 0; k3 < numMeshPtsPerDir[2]; k3++) {
      for (k2 = 0; k2 < numMeshPtsPerDir[1]; k2++) {
        for (k1 = 0; k1 < numMeshPtsPerDir[0]; k1++) {

          I = (k3 * numMeshPtsPerDir[1] * numMeshPtsPerDir[0]) + (k2
              * numMeshPtsPerDir[0]) + k1;

          kk1 = (numMeshPtsPerDir[0] - k1) % numMeshPtsPerDir[0];
          kk2 = (numMeshPtsPerDir[1] - k2) % numMeshPtsPerDir[1];
          kk3 = (numMeshPtsPerDir[2] - k3) % numMeshPtsPerDir[2];

          I2 = (kk3 * numMeshPtsPerDir[1] * numMeshPtsPerDir[0]) + (kk2
              * numMeshPtsPerDir[0]) + kk1;

          /* added the vec_k and vec_L data here : these were missing! @PJA: Dec. 17th, 2008 */
          vec_k[0] = k1; vec_k[1] = k2; vec_k[2] = k3;

          //vec_L[0] = numMeshPtsPerDir[0] * meshDeltaX;
          //vec_L[1] = numMeshPtsPerDir[1] * meshDeltaX;
          //vec_L[2] = numMeshPtsPerDir[2] * meshDeltaX;

          /* Compute the effective fluid fluctuations to construct
             appropriate thermal forcing of the particle equations
             as prescribed by the principle of detailed balance.

             Hydrodynamic coupling:

             H = R_a \wp L^{-1} R_a^T/ meshDeltaX^d ,

             where R_a is delta function averaging procedure.
             \wp is the projection operator.

             Detailed balance requires:

             QQ' = 2*KB*T*H.

             This requires for the fluid generate random variable
             with covariance:

             Q = sqrt(2*KB*T/meshDeltaX^d)*R_a \wp L^{-1/2}.

             where we used that \wp L^{-1} = \wp L^{-1} \wp^T.

             WARNING: (need to double check this is the
             correct distribution for uniform mesh operator !!!)

             WARNING: Need to set the conjugate modes to be equal
             and self-conjugate to be real.

           */
          if ((k1 != 0) || (k2 != 0) || (k3 != 0)) {
            /*
                 Stokes flow subject to Lees-Edwards boundary
                 conditions (periodic images are sheared
                 relative to base image).
                 Finite difference approximation of the Laplacian
                 operator is used under the change of variable
                 which moves in the coordinate system of a
                 sheared box deforming with the flow.  This
                 coordinate system allows for the flow field
                 to be represented on a periodic domain with
                 a jump boundary condition.  The Laplacian
                 becomes under this change of variable for
                 shear along the z-axis (j = 2):

                 TeX:
                 $\mathbf{x} = \phi(\mathbf{q}) = (q_1 + \dot{\gamma}tq_3,q_2,q_3)$,
                 $\mathbf{q} = \phi^{-1}(\mathbf{x}) = (x_1 - \dot{\gamma}tx_3,x_2,x_3)$
                 and Laplacian can be expressed as
                 $$\mathcal{L}_{\mathbf{x}} \mathbf{u}^{(j)} 
                 = \sum_{\ell=1}^3 
                 \frac{\partial^2 \mathbf{u}^{(j)} }
                 {\partial \mathbf{x}^{(\ell),2}}
                 = \mathcal{L}_{\mathbf{q}}(t) \mathbf{w}^{(d)}
                 = \sum_{\ell=1}^3 \sum_{a,b = 1}^3 
                 (\delta_{\ell,a} - \dot{\gamma}t \delta_{\ell,3} \delta_{a,1})
                 \frac{\partial^2 \mathbf{w}^{(j)}}
                 {\partial \mathbf{q}^{(a)} \partial \mathbf{q}^{(b)}}
                 (\delta_{\ell,b} - \dot{\gamma}t \delta_{\ell,3} \delta_{b,1})$$

                 We shall use a finite difference approximation of the
                 partial derivatives.  The boundary conditions are
                 incorporated by adding the appropriate jump condition
                 to the values used in the finite difference stencil.
                 Mathematically, this can be taken into account by
                 introducing a source term in the steady-state Stokes
                 equations (move known values to RHS in the linear
                 equations) or even by a linear flow in quasi-steady 
                 regime.  The matrix is then cylcic and inverted
                 using the FFT.
            */
            L_shearDir = numMeshPtsPerDir[shearDir] * meshDeltaX;
            for (j = 0; j < num_dim; j++) {

              /* compute the fourier representation
              of the finite difference operator */
              a_k_j[j] = 0.0;
              for (ell = 0; ell < num_dim; ell++) {
              
                for (a = 0; a < num_dim; a++) {
                  for (b = 0; b < num_dim; b++) {

                    k_a = vec_k[a]; k_b = vec_k[b];

                    /* adjust the wavenumber
                       for use in the spectral
                       representation.  We use
                       that modes are equivalent
                       on the lattice with wavevector
                       modulo N.
                     */
                    //N_a = numMeshPtsPerDir[a];
                    //N_b = numMeshPtsPerDir[b];
                    N_a = numMeshPtsPerDir[a]; N_b = numMeshPtsPerDir[b];

                    //if (k_a > (N_a - 1) / 2) {
                    //  k_a = k_a - N_a;
                    //}

                    //if (k_b > (N_b - 1) / 2) {
                    //  k_b = k_b - N_b;
                    //}

                    //L_a = vec_L[a];
                    //L_b = vec_L[b];
                    
                    if (a == ell) {
                      delta_a_ell = 1;
                    } else {
                      delta_a_ell = 0;
                    }
                    
                    if (b == ell) {
                      delta_b_ell = 1;
                    } else {
                      delta_b_ell = 0;
                    }

                    if (a == shearVelDir) {
                      delta_a_velDir = 1;
                    } else {
                      delta_a_velDir = 0;
                    }

                    if (b == shearVelDir) {
                      delta_b_velDir = 1;
                    } else {
                      delta_b_velDir = 0;
                    }

                    if (ell == shearDir) {
                      delta_ell_shearDir = 1;
                    } else {
                      delta_ell_shearDir = 0;
                    }

                    if (a != b) { /* uses composition of central
                                     differences in x_a and x_b directions
                                     to achieve second order accuracy */
                      rep_d_2_w_j__d_q_a__d_q_b = -(sin(2*UNIT_pi*k_a/N_a)*sin(2*UNIT_pi*k_b/N_b))
                                                  /(meshDeltaX*meshDeltaX);

                    } else { /* uses second order central differences in x_a direction */
                      rep_d_2_w_j__d_q_a__d_q_b = -2.0*(1 - cos(2*UNIT_pi*k_a/N_a))/(meshDeltaX*meshDeltaX);
                    }

                    a_k_j[j] += mu*(delta_a_ell - dot_gamma_t*delta_ell_shearDir*delta_a_velDir)
                                  *(delta_b_ell - dot_gamma_t*delta_ell_shearDir*delta_b_velDir)
                                  *rep_d_2_w_j__d_q_a__d_q_b;

                  } /* end of b loop */
                } /* end of a loop */
                
              } /* end of ell loop */

            } /* end of j loop */

            /*
               if ((vec_k[0] == 1) && (vec_k[1] == 1) && (vec_k[2] == 1)) {
               printf("WARNING: %s : %s \n",error_str_code, error_str_func);
               printf("Set the flow explicitly to the shear flow. \n");
               }

               for (j = 0; j < num_dim; j++) {
               SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_k[j][I][0]
               = shearFlow_k_j[0][j]/normalizeFluidDensity_k;
               SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_k[j][I][1]
               = shearFlow_k_j[1][j]/normalizeFluidDensity_k;
               }
             */

            for (j = 0; j < num_dim; j++) {

              if (a_k_j[j] > 1e-9) {
                stringstream message;

                message << "a_k_j[" << j << "] = " << a_k_j[j] << " > 0 failed to hold." << endl;
                message << "Rough check on the definiteness of the difference operator." << endl;
                message << "need to be careful of round-off errors here." << endl;
                SELM_Package::packageError(error_str_code, error_str_func, message);

              }

            } /* end of the j loop */

            /*
               a         = 2.0*mu/(meshDeltaX*meshDeltaX);
               c_k1      = 1 - cos(2*UNIT_pi*k1/numMeshPtsPerDir[0]);
               c_k2      = 1 - cos(2*UNIT_pi*k2/numMeshPtsPerDir[1]);
               c_k3      = 1 - cos(2*UNIT_pi*k3/numMeshPtsPerDir[2]);

               a_k       = a*(c_k1 + c_k2 + c_k3);
             */

            /* (1.0/a_k) is -L^{-1} *//* WARNING: check this carefully,
               since it has wavenumber dependence */

            for (j = 0; j < num_dim; j++) {

              /* Note that since the discrete Fourier transform we use is not unitary
               an extra factor related to the number of lattice sites in each direction
               arises to obtain the correct covariance structure in
               physical space.  This can be readily deduced by consider the
               covariance structure of a random field having Kronecker delta correlation
               in space represented by applying our discrete Fourier transform.

               Self-conjugate modes give different results than the modes with distinct
               conjugates.

               */

              if (I == I2) { /* indicates mode is self-conjugate */

                /* variance for mode (components) */

                /* sigma_real^2 = deltaT*(2*KB*T/dX^d)*(L^{-1})/(N^d)
               sigma_imag^2 = 0
                 */
                G_real_kk = deltaT * (2.0 * KB * T / cellVolume) * (1.0
                    / (-a_k_j[j] * prodNumMeshPts));
                G_imag_kk = 0.0;

              } else { /* mode is not self-conjugate */

                /* variance for mode (components) */

                /* sigma_real^2 = deltaT*(2*KB*T/dX^d)*(L^{-1})/(2N^d)
               sigma_imag^2 = deltaT*(2*KB*T/dX^d)*(L^{-1})/(2N^d)
                 */
                G_real_kk = deltaT * (2.0 * KB * T / cellVolume) * (1.0
                    / (-a_k_j[j] * 2.0 * prodNumMeshPts));
                G_imag_kk = deltaT * (2.0 * KB * T / cellVolume) * (1.0
                    / (-a_k_j[j] * 2.0 * prodNumMeshPts));

              }

              sqrt_G_real_kk = sqrt(G_real_kk);
              sqrt_G_imag_kk = sqrt(G_imag_kk);

              eta_real = (double) fixSELM->random->gaussian(); /*gaussdev(gaussDevState); */
              eta_imag = (double) fixSELM->random->gaussian(); /*gaussdev(gaussDevState); */

              fluidStochForceDensity_k[j][I][0] = sqrt_G_real_kk * eta_real
                  / deltaT;

              fluidStochForceDensity_k[j][I][1] = sqrt_G_imag_kk * eta_imag
                  / deltaT;

            } /* end of j loop */

          } else { /* zero mode set to zero */

            for (d = 0; d < num_dim; d++) {
              fluidStochForceDensity_k[d][I][0] = 0.0;

              fluidStochForceDensity_k[d][I][1] = 0.0;
            }

          } /* end check for zero mode */

        } /* end k1 loop */
      } /* end k2 loop */
    } /* end k3 loop */

    /* -- Set the conjugate modes to be equal.
       Sets each mode equal to its conjugate mode
       this enforces the constraint that the velocity
       field is real valued.

       (@optimize) Algorithm is redundant below, sets each
       mode twice.

       (@optimize) The algorithm should be unnecessary
       in the case thermal effects are not used since
       the force field modes come from a real field.

     */
    for (k3 = 0; k3 < numMeshPtsPerDir[2]; k3++) {
      for (k2 = 0; k2 < numMeshPtsPerDir[1]; k2++) {
        for (k1 = 0; k1 < numMeshPtsPerDir[0]; k1++) {

          I1 = (k3 * numMeshPtsPerDir[1] * numMeshPtsPerDir[0]) + (k2
              * numMeshPtsPerDir[0]) + k1;

          kk1 = (numMeshPtsPerDir[0] - k1) % numMeshPtsPerDir[0];
          kk2 = (numMeshPtsPerDir[1] - k2) % numMeshPtsPerDir[1];
          kk3 = (numMeshPtsPerDir[2] - k3) % numMeshPtsPerDir[2];

          I2 = (kk3 * numMeshPtsPerDir[1] * numMeshPtsPerDir[0]) + (kk2
              * numMeshPtsPerDir[0]) + kk1;

          for (d = 0; d < num_dim; d++) {

            /* set the real parts equal */
            fluidStochForceDensity_k[d][I1][0]
                                            = fluidStochForceDensity_k[d][I2][0];

            /* set the imaginary parts (conj to each other) */
            /* if the condition requires u_k to be conj to
               itself then set the imaginary part is zero,
               otherwise set the imaginary part conj. */
            if (I1 == I2) { /* self-conjugate mode */

              /* imaginary part set to zero */
              fluidStochForceDensity_k[d][I1][1] = 0;

            } else { /* non-self-conjugate mode */

              /* imaginary part set to conjugate */
              fluidStochForceDensity_k[d][I1][1]
                                              = -fluidStochForceDensity_k[d][I2][1];

            }

          } /* of d loop */

        } /* end k1 loop */
      } /* end k2 loop */
    } /* end k3 loop */

  /* end num_dim  == 3 */
  } else {
    printf("WARNING: %s : %s", error_str_code, error_str_func);
    printf("num_dim = %d \n", num_dim);
    printf("Stochastic field generation computation not implemented in this case yet. \n");
  }


}


void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::projectField(int num_dim, double meshDeltaX,
                                                             int *numMeshPtsPerDir,
                                                             int shearDir, int shearVelDir,
                                                             double shearDist,
                                                             fftw_complex **field_u_k) {

  const char *error_str_func = "projectField()";

  int d;
  int i, j, k;
  int I;
  int k1, k2, k3;
  double a;
  double a_k;
  double c_k1, c_k2, c_k3;
  double deltaT;
  double g_k_real[3];
  double g_k_imag[3];

  double g_k_sq;

  double q_k_real[3];
  double q_k_imag[3];

  double q_k_dot_u_real;
  double q_k_dot_u_imag;

  double q_k_dot_f_real;
  double q_k_dot_f_imag;

  double q_k_sq;

  double vec_proj_real[3];
  double vec_proj_imag[3];

  double normalizeFluidDensity_k;

  double L_shearDir;
  double dot_gamma_t;
  
  double conj_q_k_T_dot_w_k__real, conj_q_k_T_dot_w_k__imag; 
  double a_real, a_imag;

  if (num_dim == 3) {
    /*
     if (shearDir != 2) {
       printf("WARNING: %s : %s \n",error_str_code, error_str_func);
       printf("Only shearDir == 2 is currently debugged. \n");
     }
     */
    L_shearDir = numMeshPtsPerDir[shearDir]*meshDeltaX;
    dot_gamma_t = shearDist/L_shearDir;

    for (k3 = 0; k3 < numMeshPtsPerDir[2]; k3++) {
      for (k2 = 0; k2 < numMeshPtsPerDir[1]; k2++) {
        for (k1 = 0; k1 < numMeshPtsPerDir[0]; k1++) {

          I = (k3*numMeshPtsPerDir[1]*numMeshPtsPerDir[0]) 
            + (k2*numMeshPtsPerDir[0]) + k1;

          if ((k1 != 0) || (k2 != 0) || (k3 != 0)) {
            /* modify the projection method */

            /* note that g_k_real = 0 */
            g_k_imag[0] = sin(2.0*UNIT_pi*k1/numMeshPtsPerDir[0])/(meshDeltaX);
            g_k_imag[1] = sin(2.0*UNIT_pi*k2/numMeshPtsPerDir[1])/(meshDeltaX);
            g_k_imag[2] = sin(2.0*UNIT_pi*k3/numMeshPtsPerDir[2])/(meshDeltaX);

            g_k_sq = (g_k_imag[0]*g_k_imag[0]) + (g_k_imag[1]*g_k_imag[1]) + (g_k_imag[2]*g_k_imag[2]);

            /* construct the shear coordinate diverence operator D(t) Fourier representation 
               as D[w_k] = conj(q_k)^T*w_k.  
               
               This will allow us a natural interpretation in terms of projection, since 
               conj(q_k)^Tw_k = <q_k,w_k> dot-product for complex vectors.
               
               Note: the sign for q_k = -g_k, since g_k is purely imaginary and we want 
               divergence representated as a complex-valued dot-product.
            
            */
            q_k_imag[0] = -g_k_imag[0]; q_k_imag[1] = -g_k_imag[1]; q_k_imag[2] = -g_k_imag[2];

            q_k_imag[shearDir] += dot_gamma_t*g_k_imag[shearVelDir];  // note sign + from the conjuage

            q_k_sq = (q_k_imag[0]*q_k_imag[0]) + (q_k_imag[1]*q_k_imag[1]) + (q_k_imag[2]*q_k_imag[2]);

            /* use projection to enforce the incompressibility constraint on u_det_k_np1.

               Note that projection requires we compute

               v_k = (I - (q_k*conj(q_k)^T/\|q_k\|^2))*w_k. 

               This removes the compressible model q_k.
               
               Let q_k = \alpha + i*\beta = i*\beta, since \alpha = 0.
               w_k = \gamma + i*\lambda
               
               conj(q_k)^T*w_k = -i*\beta^T*\gamma + \beta^T*\lambda.
               
               Then use complex multiplication to combine into the projection
               expression to obtain the final v_k.
               
               Note:
               \alpha = 0, \beta = q_k_imag[]; \gamma = field_u_k[d][I][0], \lambda = field_u_k[d][I][1].

            */
            conj_q_k_T_dot_w_k__real = 0.0; conj_q_k_T_dot_w_k__imag = 0.0;            
            for (d = 0; d < num_dim; d++) {
              conj_q_k_T_dot_w_k__real += q_k_imag[d]*field_u_k[d][I][1];
              conj_q_k_T_dot_w_k__imag += -q_k_imag[d]*field_u_k[d][I][0];
            }
            
            // a = conj(q_k)^T*w_k/\|q_k\|^2.
            a_real = conj_q_k_T_dot_w_k__real/q_k_sq; 
            a_imag = conj_q_k_T_dot_w_k__imag/q_k_sq; 
                                                                                      
            for (d = 0; d < num_dim; d++) {
              vec_proj_real[d] = -q_k_imag[d]*a_imag; // q_k*conj(q_k)^T*w_k/\|q_k\|^2.
              field_u_k[d][I][0] -= vec_proj_real[d]; // (I - q_k*conj(q_k)^T*w_k/\|q_k\|^2)*w_k

              vec_proj_imag[d] = q_k_imag[d]*a_real;  // q_k*conj(q_k)^T*w_k/\|q_k\|^2.
              field_u_k[d][I][1] -= vec_proj_imag[d]; // (I - q_k*conj(q_k)^T*w_k/\|q_k\|^2)*w_k

            } /* end of d loop */

          } /* end of check not zero mode */

        } /* end k1 loop */
      } /* end k2 loop */
    } /* end k3 loop */

    /* end num_dim  == 3 */
  } else {
    printf("WARNING: %s : %s", error_str_code, error_str_func);
    printf("num_dim = %d \n", num_dim);
    printf(
        "Incompressible computation not implemented in current version of the codes. \n");

    printf("WARNING: %s : %s", error_str_code, error_str_func);
    printf("NOTE: \"Stokes Paradox\" Occurs for Steady-State Incompressible \n");
    printf("Fluid Velocity Fields in 2D! \n");
  }

}


/* Compute the Fourier coefficients of the pressure field.
 This is obtain from:

 p_k = -(g_k^T/(|g_k|^2))*f_k

 NOTE: The pressure reported projects the force density,
 which corresponds to the pressure in the hydrodynamic
 equations.  This is NOT the same as the "pressure" obtained
 from the Hodge decomposition and imposition of incompressibility
 for the fluid velocity field, which is applied to thr steady-state
 solution.  For the uniform mesh case we may perform the projection
 at either stage to enforce incompressibility by the fact that the
 projection operator and discrete Laplacian commute.

 */
void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::computePressure(int num_dim,
                                                                double meshDeltaX,
                                                                int *numMeshPtsPerDir,
                                                                int shearDir,
                                                                int shearVelDir,
                                                                double shearDist,
                                                                fftw_complex **forceDensity_k,
                                                                fftw_complex *pressure_k) {

  const char *error_str_func = "computePressure()";

  int d;
  int i, j, k;
  int I;
  int k1, k2, k3;
  double a;
  double a_k;
  double c_k1, c_k2, c_k3;

  double g_k_real[3];
  double g_k_imag[3];

  double g_k_dot_u_real;
  double g_k_dot_u_imag;

  double g_k_dot_f_real;
  double g_k_dot_f_imag;

  double g_k_sq;

  double g_k_dot_u_real_over_g_k_sq;
  double g_k_dot_u_imag_over_g_k_sq;

  double g_k_dot_f_real_over_g_k_sq;
  double g_k_dot_f_imag_over_g_k_sq;

  double q_k_real[3];
  double q_k_imag[3];

  double q_k_dot_u_real;
  double q_k_dot_u_imag;

  double q_k_dot_f_real;
  double q_k_dot_f_imag;

  double q_k_sq;

  double q_k_dot_u_real_over_q_k_sq;
  double q_k_dot_u_imag_over_q_k_sq;

  double q_k_dot_f_real_over_q_k_sq;
  double q_k_dot_f_imag_over_q_k_sq;

  double vec_proj_real[3];
  double vec_proj_imag[3];

  double normalizeFluidDensity_k;

  double L_shearDir;

  normalizeFluidDensity_k = 1;
  for (d = 0; d < num_dim; d++) {
    normalizeFluidDensity_k = normalizeFluidDensity_k * numMeshPtsPerDir[d];
  }

  if (num_dim == 3) {
    /*
if (shearDir != 2) {
printf("WARNING: %s : %s \n",error_str_code, error_str_func);
printf("Only shearDir == 2 is currently debugged. \n");
}
     */
    L_shearDir = numMeshPtsPerDir[shearDir] * meshDeltaX;

    for (k3 = 0; k3 < numMeshPtsPerDir[2]; k3++) {
      for (k2 = 0; k2 < numMeshPtsPerDir[1]; k2++) {
        for (k1 = 0; k1 < numMeshPtsPerDir[0]; k1++) {

          I = (k3 * numMeshPtsPerDir[1] * numMeshPtsPerDir[0]) + (k2
              * numMeshPtsPerDir[0]) + k1;

          if ((k1 != 0) || (k2 != 0) || (k3 != 0)) {

            /* note that
g_k_real = 0 */
            g_k_imag[0] = sin(2.0 * UNIT_pi * k1 / numMeshPtsPerDir[0])
    / meshDeltaX;
            g_k_imag[1] = sin(2.0 * UNIT_pi * k2 / numMeshPtsPerDir[1])
    / meshDeltaX;
            g_k_imag[2] = sin(2.0 * UNIT_pi * k3 / numMeshPtsPerDir[2])
    / meshDeltaX;

            g_k_sq = (g_k_imag[0] * g_k_imag[0]) + (g_k_imag[1] * g_k_imag[1])
    + (g_k_imag[2] * g_k_imag[2]);

            /* construct the shear coordinate divergence
               operator D(t) Fourier representation */
            q_k_imag[0] = g_k_imag[0];
            q_k_imag[1] = g_k_imag[1];
            q_k_imag[2] = g_k_imag[2];

            q_k_imag[shearDir] = g_k_imag[shearDir] - (shearDist / L_shearDir)
    * g_k_imag[shearVelDir];

            q_k_sq = (q_k_imag[0] * q_k_imag[0]) + (q_k_imag[1] * q_k_imag[1])
    + (q_k_imag[2] * q_k_imag[2]);

            q_k_dot_f_real = 0; /* real part of g^T f */
            for (d = 0; d < num_dim; d++) {
              q_k_dot_f_real += -q_k_imag[d] * forceDensity_k[d][I][1]
                                                                    / normalizeFluidDensity_k;
            }
            q_k_dot_f_real_over_q_k_sq = q_k_dot_f_real / q_k_sq;

            q_k_dot_f_imag = 0;
            for (d = 0; d < num_dim; d++) {
              q_k_dot_f_imag += q_k_imag[d] * forceDensity_k[d][I][0]
                                                                   / normalizeFluidDensity_k;
            }
            q_k_dot_f_imag_over_q_k_sq = q_k_dot_f_imag / q_k_sq;

            /* perform projection */
            /*
for (d = 0; d < num_dim; d++) {
vec_proj_real[d]            = -g_k_imag[d]*g_k_dot_f_imag_over_g_k_sq;
SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_k[d][I][0]
+= vec_proj_real[d];

vec_proj_imag[d]            = g_k_imag[d]*g_k_dot_u_real_over_g_k_sq;
SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidDriftVel_k[d][I][1]
+= vec_proj_imag[d];

}
             */

            pressure_k[I][0] = -q_k_dot_f_real_over_q_k_sq;

            pressure_k[I][1] = -q_k_dot_f_imag_over_q_k_sq;

            /* end check for not zero mode */
          } else { /* for zero mode do the following */
            pressure_k[I][0] = 0.0;
            pressure_k[I][1] = 0.0;
          }

        } /* end k1 loop */
      } /* end k2 loop */
    } /* end k3 loop */

    /* end num_dim  == 3 */
  } else {
    printf("WARNING: %s : %s", error_str_code, error_str_func);
    printf("num_dim = %d \n", num_dim);
    printf(
        "Incompressible computation not implemented in current version of the codes. \n");

    printf("WARNING: %s : %s", error_str_code, error_str_func);
    printf("NOTE: \"Stokes Paradox\" Occurs for Steady-State Incompressible \n");
    printf("Fluid Velocity Fields in 2D! \n");
  }

}


void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::bri1_unitCellRectImageShearPeriodic(double *periodL, double *meshCenterX0,
                                                                                    int shearDir, int shearVelDir, double shearDist,
                                                                                    int num_dim, double *X_orig, double *X_unitCell) {

  const char *error_str_func = "bri1_unitCellRectImageShearPeriodic()";

  IB_appl1_unitCellRectImageShearPeriodic(periodL, meshCenterX0,
                                          shearDir, shearVelDir, shearDist,
                                          num_dim, X_orig, X_unitCell);

}



void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::IB_appl1_unitCellRectImageShearPeriodic(double *periodL, double *meshCenterX0,
                                                                                        int shearDir, int shearVelDir,
                                                                                        double shearDist, int num_dim,
                                                                                        double *X_orig, double *X_unitCell) {

  const char *error_str_func = "IB_appl1_unitCellRectImageShearPeriodic()";

  int d;
  int k_shearDir, k_shearVelDir, k_shearRemDir;

  double XX1[3];

  double L_shearDir;
  double L_shearVelDir;
  double L_shearRemDir;

  double diffShearDir;
  double diffShearVelDir;
  double diffShearRemDir;

  int shearRemDir; /* remaining direction */

  /* We shall use perspective of the system described by
   * the rectangular unit cell which is used
   * to tile space with shifts in the shearVelDir in the
   * shearDir in the tiling.  See Shear SIB paper figures.
   */

  if (shearDir == shearVelDir) {
    stringstream message;

    message << "The shear direction and shear velocity direction" << endl;
    message << "are not allowed to be in the same direction." << endl;
    message << endl;
    message << "Note: One way this error can occur is when one does" << endl;
    message << "not want to use the shear features in a simulation" << endl;
    message << "and simply sets the shear displacement and directions" << endl;
    message << "all set to zero.  Instead, make a valid choice for the" << endl;
    message << "shear and simply be sure to set the shearDist to zero." << endl;
    message << "For example, shearDir = 2, shearVelDir = 0, shearDist = 0." << endl;
    message << endl;
    message << "The values used in the calling routine were:" << endl;
    message << "  shearDir      = " << shearDir << endl;
    message << "  shearVelDir   = " << shearVelDir << endl;
    message << "  shearDist     = " << shearDist << endl;
    message << "  L_shearVelDir = " << L_shearVelDir << endl;

    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  for (d = 0; d < num_dim; d++) {
    if ((d != shearDir) && (d != shearVelDir)) {
      shearRemDir = d;
    }
  }

  L_shearDir = periodL[shearDir];
  L_shearVelDir = periodL[shearVelDir];
  L_shearRemDir = periodL[shearRemDir];

  if (shearDist > L_shearVelDir) {
    stringstream message;
    message << "We assume that shearDist never exceeds domain length." << endl;
    message << "We require shearDist < L_shearVelDir." << endl;
    message << "  shearDist     = " << shearDist << endl;
    message << "  L_shearVelDir = " << L_shearVelDir << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  /* determine the unit cell image of X1 and call it XX1 */
  for (d = 0; d < num_dim; d++) {
    XX1[d] = X_orig[d];
  }

  diffShearDir = XX1[shearDir] - meshCenterX0[shearDir];
  k_shearDir = floor(fabs(diffShearDir / L_shearDir) + 0.5); /* round */
  if (diffShearDir < 0) {
    k_shearDir = -1 * k_shearDir;
  }

  X_unitCell[shearVelDir] = XX1[shearVelDir] - k_shearDir * shearDist; /* adjust cell alignments */
  X_unitCell[shearDir] = XX1[shearDir] - k_shearDir * L_shearDir; /* move point down cell rows */

  diffShearVelDir = X_unitCell[shearVelDir] - meshCenterX0[shearVelDir];
  k_shearVelDir = floor(fabs(diffShearVelDir / L_shearVelDir) + 0.5); /* round */
  if (diffShearVelDir < 0) {
    k_shearVelDir = -1 * k_shearVelDir;
  }

  X_unitCell[shearVelDir] = X_unitCell[shearVelDir] - k_shearVelDir
      * L_shearVelDir; /* move point to unit cell in shearVelDir */

  diffShearRemDir = XX1[shearRemDir] - meshCenterX0[shearRemDir];
  k_shearRemDir = floor(fabs(diffShearRemDir / L_shearRemDir) + 0.5); /* round */
  if (diffShearRemDir < 0) {
    k_shearRemDir = -1 * k_shearRemDir;
  }

  X_unitCell[shearRemDir] = XX1[shearRemDir] - k_shearRemDir * L_shearRemDir; /* move point to unit cell in shearVelDir */

  /* X_unitCell == XX1 */

}



/* == Compute the fluid velocity of the control points for FFTW3 uniform mesh
 with Lees-Edwards boundary conditions. */
void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::computeControlPtsVel_SHEAR_FFTW3(SELM_Lagrangian       *SELM_LagrangianData,
                                                                                        SELM_Eulerian         *SELM_EulerianData,
                                                                                        SELM_CouplingOperator *SELM_CouplingOperatorsData) {

  const char *error_str_func = "computeControlPtsVel_SHEAR_FFTW3()";

  int k;

  /* call the coupling operators */
  SELM_CouplingOperatorsData->computeOperator("GAMMA",
                                              (SELM_Lagrangian *) SELM_LagrangianData,
                                              (SELM_Eulerian *)   SELM_EulerianData);

  /* process the operator data according to the specific type of eulerian and lagrangian data */
  if (SELM_LagrangianData->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {

    SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE
      = (SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *)SELM_LagrangianData;

    /* copy the results of this operation to the control point velocities */
    for (k = 0; k < SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->numEntriesOpGammaVel; k++) {
      SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[k] = SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->opGammaVel[k];
    }
    
  } else if (SELM_LagrangianData->type == SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {    
  
    SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE
      = (SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *)SELM_LagrangianData;

    /* copy the results of this operation to the control point velocities */
    for (k = 0; k < SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->numEntriesOpGammaVel; k++) {
      SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Vel[k] = SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->opGammaVel[k];
    }
  
  } else if (SELM_LagrangianData->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID) {    
  
    SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID
      = (SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *)SELM_LagrangianData;

    /* copy the results of this operation to the control point velocities */
    for (k = 0; k < SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->numEntriesOpGammaVel; k++) {
      SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Vel[k] = SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->opGammaVel[k];
    }

  } else {
    stringstream message;
    message << "No methods implemented to handle this type of Lagrangian data yet." << endl;
    message << "  typeStr = " << SELM_LagrangianData->typeStr << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

}


/* setup for the time step (apply force to the mesh, etc...*/
void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::IB_appl1_start_time_step_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3() {

  const char *error_str_func = "IB_appl1_start_time_step_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3";

  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3                       *SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3;
  //SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE                         *SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE;

  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3::SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_ExtrasType *SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras;

  //SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_ExtrasType *SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras;

  int I;
  int i, j, k, d;
  int N;
  int flagDebugTests = 0;
  int flagDebugStillCompIBForces = 0;

  double *meshCenterX0;
  int *numMeshPtsPerDir;
  double L_shearDir;

  int shearDir;
  int shearVelDir;
  double shearRate;
  double shearDist;

  int num_dim;

  double meshDeltaX;
  
  int flagShearMode;

  #ifdef USE_PACKAGE_FFTW3

  /* == Dereference the data */
  if (fixSELM->SELM_Eulerian_List[0]->type == SELM_Eulerian_Types::TYPE_LAMMPS_SHEAR_UNIFORM1_FFTW3) {

    SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3    = (SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3 *) fixSELM->SELM_Eulerian_List[0];
    SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras = SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3->SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras;

    /* get shear data from appropriate data structure */
    num_dim                               = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->num_dim;

    /* @optimization: much of the below can be simplified here.  Just done for now for safety to sync data structures. */
    flagShearMode = SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagShearMode;
    if (flagShearMode == SHEAR_MODE_TYPE_RM_SHEAR1) {

      ShearData_RM_SHEAR1_Type *shearData_RM_SHEAR1 = (ShearData_RM_SHEAR1_Type *)SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->shearData;

      /* set the shear information for the Eulerian DOF */
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearRate = shearData_RM_SHEAR1->shearRate;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDir = shearData_RM_SHEAR1->shearDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearVelDir = shearData_RM_SHEAR1->shearVelDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist = shearData_RM_SHEAR1->shearDist;

      /* use the shear information consistent with the Eulerian DOF */
      shearRate = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearRate;
      shearDir = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDir;
      shearVelDir = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearVelDir;
      shearDist = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist;
      
    } else if (flagShearMode == SHEAR_MODE_TYPE_RM_OSC1) {

      ShearData_RM_OSC1_Type *shearData_RM_OSC1 = (ShearData_RM_OSC1_Type *)SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->shearData;

      /* set the shear information for the Eulerian DOF */
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearRate = shearData_RM_OSC1->shearRate;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDir = shearData_RM_OSC1->shearDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearVelDir = shearData_RM_OSC1->shearVelDir;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist = shearData_RM_OSC1->shearDist;

      /* use the shear information consistent with the Eulerian DOF */
      shearRate = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearRate;
      shearDir = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDir;
      shearVelDir = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearVelDir;
      shearDist = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist;      

    } else {
      stringstream message;
      message << "Expecting shear mode of type: " << SHEAR_MODE_TYPE_STR_RM_SHEAR1 << endl;
      message << "Instead shear mode was of type: " << SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagShearModeStr << endl;
      SELM_Package::packageError(error_str_code, error_str_func, message);
    }

  } else {
    stringstream message;
    message << "Expecting mesh of type: " << SELM_Eulerian_Types::TYPE_STR_LAMMPS_SHEAR_UNIFORM1_FFTW3 << endl;
    message << "Instead mesh was used of type: " << fixSELM->SELM_Eulerian_List[0]->typeStr <<  endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  /* == Initialize the fluid force density to zero (do this only once) */
  N = 1;
  for (d = 0; d < SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->num_dim; d++) {
    N = N * SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->numMeshPtsPerDir[d];
  }
  for (d = 0; d < SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->num_dim; d++) {
    for (j = 0; j < N; j++) {
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidForceDensity_m[d][j][0] = 0.0;
      SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->fluidForceDensity_m[d][j][1] = 0.0;
    }
  }

  /* @@@ TODO: loop over the coupling operators to allow them to compute how Lagrangian based forces are
   * applied to the fluid through the Lambda operators.
   */
  for (int couplingOpI = 0; couplingOpI < fixSELM->SELM_CouplingOperator_List_N; couplingOpI++) {

    /* == Assumes the control point forces were computed by LAMMPS already. */

    SELM_CouplingOperator *couplingOp = fixSELM->SELM_CouplingOperator_List[couplingOpI];

    /* maybe make below more generic */
    if (couplingOp->type == SELM_CouplingOperator_Types::TYPE_LAMMPS_SHEAR_UNIFORM1_FFTW3_TABLE1) {

      SELM_Lagrangian *lagrangian = NULL;
      SELM_Eulerian   *eulerian   = NULL;

      SELM_CouplingOperator_LAMMPS_SHEAR_UNIFORM1_FFTW3_TABLE1 *op
        = (SELM_CouplingOperator_LAMMPS_SHEAR_UNIFORM1_FFTW3_TABLE1 *)couplingOp;

      /* loop over the coupled entities */
      for (int k = 0; k < op->numCoupleList; k++) {
        lagrangian = op->lagrangianList[k];
        eulerian = op->eulerianList[k];

        /* == Compute the force density arising from the control points. */
        op->computeOperator("LAMBDA",
                            lagrangian,
                            eulerian);

        /* == Dereference the data */
        if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {

          SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE
            = (SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *) fixSELM->SELM_Lagrangian_List[k];

          /* get data from the time stepping data structures */
          //deltaT = this->SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->deltaT;
          
        } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {

          SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE
            = (SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *) fixSELM->SELM_Lagrangian_List[k];

          /* get data from the time stepping data structures */
          //deltaT = this->SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->deltaT;
          
        } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID) {

          SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID
            = (SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *) fixSELM->SELM_Lagrangian_List[k];

          /* get data from the time stepping data structures */
          //deltaT = this->SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->deltaT;          

        } else {
          stringstream message;
          message << "Expecting control points of type: "
                  << SELM_Lagrangian_Types::TYPE_STR_LAMMPS_ATOM_ANGLE_STYLE << endl;
          message << "Instead mesh was used of type: "
                  << lagrangian->typeStr << endl;
          SELM_Package::packageError(error_str_code, error_str_func, message);
        } /* end else */

      } /* end of k loop */

    } /* end if LAMMPS_SHEAR_UNIFORM1_FFTW3_TABLE1 */

  } /* end of loop over coupling operators */

#else

  stringstream message;
  message << "Compiler flag USE_PACKAGE_FFTW3 was not set." << endl;
  message << "None of the FFTW3 routines were not compiled in the codes." << endl;
  SELM_Package::packageError(error_str_code, error_str_func, message);

#endif


}






void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::writeSimulationDataToDisk(char *baseFilename, int timeIndex) {

  const char *error_str_func = "writeSimulationDataToDisk()";

  FILE *fid;
  char  filename[1000];

  /* open the file for writing the data */
  sprintf(filename,"%s_%.9d.SELM_Integrator_%s", baseFilename, timeIndex, typeStr);
  fid = fopen(filename,"w");

  if (fid == NULL) {

    stringstream message;
    message << "Could not open file, error occured." << endl;
    message << "  filename = " << filename << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);

  }

  fprintf(fid, "-- SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3 : Simulation Data -- \n");
  fprintf(fid, "\n");

  fprintf(fid,"flagShearMode %d \n",
          SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3_Extras->flagShearMode);

  /* close the file */
  fclose(fid);
}


void SELM_Integrator_LAMMPS_SHEAR_QUASI_STEADY1_FFTW3::syncShearDomainWithLammpsDomain() {

  const char *error_str_func = "syncShearDomainWithLammpsDomain()";

  //double xy, xz, yz; /* tild vectors in LAMMPS */
  //double xprd, yprd, zprd; /* box dimensions in lammps */
  
  Domain *domain = lammps->domain;
  
  double  shearRate, shearDist, meshDeltaX;
  int     num_dim;
  int    *numMeshPtsPerDir;
  double *meshCenterX0;

  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3                                                       *SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3;
  SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3::SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_ExtrasType *SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras;

  /* get the triclinic flag : 0 - orthogonal, 1 - triclinic box */
  
  /* get data from the integrator */
  /* == Dereference the data */
  if (fixSELM->SELM_Eulerian_List[0]->type == SELM_Eulerian_Types::TYPE_LAMMPS_SHEAR_UNIFORM1_FFTW3) {

    SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3
      = (SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3 *)fixSELM->SELM_Eulerian_List[0];

    SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras
      = SELM_EulerianData_LAMMPS_SHEAR_UNIFORM1_FFTW3->SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras;

    num_dim          = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->num_dim;
    meshDeltaX       = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->meshDeltaX;
    numMeshPtsPerDir = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->numMeshPtsPerDir;
    meshCenterX0     = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->meshCenterX0;

    shearRate        = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearRate;
    shearDist        = SELM_Eulerian_LAMMPS_SHEAR_UNIFORM1_FFTW3_Extras->shearDist;
    
    // ? shearDist = domain->xz ?
       
  } else {
    stringstream message;
    message << "Expecting mesh type: %s." << endl;
    message << "  SELM_Eulerian_Types::TYPE_STR_LAMMPS_SHEAR_UNIFORM1_FFTW3" << endl;
    message << "Instead mesh type was: " << endl;
    message << "  " << fixSELM->SELM_Eulerian_List[0]->typeStr << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  /* only perform synchronization if the shearRate is non-zero */
  if (shearRate != 0) {

    if (domain->triclinic == 0) {
      stringstream message;
      message << "For simulations with a deforming box domain (Lees-Edwards conditions)" << endl;
      message << "There was a non-zero shear rate and shear distance used in integrator" << endl;
      message << "while the LAMMPS codes had an orthogonal box specified." << endl;
      message << "To use this feature LAMMPS must us a triclinic box." << endl;
      message << "lammps->domain->triclinic = " << domain->triclinic << endl;
      SELM_Package::packageError(error_str_code, error_str_func, message);
    }

  } else {

    /* check the box is undeformed in LAMMPS, if undeformed in SELM */
    if (shearDist == 0) {

      if (domain->triclinic == 0) {
        if ((domain->xy != 0.0) && (domain->xz != 0.0) && (domain->yz != 0.0)) {
          stringstream message;
          message << "For simulations with a deforming box domain (Lees-Edwards conditions)" << endl;
          message << "The SELM and LAMMPS deformation must be setup to be the same." << endl;
          message << "This requires shearDist, shearVelDir, shearDir, be consistent with" << endl;
          message << "the xy, xz, yz in LAMMPS" << endl;
          message << "If this feature is not to be used, then setup an orthogonal box in LAMMPS" << endl;
          SELM_Package::packageError(error_str_code, error_str_func, message);
        }
      }  /* end triclinic == 0 */

    } /* end shearDist == 0 */

  } /* end of else */

}




