/* ----------------------------------------------------------------------

  Stochastic Eulerian Lagrangian Methods

  Paul J. Atzberger
  http://atzberger.org/
 
------------------------------------------------------------------------- */

#include <cstdlib>
#include "stddef.h"
#include "stdio.h"
#include "string.h"
#include <malloc.h>

/* SELM_includes */
#include "fix_SELM.h"
#include "fix_SELM_XML_Handler.h"

#include "SELM_Eulerian.h"
#include "SELM_Integrator.h"
#include "SELM_Lagrangian.h"
#include "SELM_Interaction.h"
#include "SELM_CouplingOperator.h"

/* LAMMPS includes */
#include "atom.h"
#include "force.h"
#include "update.h"
#include "respa.h"
#include "error.h"
#include "comm.h"
#include "universe.h"
#include "version.h" 
#include "random_mars.h"

/* include these to help trigger recompile each time */
#include "style_angle.h"
#include "style_atom.h"
#include "style_bond.h"
#include "style_command.h"
#include "style_compute.h"
#include "style_dihedral.h"
#include "style_dump.h"
#include "style_fix.h"
#include "style_improper.h"
#include "style_integrate.h"
#include "style_kspace.h"
#include "style_minimize.h"
#include "style_pair.h"
#include "style_region.h"

using namespace LAMMPS_NS;
using namespace FixConst;

char* FixSELM::error_str_code = "fix_SELM.cpp";

/* =========================== Task list =========================== */
/* TODO time step and synchronize (check dt) */
/* TODO Lees-Edwards domains sync */
/* TODO Parallelization issues (master fftw mesh) */
/* TODO Debug shear quasi-steady code: FENE model or lj model */

/* =========================== Class definitions =========================== */
FixSELM::FixSELM() : Fix(NULL, 0, NULL) {
  /* WARNING: May need to modify LAMMPS codes so that we have
              Fix(NULL, 0, NULL) acts like empty constructor */

  const char *error_str_func = "FixSELM()";

  int flagWarningMessage = 1;

  /* initialize select attributes */
  initAttributes();

  /* create empty class */
  lammps = NULL;

  stringstream message;
  message << "Empty FixSELM created. This should only be used" << endl;
  message << "for testing purposes.  This object does not contain" << endl;
  message << "the needed LAMMPS data structure references." << endl;
  message << "Comment out the error generation if you really want to use this." << endl;
  //SELM_Package::packageWarning(error_str_code, error_str_func, message);
  SELM_Package::packageError(error_str_code, error_str_func, message);

}




/* =========================== Class definitions =========================== */
FixSELM::FixSELM(LAMMPS *lmp, int narg, char **arg) : Fix(lmp, narg, arg)
{

  const char *error_str_func = "FixSELM()";

  char tempStr[1000];
  int  paramFileType;
  int  N;

  /* initialize select attributes */
  initAttributes();

  /* == Display some version information */
  cout << endl;
  cout << "USER-SELM Package (SVN Version = " << SELM_SVN_Version;
  cout << ", Compile Date-Time = " << SELM_Compile_Date_Time << ") " << endl;
  cout << "Simulation Start Date-Time = " << currentDateTime() << endl;

  /* == Save reference to the lammps data */
  lammps = lmp;
  SELM_Package::setLAMMPS(lammps);

  /* == Setup LAMMPS related flags. */
  time_integrate = 1; /* set to 1 for fix performing integration, 0 if fix does not */

  SELM_integrator_mask = 0; /* initial value for the mask for the time step integrator */

  /* == Parse options for the SELM integrator. */
  if (narg < 4) {
    stringstream message;
    message << "Fix SELM requires filename for parameters." <<  endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  /*
    SELM_CouplingOperator_TABLE1 *cp;
    strcpy(tempStr, SELM_CouplingOperator_TABLE1::OPERATOR_TYPE_STR_NULL);
    cp = new SELM_CouplingOperator_TABLE1;
  */

  paramFileType = PARAM_FILE_TYPE_XML;
  switch (paramFileType) {

    case PARAM_FILE_TYPE_TXT:
      /* === Parser for the parameter files for SELM. */
      SELM_parse_ParameterFile_TXT(arg[3]); /* parse the parameter file */
    break;

    case PARAM_FILE_TYPE_XML:
      /* === Parser XML File for the parameters for SELM. */
      SELM_parse_ParameterFile_XML(arg[3]); /* parse the parameter file */
    break;

    default:
      stringstream message;
      message << "The specified parameter file type is not recognized." << endl;
      message << "paramFileType = " << paramFileType << endl;
      SELM_Package::packageError(error_str_code, error_str_func, message);
      break;

  } /* end of switch */

  /* === Setup SELM data structures (parse additional files).
   */

  /* == Setup the random number generator */
  random = new RanMars(lammps, SELM_Seed + comm->me);

  /* == Setup the Lagrangian data structures */
  N = SELM_Lagrangian_List_N;
  for (int I = 0; I < N; I++) {
    SELM_Lagrangian_List[I]->setup();
  }

  /* == Setup the Eulerian data structures */
  N = SELM_Eulerian_List_N;
  for (int I = 0; I < N; I++) {
    SELM_Eulerian_List[I]->setup();
  }

  /* == Setup the CouplingOperators data structures */
  N = SELM_CouplingOperator_List_N;
  for (int I = 0; I < N; I++) {
    SELM_CouplingOperator_List[I]->setup();
  }

  /* == Setup the CouplingOperators data structures */
  N = SELM_Interaction_List_N;
  for (int I = 0; I < N; I++) {
    SELM_Interaction_List[I]->setup();
  }

  /* == Setup the Integrator data structures */
  this->SELM_IntegratorData->setup();

  /* == Write the initial integration data */
  writeAllSimulationData(lammps->update->ntimestep);


  /* == Write some data to XML info file */
  /* record the LAMMPS version and SELM versions
   * among other information */
  writeInfo();

//
//  /* ------------------------------------------------------- */
//  /* == Mesh data structure */
//  /* ------------------------------------------------------- */
//  switch (typeMesh) {
//
//  case SELM_EULERIAN_TYPE_FLUID_SHEAR_UNIFORM1_FFTW3:
//    SELM_EulerianData = new SELM_Eulerian_FLUID_SHEAR_UNIFORM1_FFTW3();
//    break;
//
//  default:
//    printf("ERROR: %s : %s \n", error_str_code, error_str_func);
//    printf("The specified Eulerian degrees of freedom type was not recognized. \n");
//    printf("Some valid types include: \n");
//    printf("%s \n", SELM_EULERIAN_TYPE_STR_FLUID_SHEAR_UNIFORM1_FFTW3);
//    packageError(1, 0);
//    break;
//
//  } /* end typeMesh switch */
//
//  /* -- Parse the data from file */
//  SELM_EulerianData->parse_ParameterFile(SELM_BaseFilename);
//
//  /* -- Setup the mesh */
//  SELM_EulerianData->setup();
//
//  /* ------------------------------------------------------- */
//  /* == Setup the integrator  */
//  /* ------------------------------------------------------- */
//  switch (typeIntegrator) {
//
//  case SELM_INTEGRATOR_TYPE_SHEAR_QUASI_STEADY1_FFTW3:
//    SELM_IntegratorData = new SELM_Integrator_SHEAR_QUASI_STEADY1_FFTW3(lammps, this);
//    break;
//
//  default:
//    printf("ERROR: %s : %s \n", error_str_code, error_str_func);
//    printf("The time stepper type specified was not recognized. \n");
//    printf("Some valid types include: \n");
//    printf("%s \n", SELM_INTEGRATOR_TYPE_STR_SHEAR_QUASI_STEADY1_FFTW3);
//    packageError(1, 0);
//    break;
//
//  } /* end typeIntegrator switch */
//
//  /* -- Parse the data from file */
//  SELM_IntegratorData->parse_ParameterFile(SELM_BaseFilename);
//
//  /* -- Setup the integrator */
//  SELM_IntegratorData->setup();
//
//
//  /* ------------------------------------------------------- */
//  /* == Setup the control points  */
//  /* ------------------------------------------------------- */
//  switch (typeControlPts) {
//
//  case SELM_LAGRANGIAN_TYPE_CONTROLPTS_BASIC1:
//    SELM_LagrangianData = new SELM_Lagrangian_CONTROLPTS_BASIC1(lammps, this);
//    break;
//
//  default:
//    printf("ERROR: %s : %s \n", error_str_code, error_str_func);
//    printf("The control pts type specified was not recognized. \n");
//    printf("Some valid types include: \n");
//    printf("%s \n", SELM_LAGRANGIAN_TYPE_STR_CONTROLPTS_BASIC1);
//    packageError(1, 0);
//    break;
//
//  } /* end typeControlPts switch */
//
//  /* -- Parse the data from file */
//  SELM_LagrangianData->parse_ParameterFile(SELM_BaseFilename);
//
//  /* -- Setup the control points */
//  SELM_LagrangianData->setup();
//
//
//  /* ------------------------------------------------------- */
//  /* == Setup the operators  */
//  /* ------------------------------------------------------- */
//  switch (typeCouplingOperators) {
//
//  case SELM_COUPLINGOPERATORS_TYPE_TABLE1:
//    SELM_CouplingOperatorsData = new SELM_CouplingOperators_TABLE1();
//    break;
//
//  default:
//    printf("ERROR: %s : %s \n", error_str_code, error_str_func);
//    printf("The coupling operators type specified was not recognized. \n");
//    printf("Some valid types include: \n");
//    printf("%s \n", SELM_COUPLINGOPERATORS_TYPE_STR_TABLE1);
//    packageError(1, 0);
//    break;
//
//  } /* end typeControlPts switch */
//
//  /* -- Parse the data from file */
//  SELM_CouplingOperatorsData->parse_ParameterFile(SELM_BaseFilename);
//
//  /* -- Setup the control points */
//  SELM_CouplingOperatorsData->setup();


}

/* destructor */
FixSELM::~FixSELM() {

  /* write some final information to disk */
  writeFinalInfo();

}

void FixSELM::initAttributes() {

  #ifndef SVN_REV
    SELM_SVN_Version        = "(was not available)"; /* indicates not SVN version info. available */
  #else
    SELM_SVN_Version        = SVN_REV;
  #endif

  #ifndef COMPILE_DATE_TIME
    SELM_Compile_Date_Time  = "(was not available)"; /* indicates date and time info. not available */
  #else
    SELM_Compile_Date_Time  = COMPILE_DATE_TIME;
  #endif

  SELM_Run_Description    = "";
  SELM_BasePath           = "";
  SELM_dir_sim_data       = "";
  
  SELM_BaseFilename       = "";

}

void FixSELM::setup(int vflag) { /* lammps setup */

  SELM_IntegratorData->setup_LAMMPS(vflag);

}

/* ---------------------------------------------------------------------- */
int FixSELM::setmask()
{
  /*
    mask |= INITIAL_INTEGRATE;
    mask |= FINAL_INTEGRATE;
   */

  SELM_integrator_mask = 0;

  /* pass to integrator to handle */
  SELM_IntegratorData->set_LAMMPS_mask(&SELM_integrator_mask);

  return SELM_integrator_mask; /* syncronize the SELM mask with that returned to LAMMPS */
}

/* ---------------------------------------------------------------------- */
void FixSELM::pre_exchange()
{

  /* pass to integrator to handle */
  SELM_IntegratorData->pre_exchange();

}

/* ---------------------------------------------------------------------- */
void FixSELM::end_of_step()
{

  /* pass to integrator to handle */
  SELM_IntegratorData->end_of_step();

}

/* ---------------------------------------------------------------------- */

void FixSELM::init()
{

  const char *error_str_func = "init()";


  /* == Initialize the SELM integrators. */

  /* update->integrate->step; (time step from LAMMPS) */

  /* == Check the integration style is Verlet, if not report error. */
  if (strcmp(update->integrate_style, "verlet") != 0) {
    stringstream message;
    message << "SELM requires for now use of the verlet integrate_style." <<  endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  /* == Initialize data structures for SELM. */
  /* integrator trigger fix_init() for any associated initialization */
  SELM_IntegratorData->init_from_fix();

}

/* ----------------------------------------------------------------------
   allow for both per-type and per-atom mass
------------------------------------------------------------------------- */
//void FixSELM::integrate_initialize() {

//  SELM_IntegratorData->integrate_initialize();

//}


void FixSELM::initial_integrate(int vflag)
{

  /* == Get the current information about the box deformation and shear. */


  /* == Get the current information about the particles. */


  /* == Determine the hydrodynamic and particle forces. */

  /* == Perform the update of the fluid and particles */
  SELM_IntegratorData->integrate_initial();

  /* if only initial called then write data to disk here,
   * otherwise wait for the final integration step to be
   * called. */
  if (SELM_integrator_mask & FINAL_INTEGRATE == 0) {
    int timeIndex = lammps->update->ntimestep;
    writeAllSimulationData(timeIndex);
  } else {
    /* wait until integrate_final is called */
  }

}

/* ---------------------------------------------------------------------- */

void FixSELM::final_integrate()
{

  /* == Perform the update of the Eulerian-Lagrangian system */
  SELM_IntegratorData->integrate_final();

  /* == Write the simulation data to disk */
  if (SELM_integrator_mask & FINAL_INTEGRATE == 0) {
    /* do nothing, should not encounter this case. */
    /* maybe report error. */
  } else {
    int timeIndex = lammps->update->ntimestep;
    writeAllSimulationData(timeIndex);
  }

}

/* ---------------------------------------------------------------------- */

void FixSELM::reset_dt()
{

  const char *error_str_func = "reset_dt()";

  stringstream message;
  message << "The fix_SELM does not implement reset_dt() yet." <<  endl;
  SELM_Package::packageError(error_str_code, error_str_func, message);

}

void FixSELM::post_force(int vflag) {
  SELM_IntegratorData->post_force(vflag);
}


/*****************************************************************************************/
/* Supporting SELM codes */
/*****************************************************************************************/
void FixSELM::packageError(int code, void *extras) {
    std::exit(code);
}



/* parses the parameters from a file */
void FixSELM::SELM_parse_ParameterFile_XML(char *filename) {

  const char *error_str_func = "SELM_parse_ParameterFile_XML()";

  FixSELM *fixSELM_Data = NULL; /* for testing */

  fix_SELM_XML_Handler *fix_SELM_DataHandler
    = new fix_SELM_XML_Handler(this);

  Atz_XML_SAX_Handler_Multilevel *dataHandler
    = new Atz_XML_SAX_Handler_Multilevel(fix_SELM_DataHandler);

  Atz_XML_Parser::parse(filename, dataHandler);

  fixSELM_Data = (FixSELM *) fix_SELM_DataHandler->XML_getData();

}

/* parses the parameters from a file */
void FixSELM::SELM_parse_ParameterFile_TXT(char *filename) {

  const char *error_str_code = "fix_SELM.c";
  const char *error_str_func = "SELM_parse_ParameterFile_TXT()";

//  FILE *fid;
//
//  char tempStr[1000];
//  char c;
//  int  k;
//
//  /* open the file */
//  fid = fopen(filename, "r");
//
//  if (fid == 0) {
  //    printf("ERROR: %s : %s \n", error_str_code, error_str_func);
//    printf("Could not open file, error occured. \n");
//    printf("  filename = %s \n", filename);
//    packageError(1, 0);
//  }
//
//  /* read the first line of the text file. */
//  /* this contains comments (# symbol in given example file). */
//  c = 0;
//  while (c != '\n') {
//    fscanf(fid, "%c", &c);
//  }
//
//  /* read the second line of the text file. */
//  /* this contains comments (# symbol in given example file). */
//  c = 0;
//  while (c != '\n') {
//    fscanf(fid, "%c", &c);
//  }
//
//  /* read the version of the file */
//  fscanf(fid,"%s",  (char *)tempStr);
//  fscanf(fid,"%d",  &SELM_Version);
//
//  /* read user defined description of the run being performed */
//  fscanf(fid,"%s",  (char *)tempStr);
//  fscanf(fid,"%d",  &SELM_Run_DescriptionLength);
//
//  SELM_Run_Description = (char *)malloc(sizeof(char)*SELM_Run_DescriptionLength + 1);
//
//  fscanf(fid,"%s",  (char *)tempStr);
//  for (k = 0; k < SELM_Run_DescriptionLength + 3; k++) { /* +3 for: space + { + } */
//    fscanf(fid,"%c",&c);
//    if ((k > 1) && (k < SELM_Run_DescriptionLength + 2)) {
//      SELM_Run_Description[k - 2] = c;
//    }
//  }
//  SELM_Run_Description[SELM_Run_DescriptionLength] = 0;
//
//  /* Base filename */
//  fscanf(fid,"%s",  (char *) tempStr);
//  fscanf(fid,"%s",  (char *) SELM_BaseFilename);
//
//  /* random Seed value */
//  fscanf(fid,"%s",  (char *)tempStr);
//  fscanf(fid,"%d",  &SELM_Seed);
//
//  /* parse the mesh related data */
//  fscanf(fid, "%s", (char *) tempStr);
//  fscanf(fid, "%s", (char *) typeStrMesh);
//
//  typeMesh = -1;
//  if (strcmp(typeStrMesh, SELM_EULERIAN_TYPE_STR_FLUID_SHEAR_UNIFORM1_FFTW3) == 0) {
//    typeMesh = SELM_EULERIAN_TYPE_FLUID_SHEAR_UNIFORM1_FFTW3;
//  }
//
//  /* parse the integrator related data */
//  fscanf(fid, "%s", (char *) tempStr);
//  fscanf(fid, "%s", (char *) typeStrIntegrator);
//
//  typeIntegrator = -1;
//  if (strcmp(typeStrIntegrator, SELM_INTEGRATOR_TYPE_STR_SHEAR_QUASI_STEADY1_FFTW3) == 0) {
//    typeIntegrator = SELM_INTEGRATOR_TYPE_SHEAR_QUASI_STEADY1_FFTW3;
//  }
//
//  /* parse the control points related data */
//  fscanf(fid, "%s", (char *) tempStr);
//  fscanf(fid, "%s", (char *) typeStrControlPts);
//
//  typeControlPts = -1;
//  if (strcmp(typeStrControlPts, SELM_LAGRANGIAN_TYPE_STR_CONTROLPTS_BASIC1) == 0) {
//    typeControlPts = SELM_LAGRANGIAN_TYPE_CONTROLPTS_BASIC1;
//  }
//
//  /* parse the coupling operator related data */
//  fscanf(fid, "%s", (char *) tempStr);
//  fscanf(fid, "%s", (char *) typeStrCouplingOperators);
//
//  typeCouplingOperators = -1;
//  if (strcmp(typeStrCouplingOperators, SELM_COUPLINGOPERATORS_TYPE_STR_TABLE1) == 0) {
//    typeCouplingOperators = SELM_COUPLINGOPERATORS_TYPE_TABLE1;
//  }
//
//  /* close the parameter file */
//  fclose(fid);

}

void FixSELM::writeAllSimulationData(int timeIndex) {

  //char baseFilename[1000];
  //char baseFilename[1000];
    
  //sprintf(baseFilename,"%s",SELM_BaseFilename);
  
  char *dir_output = this->SELM_dir_sim_data;
  
  // write the Lagrangian DOF to disk
  for (int k = 0; k < SELM_Lagrangian_List_N; k++) {
    if ( (this->SELM_Lagrangian_List[k]->flagWriteSimulationData) &&
         (timeIndex % this->SELM_Lagrangian_List[k]->saveSkipSimulationData == 0) ) {
      this->SELM_Lagrangian_List[k]->writeSimulationDataToDisk(dir_output, // WARNING: changed from this->SELM_BaseFilename (might need to change downstream codes)
                                                               timeIndex);
    }
  } // end k loop

  // write the Eulerian DOF to disk
  for (int k = 0; k < SELM_Eulerian_List_N; k++) {
    if ( (this->SELM_Eulerian_List[k]->flagWriteSimulationData) &&
         (timeIndex % this->SELM_Eulerian_List[k]->saveSkipSimulationData == 0) ) {
      this->SELM_Eulerian_List[k]->writeSimulationDataToDisk(dir_output, // WARNING: changed from this->SELM_BaseFilename (might need to change downstream codes)
                                                             timeIndex);
    }
  } // end k loop

  // write the Integrator data to disk
  if ( (this->SELM_IntegratorData->flagWriteSimulationData) &&
       (timeIndex % this->SELM_IntegratorData->saveSkipSimulationData == 0) ) {
    this->SELM_IntegratorData->writeSimulationDataToDisk(dir_output, // WARNING: changed from this->SELM_BaseFilename (might need to change downstream codes)
                                                         timeIndex);
  } // end if

//  /* write control points data to disk */
//  if ((SELM_LagrangianData->flagWriteSimulationData) &&
//      (timeIndex % SELM_LagrangianData->saveSkipSimulationData == 0)) {
//    SELM_LagrangianData->writeSimulationDataToDisk(baseFilename, timeIndex);
//  }
//
//  /* write mesh data to disk */
//  if ((SELM_EulerianData->flagWriteSimulationData) &&
//      (timeIndex % SELM_EulerianData->saveSkipSimulationData == 0)) {
//    SELM_EulerianData->writeSimulationDataToDisk(baseFilename, timeIndex);
//  }
//
//  /* write coupling operator data to disk */
//  if ((SELM_CouplingOperatorsData->flagWriteSimulationData) &&
//      (timeIndex % SELM_CouplingOperatorsData->saveSkipSimulationData == 0)) {
//    SELM_CouplingOperatorsData->writeSimulationDataToDisk(baseFilename, timeIndex);
//  }
//
//  /* write integrator data to disk */
//  if ((SELM_IntegratorData->flagWriteSimulationData) &&
//      (timeIndex % SELM_IntegratorData->saveSkipSimulationData == 0)) {
//    SELM_IntegratorData->writeSimulationDataToDisk(baseFilename, timeIndex);
//  }

}

void FixSELM::writeInfo() {

  char filename[10000];

  const char *error_str_func = "writeInfo()";

  FILE *fid;

  /* open the file for writing the data */
  sprintf(filename, "%s.SELM_Info", SELM_BaseFilename);
  fid = fopen(filename,"w");

  if (fid == NULL) {
    stringstream message;
    message << "Could not open file to write error occured." << endl;
    message << "  filename = " << filename << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  /* write XML format */
  stringstream output;
  output << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << endl;
  output << "<SELM_Info>" << endl;
  //output << "<LAMMPS_Version value=" << "\"" <<  lammps->universe->version << "\"" << "/>" << endl;
  output << "<LAMMPS_Version value=" << "\"" << LAMMPS_VERSION << "\"" << "/>" << endl;  // PJA: new version of LAMMPS defines hard-code
  output << "<SELM_SVN_Version value=" << "\"" <<  SELM_SVN_Version << "\"" << "/>" << endl;
  output << "<SELM_Compile_Date_Time value=" << "\"" <<  SELM_Compile_Date_Time << "\"" << "/>" << endl;
  output << "<Simulation_Start_Date_Time value=" << "\"" <<  currentDateTime() << "\"" << "/>" << endl;
  output << "</SELM_Info>" << endl;

  /* write the output to disk */
  fprintf(fid, output.str().c_str());

   /* close the file */
  fclose(fid);

}

void FixSELM::writeFinalInfo() {

  char filename[10000];

  const char *error_str_func = "writeInfo()";

  FILE *fid;

  /* open the file for writing the data */
  sprintf(filename, "%s.SELM_InfoExtra", SELM_BaseFilename);
  fid = fopen(filename,"w");

  if (fid == NULL) {
    stringstream message;
    message << "Could not open file to write error occured." << endl;
    message << "  filename = " << filename << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  /* delete final line of current file */

  /* write some additional information */
  /* write XML format */
  stringstream output;
  output << "<SELM_InfoExtra>" << endl;
  output << "<Simulation_Stop_Date_Time value=" << "\"" <<  currentDateTime() << "\"" << "/>" << endl;
  output << "</SELM_InfoExtra>" << endl;

  /* write the output to disk */
  fprintf(fid, output.str().c_str());

  /* close the file */
  fclose(fid);

}

// Get current date/time, (format such as YYYY-MM-DD.HH:mm:ss)
const string FixSELM::currentDateTime() {

  time_t     now = time(0);
  struct tm  tstruct;

  char       buf[80];

  tstruct = *localtime(&now);

  // Visit http://www.cplusplus.com/reference/clibrary/ctime/strftime/
  // for more information about date/time format
  //strftime(buf, sizeof(buf), "%Y-%m-%d.%X", &tstruct);
  strftime(buf, sizeof(buf), "%m-%d-%Y %X", &tstruct);

  return buf;
}




