/* ----------------------------------------------------------------------

  Stochastic Eulerian Lagrangian Methods

  Paul J. Atzberger
  http://atzberger.org/
 
------------------------------------------------------------------------- */

#ifdef FIX_CLASS

FixStyle(SELM,FixSELM)

#else

#ifndef LMP_FIX_SELM_H
#define LMP_FIX_SELM_H

#ifndef FFT_FFTW /* determine if FFTW package specified */
  #warning "No FFTW package specified for SELM codes.  The SELM functionality will be disabled."
#else
  #define USE_PACKAGE_FFTW3
#endif


#include "mpi.h"
#include "fix.h"
#include "stddef.h"
#include <cstdlib>
#include "stdio.h"
#include "string.h"
#include "SELM_Eulerian.h"
#include "SELM_Integrator.h"
#include "SELM_Lagrangian.h"
#include "SELM_CouplingOperator.h"
#include "SELM_Interaction.h"
#include "random_mars.h"
#include "lammps.h"

#include "Atz_XML_Package.h"

#ifdef USE_PACKAGE_FFTW3
  #include <fftw3.h>
#endif

using namespace std;

namespace LAMMPS_NS {

class FixSELM : public Fix {

 public:

  static char  *error_str_code;

  /* ================= constants ================= */
  /* constants (values for non-integers defined in .cpp) */
  static const int MAX_STR_LEN = 10000;

  /* parameter file types */
  static const int    PARAM_FILE_TYPE_NULL       = 0;
  static const int    PARAM_FILE_TYPE_TXT        = 1;
  static const int    PARAM_FILE_TYPE_XML        = 2;

  /* ================= function prototypes ================= */
  FixSELM(class LAMMPS *, int, char **);

  FixSELM(); /* used to construct empty object,
                primarily for testing purposes */
  ~FixSELM();

  void         initAttributes();

  int          setmask();
  virtual void init();
  void         setup(int vflag);
  virtual void initial_integrate(int);
  virtual void final_integrate();
  void         reset_dt();
  void         post_force(int vflag);
  
  // additional methods/hooks
  void pre_exchange();
  void end_of_step();

  /* =========================== Function Calls =========================== */

  /* Generates messages when encountering errors. */
  void packageError(int code, void *extras);

  /* SELM parse parameter file */
  void SELM_parse_ParameterFile_TXT(char *filename);

  /* SELM parse parameter file */
  void SELM_parse_ParameterFile_XML(char *filename);

  /* write the SELM simulation data to disk */
  void writeAllSimulationData(int timeIndex);
  void writeInfo();
  void writeFinalInfo();
  const string currentDateTime();

  /* =========================== Variables =========================== */
  LAMMPS                                                    *lammps; /* lammps data */
  int                                                        SELM_integrator_mask; /* mask set by SELM codes */

  int                                                        SELM_Version;
  char*                                                      SELM_SVN_Version;
  char*                                                      SELM_Compile_Date_Time;
  int                                                        SELM_Run_DescriptionLength;
  char                                                      *SELM_Run_Description;

  char                                                      *SELM_BasePath;
  char                                                      *SELM_dir_sim_data;
  char                                                      *SELM_BaseFilename;

  int                                                        SELM_Seed;
  RanMars                                                   *random; /* random number generator */

  int                                                        SELM_Eulerian_List_N;
  SELM_Eulerian                                            **SELM_Eulerian_List;

  int                                                        SELM_Lagrangian_List_N;
  SELM_Lagrangian                                          **SELM_Lagrangian_List;

  int                                                        SELM_CouplingOperator_List_N;
  SELM_CouplingOperator                                    **SELM_CouplingOperator_List;

  int                                                        SELM_Interaction_List_N;
  SELM_Interaction                                         **SELM_Interaction_List;

  SELM_Integrator                                           *SELM_IntegratorData;

};

}

#endif
#endif
