/* ----------------------------------------------------------------------

  Stochastic Eulerian Lagrangian Methods

  Paul J. Atzberger
  http://atzberger.org/
 

------------------------------------------------------------------------- */

#include <cstdlib>
#include "stddef.h"
#include "stdio.h"
#include "string.h"
#include <malloc.h>

/* SELM_includes */
#include "fix_SELM_statistic_MSD.h"

/* LAMMPS includes */
#include "atom.h"
#include "force.h"
#include "update.h"
#include "respa.h"
#include "error.h"
#include "comm.h"
#include "random_mars.h"

using namespace LAMMPS_NS;
using namespace FixConst;

char* FixSELM_statistic_MSD::error_str_code = "fix_SELM_statMSD.cpp";

/* =========================== Task list =========================== */
/* TODO compute the MSD(t) statistics                                */
/* TODO write the MSD statistics to disk periodically                */

/* =========================== Class definitions =========================== */
FixSELM_statistic_MSD::FixSELM_statistic_MSD() : Fix(NULL, 0, NULL) {
  /* WARNING: May need to modify LAMMPS codes so that we have
              Fix(NULL, 0, NULL) acts like empty constructor */

  const char *error_str_func = "FixSELM_statistic_MSD()";

  int flagWarningMessage = 1;

  /* initialize select attributes */
  //initAttributes();

  /* create empty class */
  lammps = NULL;

  if (flagWarningMessage) {
    printf("WARNING: %s : %s \n", error_str_code, error_str_func);
    printf("  Empty FixSELM_statistic_MSD created. This should only be used \n");
    printf("  for testing purposes.  This object does not contain \n");
    printf("  the needed LAMMPS data structure references. \n");
  }

}


/* =========================== Class definitions =========================== */
FixSELM_statistic_MSD::FixSELM_statistic_MSD(LAMMPS *lmp, int narg, char **arg) : Fix(lmp, narg, arg)
{

  const char *error_str_func = "FixSELM_statistic_MSD()";

  char tempStr[1000];
  int  N;

  /* initialize select attributes */
  //initAttributes();

  /* == Save reference to the lammps data */
  lammps = lmp;
  SELM_Package::setLAMMPS(lammps);

  /* == Setup LAMMPS related flags. */
  time_integrate = 0; /* set to 1 for fix performing integration, 0 if fix does not */

  /* == Parse options for the SELM integrator. */
  if (narg < 7) {
	stringstream message;
	message << "Fix_SELM_statMSD requires filename for parameters." <<  endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  /* show the arguments specified */
  cout << "Fix_SELM_statMSD:  nevery = " <<  arg[3] << endl;
  cout << "Fix_SELM_statMSD:  ndelaysteps = " <<  arg[4] << endl;
  cout << "Fix_SELM_statMSD:  nfreq = " <<  arg[5] << endl;
  cout << "Fix_SELM_statMSD:  str = " <<  arg[6] << endl;

  nevery      = atoi(arg[3]);
  ndelaysteps = atoi(arg[4]);
  nfreq       = atoi(arg[5]);

  nvalid      = 1;

  global_freq = nfreq;
  time_depend = 1;

  /* filename = arg[3] */

/*
  nevery = atoi(arg[3]);
  nrepeat = atoi(arg[4]);
  nfreq = atoi(arg[5]);
 *
 */

/*      stringstream message;
      message << "The specified parameter file type is not recognized." << endl;
      message << "paramFileType = " << paramFileType << endl;
      SELM_Package::packageError(error_str_code, error_str_func, message);
      break;
*/

  /* === Setup SELM data structures (parse additional files).
   */

  /* -- open file for appending data during the simulation at statistics are
   * computed.
   */

  /* == Write the initial integration data */
  //writeAllSimulationData(lammps->update->ntimestep);



}

/* ---------------------------------------------------------------------- */
int FixSELM_statistic_MSD::setmask()
{
  /*
    mask |= INITIAL_INTEGRATE;
    mask |= FINAL_INTEGRATE;
   */

  int mask = 0;

  mask |= END_OF_STEP;

  cout << "Fix_SELM_statMSD:  setmask = " <<  mask << endl;

  return mask;
}

/* ---------------------------------------------------------------------- */

void FixSELM_statistic_MSD::init()
{

  const char *error_str_func = "init()";

  /* fix_ave_time.cpp does following */
  // set current indices for all computes,fixes,variables

  /* == Initialize the SELM integrators. */

  /* update->integrate->step; (time step from LAMMPS) */

  /* == Initialize data structures for SELM. */

  cout << "Fix_SELM_statMSD:  init() " << endl;

  Atom *atom  = lammps->atom;

  int N       = atom->natoms;
  int num_dim = 3;

  this->x_old  = (double *)malloc(sizeof(double)*N*num_dim);
  this->deltaX = (double *)malloc(sizeof(double)*N*num_dim);

  /* set old positions */
  for (int k = 0; k < N; k++) {
    x_old[k*num_dim + 0] = atom->x[0][k];
	x_old[k*num_dim + 1] = atom->x[1][k];
	x_old[k*num_dim + 2] = atom->x[2][k];
  }

  nvalid += ndelaysteps; /* set next valid time step */

}


/* ----------------------------------------------------------------------
   only does something if nvalid = current timestep
------------------------------------------------------------------------- */


void FixSELM_statistic_MSD::setup(int vflag)
{
	end_of_step();

	cout << "Fix_SELM_statMSD:  setup() " << endl;
}


/* ---------------------------------------------------------------------- */

void FixSELM_statistic_MSD::end_of_step()
{

  Atom *atom = lammps->atom;
  int num_dim = 3;
  int N = atom->natoms;

  // skip if not step which requires doing something

  bigint ntimestep = update->ntimestep;
  //if (ntimestep != nvalid) return;
  if (ntimestep % this->ndelaysteps != 0) return;

  //if (mode == SCALAR) invoke_scalar(ntimestep);
  //else invoke_vector(ntimestep);

  cout << "Fix_SELM_statMSD:  end_of_step() " << endl;

  cout << "Fix_SELM_statMSD:  timestep = " << ntimestep << endl;

  /* compute the change in position */
  /* WARNING: need to handle the periodic boundary conditions!!!! */
  for (int k = 0; k < N; k++) {
    deltaX[k*num_dim + 0] = atom->x[0][k] - x_old[k*num_dim + 0];
    //if (fabs(deltaX[k*num_dim + 0]) > L/2.0) {  /* find position mod L/2 */
    //} 
    deltaX[k*num_dim + 1] = atom->x[1][k] - x_old[k*num_dim + 1];
    deltaX[k*num_dim + 2] = atom->x[2][k] - x_old[k*num_dim + 2];
  }

  /* set old positions */
  for (int k = 0; k < N; k++) {
    x_old[k*num_dim + 0] = atom->x[0][k];
	x_old[k*num_dim + 1] = atom->x[1][k];
	x_old[k*num_dim + 2] = atom->x[2][k];
  }

  cout << "Fix_SELM_statMSD:  deltaX[0] = " << deltaX[0] << endl;
  cout << "Fix_SELM_statMSD:  deltaX[1] = " << deltaX[1] << endl;
  cout << "Fix_SELM_statMSD:  deltaX[2] = " << deltaX[2] << endl;

}



/*****************************************************************************************/
/* Supporting SELM codes */
/*****************************************************************************************/

void FixSELM_statistic_MSD::writeAllSimulationData(int timeIndex) {

	/*
  char baseFilename[1000];

  sprintf(baseFilename, "%s", SELM_BaseFilename);

  // write the Lagrangian DOF to disk
  for (int k = 0; k < SELM_Lagrangian_List_N; k++) {
    //if (this->SELM_Lagrangian_List[k]->flagWriteSimulationData) {
      this->SELM_Lagrangian_List[k]->writeSimulationDataToDisk(this->SELM_BaseFilename, timeIndex);
    //}
  } // end k loop

   // write the Eulerian DOF to disk
  for (int k = 0; k < SELM_Eulerian_List_N; k++) {
    //if (this->SELM_Eulerian_List[k]->flagWriteSimulationData) {
      this->SELM_Eulerian_List[k]->writeSimulationDataToDisk(this->SELM_BaseFilename,
                                                             timeIndex);
    //}
  } // end k loop
   */

}






